---------------------------------------------------------------------------
--- LOCAL VARS ------------------------------------------------------------
---------------------------------------------------------------------------

-- What section to use in page headers in this file.
local thisPageSection = "CATEGORIES"                         

-- Error messages
local ERR_AUTH_NOADDPRIV = [[You are not allowed to add new categories to the system. Please check with your administrator.]]
local ERR_AUTH_NODELPRIV = [[You are not allowed to delete categories from the system. Please check with your administrator.]]
local ERR_AUTH_NOEDITPRIV = [[You are not allowed to edit categories in the system. Please check with your administrator.]]
local ERR_CAT_DUPCATEGORY = [[A category by this name already exists.]]
local ERR_DEL_NOCATEGORIES = [[You did not select any categories to delete.]]
local ERR_DEL_NOFINDCATEGORY = [[Unable to find the category to delete. It may have been removed in another session.]]
local ERR_CAT_NOCATID = [[An error has occured. I don't know what category you're trying to modify. Please try again.]]
local ERR_CAT_BADPROPNAME = [[The category name you have specified is invalid. Please try again.]]
local SUCC_CAT_NOCHANGE = [[No changes have been made.]]

-- The number of category fields to include on the add new category page.
local ADD_CATEGORY_FIELD_COUNT = 5

---------------------------------------------------------------------------
--- FUNCTIONS -------------------------------------------------------------
---------------------------------------------------------------------------

-- Display the total number of categories in the system.
local displayCategoryCount = function (db)
    -- Try to determine the count of items in the category list. If this
    -- doesn't work, then just leave.
    local totalCategories = rowsInTable (db , "CategoryList")
    if totalCategories == nil then return end

    -- Subtract 1 from the category count, because the category list 
    -- specifies the "None" category that isn't really part of the
    -- system.
    totalCategories = totalCategories - 1

    -- Figure out the proper grammar for the display.
    local label1 = "are "
    local label2 = " categories"
    if totalCategories == 1 then label1 = "is " label2 = " category" end

    -- Display it
    html:addText ('<FONT SIZE="-1">')
    html:addText ("There " , label1, totalCategories , label2, " registered with the system.")
    html:addText ("</FONT><P>")
end

---------------------------------------------------------------------------

-- Generate the form that lists current categories and the ability to delete them
-- from the system. A properties link is also supported, if possible.
local generateCategoryList = function (db)
    -- Say how many categories there are. 
    displayCategoryCount (db)

    -- If the user has javascript enabled, this confirmation will be used to get them to
    -- confirm what they are about to do.
    local deleteConfirm = "return confirm ('Any selected categories will be completely removed from the system. Are you sure this is what you want to do?');"

    -- Open the form that will allow for the deletion of categories
    html.form:open (env.scriptName , "POST", nil, nil, deleteConfirm)

    -- Open a table for the values
    html.table:open (nil, nil, nil, nil, "5", "categoryList")

    -- Create an alias for the user's permission to edit categories.
    local canEdit = userInfo.canEditCategory 

    -- Display a heading for all of the rows. Start with the
    -- name of the category.
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Category")

    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Games")

    -- If this user has edit powers, then we need two more columns.
    if canEdit then
        -- This column will hold the properties link.
        html.table:heading (true, nil, nil, "LEFT", "TOP")
        html:addText ("&nbsp;")

        -- This column will hold the delete checkbox.
        html.table:heading (true, nil, nil, "LEFT", "TOP")
        html:addText ("Delete")
    end

    -- Get ready for the next row
    html.table:newRow ()

    -- Get the complete list of categories sorted by name and excluding the
    -- default "None" category.
    --
    -- Iterate through all of them.
    local catList = getCategoryList (db, true, true)
    for index,row in catList do
        -- Include the name of this category
        html.table:cell (true, nil, nil, "LEFT" , "TOP")
        html:addText (row.categoryName)

        -- Include how many games there are in this category
        html.table:cell (true, nil, nil, "LEFT" , "TOP")

        -- If we can get a count of games that are in this category, then
        -- display a link that allows the viewing of all games in this
        -- category.
        local count = countGamesInCategory (db, row.categoryID)
        if count ~= nil then
            html:link (urls.gamesByCategory .. row.categoryID , count)
        else
            html:addText ("N/A")
        end

        -- Two extra columns for users with edit permissions.
        if canEdit then
            -- Output a properties link for this category
            html.table:cell (true, nil, nil, "LEFT" , "TOP")
            html:link (urls.categoryProperties .. row.categoryID , "Properties")

            -- Output a checkbox to delete this item
            html.table:cell (true, nil, nil, "LEFT" , "TOP")
            html.form:checkbox ("delCategoryID" , row.categoryID)
        end

        -- Get ready for the next row
        html.table:newRow ()
    end

    -- If the category list has at least 1 item and this user can edit, then output the
    -- form button that will allow deletions. This isn't neccesary if there are no categories
    -- or if the current user has no permissions.
    if table.getn (catList) > 0 and canEdit then
        -- Finish up with a submit button.
        html.table:cell (true, nil, nil, "CENTER", "TOP", nil, nil, nil, "3")
        html.form:submit ("Button", "Delete selected Categories", "category")
        html.form:hidden ("command", "CategoryDelete")
    end

    -- Close the table and form now
    html.table:close ()
    html.form:close ()
end

---------------------------------------------------------------------------

-- Generate the form that allows you to add new categories to the system.
local generateAddForm = function (db)
    -- Output a blank line of the same size as the one that the category
    -- list uses to show how many categories that there are. 
    html:addText ('<FONT SIZE="-1">&nbsp;</FONT><P>')

    -- Open the form that will allow for the addition of categories
    html.form:open (env.scriptName , "POST")

    -- Open a table for the values
    html.table:open (nil, nil, nil, nil, "5", "categoryList")

    -- Display a heading for all of the rows
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Category Name")

    -- Get ready for the next row
    html.table:newRow ()

    -- Generate a bunch of text entry fields. The number is configurable
    -- at the top of this file.
    for i = 1,ADD_CATEGORY_FIELD_COUNT do
        html.table:cell (true, nil, nil, "LEFT" , "TOP")
        html.form:text ("newCategory" .. i , "" , "40" , "60")
        html.table:newRow ()
    end

    -- Finish up with a submit button.
    html.table:cell (true, nil, nil, "CENTER", "TOP")
    html.form:submit ("Button", "Add Categories", "category")
    html.form:hidden ("command", "CategoryAdd")

    -- Close the table and form now
    html.table:close ()
    html.form:close ()
end

---------------------------------------------------------------------------

-- Takes some categoryInfo (ID and name, as in the CategoryList table), and
-- then finds all games that have that category as the primary category. For
-- any that do, the primary category is replaced with the first match in
-- the CategoryPlacement table. If the game has only one category, then
-- the category is set to None.
local fixCategorizedGames = function (db , categoryInfo)
    -- We are removing the category in categoryInfo. So, find all games for which
    -- that category is the primary category. 
    local foundGames = getGamesWithPrimaryCategory (db, categoryInfo.categoryID)

    -- Now loop over the results that we saved.
    for index,row in foundGames do
        -- For this game, find the first row in CategoryPlacement for this game.
        local secondaryCategory = getFirstSecondaryCategoryForGame (db, row.gameID)

        -- Now, if secondaryCategory is nil, then this game had only a single category. In that
        -- case, set the category in this game to be 0 (None).
        if secondaryCategory == nil then
            -- No secondary category, so just update this game to have a category that is empty.
            alterGamePrimaryCategory (db, row.gameID, 0)
            html:addText ("Primary category of " , row.name , " has been set to 'None'<BR>")
        else
            -- Here ther WAS a secondary category. First of all, delete this row from the
            -- CategoryPlacement table. Then, update the GameList entry for this game to have
            -- the secondary category as primary.
            deleteCategoryPlacementByPlacement (db, secondaryCategory.placementID)
            alterGamePrimaryCategory (db, row.gameID, secondaryCategory.categoryID)

            -- Figure out the new primary and say what happend.
            local newPrimary = getCategoryByID (db, secondaryCategory.categoryID)
            if newPrimary then
                html:addText ("Primary category of " , row.name , " has been set to '" , newPrimary.categoryName , "'<BR>")
            end
        end
    end
end

---------------------------------------------------------------------------

-- Called to process the removal of categories from the system.
local processDeleteForm = function (db)
    -- Make sure before we process anything that this user is allowed to
    -- remove categories from the system. 
    if userInfo.canEditCategory == false then
        addLog (db, "Attempted to delete categories")
        html:addText (errorText (ERR_AUTH_NODELPRIV) , "<P>")
        return 
    end

    -- Get the arguments that we need. The delete form is only handled when
    -- it is actually submitted, so it's an error if there is no categories
    -- listed for deletion.
    local args = collectFormArguments ({"delCategoryID"})
    if args == nil then
        addLog (db, "Attempted delete with no categories selected")
        html:addText (errorText (ERR_DEL_NOCATEGORIES) , "<P>")
        return
    end

    -- The form field could be a string if only one thing is being deleted,
    -- or a table for more. Make sure it's always a table, to make the code
    -- below easier to follow.
    if type (args.delCategoryID) ~= "table" then
        args.delCategoryID = {args.delCategoryID}
    end

    -- Loop over and remove all categories given. Confirmation was received
    -- before the form submitted.
    for index,auth in args.delCategoryID do
        -- Take care not to remove the category with ID 0 ("None"). It's not
        -- listed in the form ever, but better safe than sorry.
        if tonumber (auth) ~= 0 then
            -- Try to get information on the category that is being removed.
            local delCategoryInfo = getCategoryByID (db, auth)

            -- Complain if we can't find the category to remove.
            if delCategoryInfo == nil then
                html:addText (errorText (ERR_DEL_NOFINDCATEGORY) , "<BR>")
            else
                -- We're deleting this category. Firstly, remove all mention of this category
                -- from the CategoryList and the CategoryPlacement tables.
                deleteCategoryByID (db, auth)
                deleteCategoryPlacementsByCategory (db, auth)

                -- Say what category was deleted
                html:addText (successText ("Deleted category '" .. delCategoryInfo.categoryName .. "'") , "<BR>")
                addLog (db , "Deleted category '" .. delCategoryInfo.categoryName .. "'")

                -- Now fix all games which have this category as the primary category. We do this AFTER
                -- the above so that we can output status text saying what category becomes what, and such.
                -- Once that's done, seperate this output from everything else.
                fixCategorizedGames (db , delCategoryInfo)
                html:addText ("<P>")
            end
        end
    end
end

---------------------------------------------------------------------------

-- Called to process the addition of new categories to the system.
local processAddForm = function (db)
    -- First of all, make sure that this user is allowed to edit categories
    -- before we do anything. This can only happen if someone hand-makes a
    -- URL, since the system won't present the form if the user doesn't have
    -- the correct permissions.
    if userInfo.canEditCategory == false then
        addLog (db, "Attempted to add categories")
        html:addText (errorText (ERR_AUTH_NOADDPRIV) , "<P>")
        return 
    end

    -- Figure out the names of the form arguments. The number of arguments is
    -- configurable at the top of this file.
    local argsNeeded = {}
    for i = 1,ADD_CATEGORY_FIELD_COUNT do
        table.insert (argsNeeded , "newCategory" .. i)
    end

    -- Try to get the form arguments. If we didn't get them, then we can just leave
    -- now without doing anything. This isn't an error since it just means that the
    -- form wasn't submitted.
    local args = collectFormArguments (argsNeeded)
    if args == nil then return end

    -- Check over all 5 of the categories and see if we should add them. For each one,
    -- we make sure there is not already a category by this name. We also keep track of
    -- all added items, so that we can report on it later.
    local newCategories = {}
    for i = 1,ADD_CATEGORY_FIELD_COUNT do
        -- Get this category
        local thisCategory = string.trimleadingtrailing (args["newCategory" .. i] or "")

        -- If this category isn't blank, then do something with it.
        if thisCategory ~= "" then
            -- Try to get some information for this category
            local categoryInfo = getCategoryByName (db, thisCategory)

            -- If we got nil, then someone is trying to add a new category, so do that. Otherwise,
            -- complain about a duplicate.
            if categoryInfo == nil then
                -- Add the new category to the system.
                selectOrCreateCategory (db, thisCategory)

                -- Generate a log to say what happened.
                local text = "Added category '" .. thisCategory .. "'"
                html:addText (successText (text) , "<BR>")

                -- Add this category to the list.
                table.insert (newCategories, thisCategory)
            else
                -- This category must have been a duplicate.
                html:addText (errorText (ERR_CAT_DUPCATEGORY) , "<BR>")
            end
        end
    end

    -- Seperate output from above from stuff that will follow.
    html:addText ("<P>")

    -- If any category were added, then add a log.
    if table.getn (newCategories) ~= 0 then
        addLog (db, "Added categories: " .. table.concat (newCategories, ", "))
    end
end

---------------------------------------------------------------------------

-- The commands for adding, removing and viewing properties are all mapped
-- to the same command. This is called to determine which form is being
-- submitted (if any), so that we can process the form appropriately 
-- before continuing on to display the list of categories.
local processForm = function (db)
    -- First, collect form arguments. What we really need is a command, 
    -- at the moment.
    local args = collectFormArguments ({"command"})

    -- If there are no arguments, then return. This can't really happen,
    -- since if there was no command, we wouldn't be here. But what the
    -- heck.
    if args == nil then return end

    -- If the command is just "Category", then the page is being viewed
    -- and nothing else is happening, so allow it. 
    if args.command == "Category" then
        return 
    end

    -- Based on the command, process the form, We know the command has to
    -- be one of the two commands we expect, since the edit command would
    -- not let control reach us if the command was something else.
    if args.command == "CategoryAdd" then 
        processAddForm (db)
    else
        processDeleteForm (db) 
    end
end


---------------------------------------------------------------------------

-- Take the database, the database row for the category to be modified,
-- and form args {"categoryID" , "submitted", and "newCategoryName"}, where
-- the last two fields are optional. 
--
-- If this is a valid submission, then update and alter categoryInfo on the
-- way out.
local handlePropertyFormSubmission = function (db , categoryInfo , args)
    -- If the form hasn't been submitted, then do nothing.
    if args.submitted == nil then return end

    -- The form has been submitted. If the new category name is empty or
    -- nil, then complain.
    local newName = string.trimleadingtrailing (args.newCategoryName or "")

    -- If the new name is empty, then that's bad.
    if newName == "" then
        html:addText (errorText (ERR_CAT_BADPROPNAME) , "<P>")
        return
    end

    -- if the new name matches a category that's already in the system, that
    -- also is bad.
    local otherCategory = getCategoryByName (db, newName)
    if otherCategory then
        -- There is a category by this name. If the category ID is the same as
        -- ours, then someone just submitted the form without changing anything.
        -- That's not an error.
        if tonumber (otherCategory.categoryID) == tonumber (categoryInfo.categoryID) then 
            html:addText (successText (SUCC_CAT_NOCHANGE), "<P>")
            return
        end

        -- However, if the ID's aren't the same, then this is an error.
        html:addText (errorText (ERR_CAT_DUPCATEGORY) , "<P>")
        return
    end

    -- Must be ok. Update the Database first.
    renameCategoryWithID (db, categoryInfo.categoryID, newName)

    -- Add to the system log and to the page
    addLog (db, "Renamed categiry '" .. categoryInfo.categoryName .. "' to '" .. newName .. "'")
    html:addText (successText ("Renamed category '" .. categoryInfo.categoryName .. "' to '" .. newName .. "'") , "<P>")

    -- Before returning, make sure the category name is updated, so that the generated page
    -- is correct.
    categoryInfo.categoryName = newName
end

---------------------------------------------------------------------------

-- Handle category property editing when someone clicks the properties link
-- on the main category page.
function commandCategoryProperties (db)
    -- Header for the page
    generatePageHeader (thisPageSection, "Edit Category Properties", "categoryProperties")

    -- Generate a button to take us back to the main category page
    html.form:singleButtonForm ("Return to Categories Page" , "Category" , env.scriptName , "POST" , nil , nil , "category")

    -- Before going any further, make sure that this user is allowed to edit
    -- properties. This stops people from hand making URL's to do things that
    -- the system doesn't provide an interface for.
    if userInfo.canEditCategory == false then
        addLog (db, "Attempted to edit category properties")
        html:addText (errorText (ERR_AUTH_NOEDITPRIV) , "<P>")
        return 
    end

    -- Get some form arguments. We definitly need a categoryID, so we know what
    -- category is being edited. The other fields are optional, because they're
    -- only available after the form has been submitted.
    --
    -- Leave now if we don't have at least the category we need.
    local args = collectFormArguments ({"categoryID"} , {"submitted" , "newCategoryName"})
    if args == nil then 
        html:addText (errorText (ERR_CAT_NOCATID) , "<P>")
        return
    end

    -- Try to get information on the category being edited. If this doesn't work,
    -- the fail out with an error. This is needed to generate the form.
    local categoryInfo = getCategoryByID (db, args.categoryID)
    if categoryInfo == nil then
        html:addText (errorText (ERR_CAT_NOCATID) , "<P>")
        return
    end

    -- Handle form submission, if needed.
    handlePropertyFormSubmission (db , categoryInfo , args)

    -- Now display the form
    html.form:open (env.scriptName , "POST");
    html.table:open (nil, nil, nil, nil, "5", "categoryList")
    html.table:caption ("Properties for: <B>" .. categoryInfo.categoryName .. "</B>", "CENTER")

    -- The only property currently supported for categories is the name, so generate a
    -- field to allow the editing of that. The default value will be the current name.
    html.table:heading (true , nil, nil, "RIGHT" , "CENTER")
    html:addText ("Name")
    html.table:cell (true, nil, nil, "LEFT", "CENTER")
    html.form:text ("newCategoryName" , categoryInfo.categoryName , "40" , "60")
    html.table:newRow ()

    -- Finish up the form with a submit button
    html.table:cell (true, nil, nil, "CENTER" , "TOP", nil, nil, nil, "2")
    html.form:submit ("Button" , "Update Properties" , "category")
 
    -- Table is complete now
    html.table:close ();

    -- Finish up the form now. Make sure we propogate the categoryID we have currently
    -- into the form submission.
    html.form:hidden ("submitted" , "yes")
    html.form:hidden ("command" , "CategoryProperties")
    html.form:hidden ("categoryID" , categoryInfo.categoryID)
    html.form:close ();
end

---------------------------------------------------------------------------

-- This is the main function to view and edit game categories
function commandCategory (db)
    -- Header for the page
    generatePageHeader (thisPageSection, nil, "categoryBase")

    -- Handle the form submission, if there has been one. 
    processForm (db)

    -- If this user is allowed to edit categories, then we are going to generate
    -- a special add form to allow for the addition. We want that to appear to 
    -- the right of the category list, so we will need a table to wrap everything.
    if userInfo.canEditCategory then
        html.table:open (nil, nil, nil, nil, "10")
        html.table:cell (false, "1%", nil, "LEFT", "TOP")
    end

    -- Generate the delete form, which is the form that lists all
    -- categories, with links to view games in that category and
    -- a properties link. This will also allow users with the correct
    -- permissions to delete categories as well.
    generateCategoryList (db)

    -- If this user is allowed to edit categories, then generate a
    -- form that will allow the addition of new categories to the
    -- system.
    if userInfo.canEditCategory then
        html.table:cell (true, "99%", nil, "LEFT", "TOP")
        generateAddForm (db)
        html.table:close ()
    end
end

---------------------------------------------------------------------------
--- MAIN CODE -------------------------------------------------------------
---------------------------------------------------------------------------

