---------------------------------------------------------------------------
--- LOCAL VARS ------------------------------------------------------------
---------------------------------------------------------------------------

-- What section to use in page headers in this file.
local thisPageSection = "CONFIGURATION"                         

-- Error messages
local ERR_CONF_NOPRIVEDIT = [[You are not allowed to alter the system configuration.]]
local ERR_CONF_BADTIMEOUT = [[The login timeout you have specified is not valid.]]
local ERR_CONF_BADDEFAULT = [[The default login page specified is not valid.]]
local SUCC_CONF_EDITOK = [[Your configuration changes have been saved.]]

---------------------------------------------------------------------------
--- FUNCTIONS -------------------------------------------------------------
---------------------------------------------------------------------------

-- Handles the form submission for the configuration settings change.
local processConfigChangeSubmission = function (db)
    -- Collect the form arguments. All must be specified.
    local args = collectFormArguments ({"newLoginTimeout", "newDefaultPage"});

    -- If we didn't get any arguments, then just do nothing at all. This is
    -- what happens when someone views the page naturally without submitting
    -- the form.
    if args == nil then return nil end

    -- Convert the new login timeout to be a number. If the argument is a non-number,
    -- tonumber will return nil. In that case, we pretend that the value returned
    -- is -1, so we can flag an error.
    local newValue = tonumber (args.newLoginTimeout) or -1

    -- If the value is <= 0, then complain. This will trigger both when a non-numeric
    -- argument was given, and also if the user attempts to set a negative value as the
    -- timeout.
    if newValue <= 0 then
        return errorText (ERR_CONF_BADTIMEOUT)
    end

    -- Validate the default page. It just has to be a non-empty string.
    local newPage = string.trimleadingtrailing (args.newDefaultPage)
    if newPage == "" then
        return errorText (ERR_CONF_BADDEFAULT)
    end

    -- Updage the config table.
    updateConfigInfo (db, "loginTimeout" , newValue)
    updateConfigInfo (db, "defaultPage" , newPage)

    -- Make sure our global configuration information is modified as well. This
    -- will be used to generate the form, so we want it to have the same values as
    -- the database now has.
    configInfo.loginTimeout = tostring (newValue)
    configInfo.defaultPage = newPage

    -- All done with a successful edit.
    return successText (SUCC_CONF_EDITOK)
end

---------------------------------------------------------------------------

-- Generates the dropdown items for the default page selection. itemlist is
-- a table with keys "name" and "value". Userinfo is the the value of the
-- current configuration setting, which will be selected by default.
local generateDefaultLoginPageDropdown = function (itemList, userInfo)
    -- Loop over all items
    for index,info in itemList do
        -- Is this item selected? It is if the value of this item is
        -- the same as the userInfo we were given.
        local isSelected = false
        if info.value == userInfo then isSelected = true end

        -- Generate the select option now.
        html.form:selectOption (info.name, info.value, isSelected)
    end
end

---------------------------------------------------------------------------

-- This is the main function to view and edit system condifuration
function commandConfiguration (db)
    -- Header for the page
    generatePageHeader (thisPageSection, nil, "configHelp")

    -- Before proceeding, ensure that this user actually has the
    -- permissions required to perform a configuration change or
    -- view.
    if userInfo.canConfigSystem == false then
        -- Say what they tried to do
        addLog (db , "Attempted to view/alter system configuration")

        -- Add some error text, and then return.
        html:addText (errorText (ERR_CONF_NOPRIVEDIT) , "<P>")
        return
    end

    -- Handle the form submission. If this returns something, then display it before
    -- returning. It could be success or error text. if it returns nil, this means
    -- that the form wasn't actually submitted, and so nothing was done.
    local result = processConfigChangeSubmission (db)
    if result ~= nil then html:addText (result , "<P>") end
          
    -- Open the form that will allow for the editing of configuration options.
    html.form:open (env.scriptName , "POST")

    -- Open a table for the values
    html.table:open (nil, nil, nil, nil, "5", "configList")

    -- Display a heading for all of the rows. There is a column for the setting
    -- name, a description of what it does, and it's value.
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Setting")
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Description")
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Value")
    html.table:newRow ();

    -- The name of the setting
    html.table:cell (true , nil , nil , "LEFT" , "TOP")
    html:addText ("Login Timeout")

    -- Description of the setting
    html.table:cell (false, nil , nil, "LEFT", "TOP")
    html:addText ("How long (in minutes) that your login session can be idle<BR>",
                  "before you are automatically logged out of the system.");

    -- Form element for the setting
    html.table:cell (true , nil , nil , "LEFT" , "TOP")
    html.form:text ("newLoginTimeout", configInfo.loginTimeout, "10" , "5", "configText")

    -- New row
    html.table:newRow ();

    -- The name of the setting
    html.table:cell (true , nil , nil , "LEFT" , "TOP")
    html:addText ("Login Page")

    -- Description of the setting
    html.table:cell (false, nil , nil, "LEFT", "TOP")
    html:addText ("What page should be displayed at login time.");

    -- Form element for the setting
    local elements = {{["name"] = "Player List",         ["value"] = "Player"},
                      {["name"] = "Designer List",       ["value"] = "Designer"},
                      {["name"] = "Publisher List",      ["value"] = "Publisher"},
                      {["name"] = "Category List",       ["value"] = "Category"},
                      {["name"] = "Game List",           ["value"] = "Game"},
                      {["name"] = "Session Report List", ["value"] = "Session"},
                      {["name"] = "Tools",               ["value"] = "Tool"}}
    html.table:cell (true , nil , nil , "LEFT" , "TOP")
    html.form:select ("newDefaultPage" , "1", nil, "configDrop", nil, generateDefaultLoginPageDropdown, elements, configInfo.defaultPage)

    -- New row
    html.table:newRow ();

    -- Make a submit button to finish up the form.
    html.table:cell (true, nil, nil, "CENTER", "TOP", nil, nil, nil, "3")
    html.form:submit ("Button", "Alter System Configuration", "config")
    html.form:hidden ("command", "ConfigureChange")

    -- Close the table and form now
    html.table:close ()
    html.form:close ()

    -- Based on the value of result above, we can tell what happened. If it's
    -- nil, then no form was submitted and someone just viewed the configuration.
    -- If it's non-nil, then the configuration form was submitted, so things were
    -- altered.
    local msg = "Viewed"
    if result then msg = "Altered" end

    -- Display a log
    addLog (db , msg .. " system configuration")
end


---------------------------------------------------------------------------
--- MAIN CODE -------------------------------------------------------------
---------------------------------------------------------------------------

