---------------------------------------------------------------------------
--- LOCAL VARS ------------------------------------------------------------
---------------------------------------------------------------------------

-- What section to use in page headers in this file.
local thisPageSection = "DESIGNERS"                         

-- Error Messags
local ERR_AUTH_NOADDPRIV = [[You are not allowed to add new designers to the system. Please check with your administrator.]]
local ERR_AUTH_NODELPRIV = [[You are not allowed to delete designers from the system. Please check with your administrator.]]
local ERR_AUTH_NOEDITPRIV = [[You are not allowed to edit designers in the system. Please check with your administrator.]]
local ERR_DES_DUPDESIGN = [[A designer by this name already exists.]]
local ERR_DEL_NODESIGNERS = [[You did not select any designers to delete.]]
local ERR_DEL_NOFINDDESIGN = [[Unable to find the designer to delete. It may have been removed in another session.]]
local ERR_DES_NODESID = [[An error has occured. I don't know what designer you're trying to modify. Please try again.]]
local ERR_DES_BADPROPNAME = [[The designer name you have specified is invalid. Please try again.]]
local SUCC_DES_NOCHANGE = [[No changes have been made.]]

-- The number of designer fields to include on the add new designer page.
local ADD_DESIGNER_FIELD_COUNT = 5

---------------------------------------------------------------------------
--- FUNCTIONS -------------------------------------------------------------
---------------------------------------------------------------------------

-- Count how many designers are registered with the system, and then display
-- that value. Do nothing if the information can't be obtained.
local displayDesignerCount = function (db)
    -- Count the number of rows in the designer list. If this doesn't work,
    -- then just return without doing anything.
    local totalDesigners = rowsInTable (db , "DesignerList")
    if totalDesigners == nil then return end

    -- Decrement the count by 1, since the list includes the entry for None
    -- that doesn't count.
    totalDesigners = totalDesigners - 1

    -- To be anal (and grammatically correct), figure out what to display for
    -- a singular entry or for 0 or many.
    local label1 = "are "
    local label2 = " designers"
    if totalDesigners == 1 then label1 = "is " label2 = " designer" end

    -- Output some stuff now.
    html:addText ('<FONT SIZE="-1">')
    html:addText ("There ", label1, totalDesigners , label2, " registered with the system.")
    html:addText ("</FONT><P>")
end

---------------------------------------------------------------------------

-- This will generate the list of designers currently in the system, along with
-- the associated form elements needed to delete designers from the system.
local generateDesignerList = function (db)
    -- Say how many designers there are.
    displayDesignerCount (db)

    -- If the user has javascript enabled, this confirmation will be used to get them to
    -- confirm what they are about to do.
    local deleteConfirm = "return confirm ('Any selected designers will be completely removed from the system. Are you sure this is what you want to do?');"

    -- Open the form that will allow for the deletion of designers
    html.form:open (env.scriptName , "POST", nil, nil, deleteConfirm)

    -- Open a table for the values
    html.table:open (nil, nil, nil, nil, "5", "designerList")

    -- Short-name the flag that says if the user is allowed to edit and
    -- delete designers or not.
    local canEdit = userInfo.canEditDesigner 

    -- Display a heading for all of the rows
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Designer")

    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Games")

    -- There are extra columns if the user has editing rights.
    -- One for properties, and one for the delete checkbox.
    if canEdit then
        html.table:heading (true, nil, nil, "LEFT", "TOP")
        html:addText ("&nbsp;")

        html.table:heading (true, nil, nil, "LEFT", "TOP")
        html:addText ("Delete")
    end

    -- Get ready for the next row
    html.table:newRow ()

    -- Obtain a list of all designers, sorted according to name, and
    -- excluding the default "None" which has an ID of 0.
    --
    -- Once we have the list, loop over it outputting table rows.
    local designList = getDesignerList (db, true, true)
    for index,row in designList do
        -- Include the name of this designer
        html.table:cell (true, nil, nil, "LEFT" , "TOP")
        html:addText (row.designerName)

        -- Include how many games there are by this designer.
        html.table:cell (true, nil, nil, "LEFT" , "TOP")

        -- Determine how many games there are by this designer. If 
        -- we can, then output a link that allows people to view all
        -- games by this designer.
        local count = countGamesByDesigner (db, row.designerID)
        if count ~= nil then
            html:link (urls.gamesByDesigner .. row.designerID , count)
        else
            html:addText ("N/A")
        end

        -- Add in extra columns for the properties link and the
        -- checkbox for the delete checkbox, if allowed.
        if canEdit then
            -- Properties link.
            html.table:cell (true, nil, nil, "LEFT" , "TOP")
            html:link (urls.designerProperties .. row.designerID , "Properties")

            -- Delete this checkbox.
            html.table:cell (true, nil, nil, "LEFT" , "TOP")
            html.form:checkbox ("delDesignerID" , row.designerID)
        end

        -- Get ready for the next row
        html.table:newRow ()
    end

    -- If there are 1 or more designers in the list that we just wrote out,
    -- include a submit button for deleting. We don't need to do that if
    -- there were no designers, since there's nothing to delete in that case.
    if table.getn (designList) > 0 and canEdit then
        -- Create a cell that spans all columns, and put the submit button
        html.table:cell (true, nil, nil, "CENTER", "TOP", nil, nil, nil, "4")
        html.form:submit ("Button", "Delete selected Designers", "designer")

        -- Flag the command.
        html.form:hidden ("command", "DesignerDelete")
    end

    -- Close the table and form now
    html.table:close ()
    html.form:close ()

end

---------------------------------------------------------------------------

-- Generate the form that allows you to add new designers to the system.
local generateAddForm = function (db)
    -- Output a blank line of the same size as the one that the designer
    -- list uses to show how many designers that there are. 
    html:addText ('<FONT SIZE="-1">&nbsp;</FONT><P>')

    -- Open the form that will allow for the addition of designers. 
    html.form:open (env.scriptName , "POST")

    -- Open a table for the values
    html.table:open (nil, nil, nil, nil, "5", "designerList")

    -- Display a heading for all of the rows
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Designer Name")

    -- Get ready for the next row
    html.table:newRow ()

    -- Ourput a row for each designer that we want to add.
    for i = 1,ADD_DESIGNER_FIELD_COUNT do
        html.table:cell (true, nil, nil, "LEFT" , "TOP")
        html.form:text ("newDesigner" .. i , "" , "40" , "60")
        html.table:newRow ()
    end

    -- Finish up the table with a submit button, and make sure that
    -- the command gos to the right place.
    html.table:cell (true, nil, nil, "CENTER", "TOP")
    html.form:submit ("Button", "Add Designers", "designer")
    html.form:hidden ("command", "DesignerAdd")

    -- Close the table and form now
    html.table:close ()
    html.form:close ()
end

---------------------------------------------------------------------------

-- Takes some designerInfo (ID and name, as in the DesignerList table), and
-- then finds all games that have that designer as the primary designer. For
-- any that do, the primary designer is replaced with the first match in
-- the DesignerPlacement table. If the game has only one designer, then
-- the designer is set to None.
--
-- This is used when a designer is deleted, to fix up any games that listed
-- that designer as primary.
local fixDesignedGames = function (db , designInfo)
    -- We are removing the designer in designInfo. So, find all games for which
    -- that designer is the primary designer. 
    local foundGames = getGamesWithPrimaryDesigner (db, designInfo.designerID)

    -- Now loop over the results that we saved.
    for index,row in foundGames do
        -- For this game, find the first row in DesignerPlacement for this game.
        local secondaryDesigner = getFirstSecondaryDesignerForGame (db, row.gameID)

        -- Now, if secondaryDesigner is nil, then this game had only a single designer. In that
        -- case, set the designer in this game to be 0 (None).
        if secondaryDesigner == nil then
            -- No secondary designer, so just update this game to have a designer that is empty.
            alterGamePrimaryDesigner (db, row.gameID, 0)
            html:addText ("Primary designer of " , row.name , " has been set to 'None'<BR>")
        else
            -- Here ther WAS a secondary designer. First of all, delete this row from the
            -- DesignerPlacement table. Then, update the GameList entry for this game to have
            -- the secondary designer as primary.
            deleteDesignerPlacementByPlacement (db, secondaryDesigner.placementID)
            alterGamePrimaryDesigner (db, row.gameID, secondaryDesigner.designerID)

            -- Figure out the new primary and say what happend
            local newPrimary = getDesignerByID (db, secondaryDesigner.designerID)
            if newPrimary then
                html:addText ("Primary designer of " , row.name , " has been set to '" , newPrimary.designerName , "'<BR>")
            end
        end
    end
end

---------------------------------------------------------------------------

-- Called to process the possible deletion of designers from the system.
local processDeleteForm = function (db)
    -- Make sure that this user is allowed to delete designers before we go
    -- any further, in case someone hand crafted a URL.
    if userInfo.canEditDesigner == false then
        addLog (db, "Attempted to delete designers")
        html:addText (errorText (ERR_AUTH_NODELPRIV) , "<P>")
        return 
    end

    -- To handle a delete form, we need the list of designer ID's to
    -- delete. This could be nil, a string containing a single
    -- designerID, or an array of designerID's, depending on how
    -- many designers were checked when the form was submitted.
    local args = collectFormArguments ({"delDesignerID"})

    -- We want at least a single designer ID to delete.
    if args == nil then
        addLog (db, "Attempted delete with no designers selected")
        html:addText (errorText (ERR_DEL_NODESIGNERS) , "<P>")
        return
    end

    -- Make sure that the designer ID is a table. it might not be, if only
    -- a single designer was selected for deletion.
    if type (args.delDesignerID) ~= "table" then
        args.delDesignerID = {args.delDesignerID}
    end

    -- Loop over and remove all designers given. Confirmation was received
    -- before the form submitted.
    for index,auth in args.delDesignerID do
        -- If the provided auth (as a number) is 0, then don't do anything,
        -- since that is the special ID for "No Designer". This should never
        -- happen, but it might if someone hand-crafts a form or something.
        if tonumber (auth) ~= 0 then
            -- Try to get information on the designer that is being removed.
            local delDesignInfo = getDesignerByID (db, auth)

            -- If we couldn't find information on this designer, then skip. Otherwise, do stuff.
            if delDesignInfo == nil then
                -- No such designer?
                html:addText (errorText (ERR_DEL_NOFINDDESIGN) , "<BR>")
            else
                -- We're deleting this designer. Firstly, remove all mention of this designer
                -- from the DesignerList and the DesignerPlacement tables.
                deleteDesignerByID (db, auth)
                deleteDesignerPlacementsByDesigner (db, auth)

                -- Say what designer was deleted
                html:addText (successText ("Deleted designer '" .. delDesignInfo.designerName .. "'") , "<BR>")
                addLog (db , "Deleted designer '" .. delDesignInfo.designerName .. "'")

                -- Now fix all games which have this designer as the primary designer. We do this AFTER
                -- the above so that we can output status text saying what designer becomes what, and such.
                -- After that, output a paragraph seperator to seperate this designer from the info
                -- that may come after it.
                fixDesignedGames (db , delDesignInfo)
                html:addText ("<P>")
            end
        end
    end
end

---------------------------------------------------------------------------

-- Called to process the possible submission of an "Add Designer" form.
local processAddForm = function (db)
    -- If this user isn't allowed to edit designers, he's not allowed to
    -- add a designer either. In this case we can only get here if someone
    -- manually enters the URL, since the system won't provide the add form
    -- on it's on.
    if userInfo.canEditDesigner == false then
        addLog (db, "Attempted to add designers")
        html:addText (errorText (ERR_AUTH_NOADDPRIV) , "<P>")
        return 
    end

    -- Figure out how many arguments are needed from the form. 
    -- This number is configurable at the top of the file by
    -- modifying the value of ADD_DESIGNER_FIELD_COUNT.
    local argsNeeded = {}
    for i = 1,ADD_DESIGNER_FIELD_COUNT do
        table.insert (argsNeeded , "newDesigner" .. i)
    end

    -- Get the form arguments. If we can't get them, then we can just
    -- leave, since this means the form wasn't submitted at all.
    local args = collectFormArguments (argsNeeded)
    if args == nil then return end

    -- Check over all of the designers and see if we should add them.
    -- We will keep track of how many designers were actually added, as
    -- well as the names of them. This isn't straight forward, because some
    -- designers could be blank, and thus ignored, or they might be a duplicate.
    local newDesigns = {}
    for i = 1,ADD_DESIGNER_FIELD_COUNT do
        -- Get the name of this new designer.
        local thisDesigner = string.trimleadingtrailing (args["newDesigner" .. i] or "")

        -- If this designer isn't blank, then do something with it.
        if thisDesigner ~= "" then
            -- Try to get some information for this designer, according to name.
            local designerInfo = getDesignerByName (db, thisDesigner)

            -- If we got nil, then someone is trying to add a new designer, so do that. Otherwise,
            -- complain about a duplicate.
            if designerInfo == nil then
                -- Add in the new designer.
                selectOrCreateDesigner (db, thisDesigner)

                -- Add some text to the page saying that this designer was added.
                local text = "Added designer '" .. thisDesigner .. "'"
                html:addText (successText (text) , "<BR>")

                -- Add this to the list.
                table.insert (newDesigns, thisDesigner)
            else
                -- This designer was a duplicate. That's bad.
                html:addText (errorText (ERR_DES_DUPDESIGN) , "<BR>")
            end
        end
    end

    -- Seperate output from above from stuff that will follow.
    html:addText ("<P>")

    -- If any designers were added, then add a log.
    if table.getn (newDesigns) ~= 0 then
        addLog (db, "Added designers: " .. table.concat (newDesigns, ", "))
    end
end

---------------------------------------------------------------------------

-- For designers, the commands Designer, DesignerAdd and DesignerDelete 
-- are mapped in the command map (in cmdmap.lua) to the same command. The
-- first thing that command does is call us, to determine if an add or delete
-- form has been submitted, and if so, to actually handle the submission. After
-- that, the regular processing happens, re-generating both forms.
local processForm = function (db)
    -- First, collect form arguments. What we really need is a command, 
    -- at the moment.
    local args = collectFormArguments ({"command"})

    -- If there are no arguments, then return. This can't really happen,
    -- since if there was no command, we wouldn't be here. But what the
    -- heck.
    if args == nil then return end

    -- If the command is just "Designer", then the page is being viewed
    -- and nothing else is happening, so allow it.
    if args.command == "Designer" then
        return 
    end

    -- Based on the command, process the form. We know the command has to 
    -- be one of the two commands we expect, since the properties command
    -- would not let control reach us if the command was unknown.
    if args.command == "DesignerAdd" then 
        processAddForm (db)
    else
        processDeleteForm (db) 
    end
end


---------------------------------------------------------------------------

-- Take the database, the database row for the designer to be modified,
-- and form args {"designerID" , "submitted", and "newDesignerName"}, where
-- the last two fields are optional. 
--
-- If this is a valid submission, then update and alter designerInfo on the
-- way out.
local handlePropertyFormSubmission = function (db , designerInfo , args)
    -- If the form hasn't been submitted, then do nothing.
    if args.submitted == nil then return end

    -- The form has been submitted. If the new designer name is empty or
    -- nil, then complain.
    local newName = string.trimleadingtrailing (args.newDesignerName or "")

    -- If the new name is empty, then that's bad.
    if newName == "" then
        html:addText (errorText (ERR_DES_BADPROPNAME) , "<P>")
        return
    end

    -- If the new name matches a designer that's already in the system, that
    -- also is bad.
    local otherDesigner = getDesignerByName (db, newName)
    if otherDesigner then
        -- If this is us, then the form was submitted with no changes made.
        if tonumber (otherDesigner.designerID) == tonumber (designerInfo.designerID) then 
            html:addText (successText (SUCC_DES_NOCHANGE), "<P>")
            return
        end

        -- That's not the case, so the name is a duplicate of some other designer.
        html:addText (errorText (ERR_DES_DUPDESIGN) , "<P>")
        return
    end

    -- Must be ok. Update the Database first.
    renameDesignerWithID (db, designerInfo.designerID, newName)

    -- Add to the system log and to the page
    addLog (db, "Renamed designer '" .. designerInfo.designerName .. "' to '" .. newName .. "'")
    html:addText (successText ("Renamed designer '" .. designerInfo.designerName .. "' to '" .. newName .. "'") , "<P>")

    -- Before returning, make sure the designer name is updated, so that the generated page
    -- is correct. 
    designerInfo.designerName = newName
end

---------------------------------------------------------------------------

-- Handle designer property editing. This gets invoked when someone clicks the
-- properties link next to a designer.
function commandDesignerProperties (db)
    -- Header for the page
    generatePageHeader (thisPageSection, "Edit Designer Properties", "designerProperties")

    -- Generate a button to take us back to the main designer page
    html.form:singleButtonForm ("Return to Designers Page" , "Designer" , env.scriptName , "POST" , nil , nil , "designer")

    -- Only continue if this user is allowed to edit designers.
    if userInfo.canEditDesigner == false then
        addLog (db, "Attempted to edit designers properties")
        html:addText (errorText (ERR_AUTH_NOEDITPRIV) , "<P>")
        return 
    end

    -- Get some form arguments. At the very least, we need to have the designerID
    -- of the designer that we're editing. Optionally, we care about the submitted
    -- field and the new designer ID field.
    -- 
    -- We stop processing of this command if no designer ID was given.
    local args = collectFormArguments ({"designerID"} , {"submitted" , "newDesignerName"})
    if args == nil then 
        html:addText (errorText (ERR_DES_NODESID) , "<P>")
        return
    end

    -- Get information on the designer to be edited, and leave if we can't
    -- find it.
    local designerInfo = getDesignerByID (db, args.designerID)
    if designerInfo == nil then
        html:addText (errorText (ERR_DES_NODESID) , "<P>")
        return
    end

    -- Handle form submission, if needed.
    handlePropertyFormSubmission (db , designerInfo , args)

    -- Now display the form
    html.form:open (env.scriptName , "POST");
    html.table:open (nil, nil, nil, nil, "5", "designerList")
    html.table:caption ("Properties for: <B>" .. designerInfo.designerName .. "</B>", "CENTER")

    -- The only property is the name of the designer attached to this
    -- designer ID. This allows us to rename things without screwing with
    -- any other data.
    html.table:heading (true , nil, nil, "RIGHT" , "CENTER")
    html:addText ("Name")
    html.table:cell (true, nil, nil, "LEFT", "CENTER")
    html.form:text ("newDesignerName" , designerInfo.designerName , "40" , "60")
    html.table:newRow ()

    -- Generate the cell with the submit button.
    html.table:cell (true, nil, nil, "CENTER" , "TOP", nil, nil, nil, "2")
    html.form:submit ("Button" , "Update Properties" , "designer")
 
    -- Table is complete now
    html.table:close ();

    -- Finish up the form now. We need some hidden fields that will tell us
    -- the designerID that we already know, and that the form was submitted,
    -- so that when we re-generate the page we'll know what happened.
    html.form:hidden ("submitted" , "yes")
    html.form:hidden ("command" , "DesignerProperties")
    html.form:hidden ("designerID" , designerInfo.designerID)
    html.form:close ();
end

---------------------------------------------------------------------------

-- This is the main function to view and edit game designers
function commandDesigner (db)
    -- Header for the page
    generatePageHeader (thisPageSection, nil, "designerBase")

    -- Handle the form submission, if there has been one. 
    processForm (db)

    -- If this user is allowed to edit designers, then we are going to generate
    -- a special add form to allow for the addition. We want that to appear to 
    -- the right of the designer list, so we will need a table to wrap everything.
    if userInfo.canEditDesigner then
        html.table:open (nil, nil, nil, nil, "10")
        html.table:cell (false, "1%", nil, "LEFT", "TOP")
    end

    -- Now, generate the delete form. This is the form that lists all of
    -- the designers in the system, along with a delete checkbox next to
    -- them. If the current user can't delete designers, then the delete
    -- checkboxes aren't listed. 
    generateDesignerList (db)

    -- If this user is allowed to edit designers, then generate a form that
    -- allows for the addition of new designers.
    if userInfo.canEditDesigner then
        html.table:cell (true, "99%", nil, "LEFT", "TOP")
        generateAddForm (db)
        html.table:close ()
    end
end

---------------------------------------------------------------------------
--- MAIN CODE -------------------------------------------------------------
---------------------------------------------------------------------------

