---------------------------------------------------------------------------
--- LOCAL VARS ------------------------------------------------------------
---------------------------------------------------------------------------

-- What section to use in page headers in this file.
local thisPageSection = "GAMES"                         

-- Error messges
local ERR_GAME_NOGAMEID = [[An error has occured. I don't know what game you're trying to view. Please try again.]]
local ERR_GAME_NODESIGNERID = [[An error has occured. I don't know what designer's games you're trying to view. Please try again.]]
local ERR_GAME_NOPUBLISHERID = [[An error has occured. I don't know what publisher's games you're trying to view. Please try again.]]
local ERR_GAME_NOCATEGORYID = [[An error has occured. I don't know what category of games you're trying to view. Please try again.]]
local ERR_GAME_NOADDPRIV = [[You are not allowed to add new games to the system. Please check with your administrator.]]
local ERR_GAME_NOEDITPRIV = [[You are not allowed to edit games in the system. Please check with your administrator.]]
local ERR_GAME_NODELPRIV = [[You are not allowed to delete games from the system. Please check with your administrator.]]
local ERR_GAME_NOGAMENAME = [[You need to specify a name for this game.]]
local ERR_GAME_BADEXPANSION = [[The expansion game selected is not valid.]]
local ERR_GAME_BADBGGID = [[The boardgamegeek value entered is not valid.]]
local ERR_GAME_BADMINPLAY = [[The minimum number of players entered is not valid.]]
local ERR_GAME_BADMAXPLAY = [[The maximum number of players entered is not valid.]]
local ERR_GAME_BADPLAYTIME = [[The play time of the game entered is not valid.]]
local ERR_GAME_BADDESIGNER = [[A designer you have specified is invalid.]]
local ERR_GAME_BADPUBLISHER = [[A publisher you have specified is invalid.]]
local ERR_GAME_BADCATEGORY = [[A category you have specified is invalid.]]
local ERR_DEL_NOGAMES = [[You did not select any games to delete.]]
local ERR_DEL_NOFINDGAME = [[Unable to find the game to delete. Possibly deleted from another session.]]

---------------------------------------------------------------------------
--- FUNCTIONS -------------------------------------------------------------
---------------------------------------------------------------------------

-- Given an array of game information and a gameID, return back the index
-- of the game given, or nil on error.
local indexOfGame = function (gameList, gameID)
    -- Convert the ID we want to a number
    gameID = tonumber (gameID)

    -- Scan the list. If we find the game with the ID given, then
    -- return it's index back.
    for index,game in gameList do
        if tonumber (game.gameID) == gameID then return index end
    end

    -- Failure
    return nil
end

---------------------------------------------------------------------------

-- This will generate a control bar that allows for navigation between
-- games in the system, plus other links for quickly viewing or adding
-- session reports for this game.
local generateGameControlBar = function (db, gameID, isEditPage)
    -- Get the complete game list. If it is empty, or only a single
    -- game, then pretend that it's nil
    local gameList = getGameList (db, true) or {}
    if table.getn (gameList) < 2 then gameList = nil end

    -- Open the bar
    html:addText ('<FONT SIZE="-2">')

    -- First thing is either edit/view, depending on if isEditPage is true or not.
    if isEditPage then
        wrappedLink (urls.gameDetails .. gameID , "View")
    else
        -- If this user can edit games, then add an edit link.
        if userInfo.canEditGame then
            wrappedLink (urls.gameProperties .. gameID , "Edit")
        end
    end

    -- Figure out the base URL for all of the other items. This makes sure that
    -- links in the property control bar go to the property page, and links to
    -- the viewer control bar go to the viewer.
    local baseURL = urls.gameDetails
    if isEditPage then baseURL = urls.gameProperties end

    -- Get information about the game that we're generating the bar for.
    local gameInfo = getGameByID (db, gameID)

    -- If we got a list of games, then we can generate navigation information.
    if gameList then
        -- Determine what index we are in the list of games.
        local myIndex = indexOfGame (gameList , gameID)
        if myIndex then
            -- If we're not the first game in the system, then generate a link to the
            -- previous game.
            if myIndex > 1 then
                wrappedLink (baseURL .. gameList[myIndex - 1].gameID, "Previous Game")
            else
                wrappedLink (nil, "No Previous")
            end

            -- If we're not the last game in the system, then generate a link to the
            -- next game.
            if myIndex + 1 <= table.getn (gameList) then
                wrappedLink (baseURL .. gameList[myIndex + 1].gameID, "Next Game")
            else
                wrappedLink (nil, "No Next")
            end
        end
    end

    -- We can insert some other helpful links if we know what game we are.
    if gameInfo then
        -- If this game has a non-zero bgg id, thn we generate a link to view
        -- this game on boardgamegeek.
        local geekID = tonumber (gameInfo.gamegeekID) or 0
        if geekID ~= 0 then
            wrappedLink (urls.bgg .. geekID, "View on BGG", "_new")
        end

        -- If this user can add a session report, generate a link to add a 
        -- session report for this game.
        if userInfo.canAddSession then
            wrappedLink (urls.sessionAdd .. gameInfo.gameID, "Add New Session")
        end

        -- Generate a link to view session reports for this game, if there are
        -- any reports filed.
        local sessionCount = countSessionReportsForGameID (db, gameInfo.gameID)
        if sessionCount ~= nil and sessionCount > 0 then
            wrappedLink (urls.gamesBySession .. gameInfo.gameID, "View Existing Sessions")
        else
            wrappedLink (nil, "No Sessions Reported")
        end
    end

    -- Close the bar
    html:addText ("</FONT>")
end

---------------------------------------------------------------------------

-- Takes a list of session reports and figures out the average playtime.
--
-- This will only take sessions with a recorded duration into account. 
--
-- The return value is the average, in minutes, or 0 if there was not enough
-- information to calculate an average.
local calculateAveragePlayTime = function (sessionList)
    -- Leave if there are no sessions.
    if table.getn (sessionList) == 0 then return 0 end

    -- Scan across all entries, and for each one, if there is a duration,
    -- add it to the total and count the entry.
    local totalDuration = 0
    local entriesSeen = 0

    for index,info in sessionList do
        if info.sessionDuration ~= 0 then
            totalDuration = totalDuration + info.sessionDuration
            entriesSeen = entriesSeen + 1
        end
    end

    -- If the total duration is 0, then all session reports have no
    -- duration, so return back 0.
    if totalDuration == 0 then return 0 end

    -- Return the average
    return math.floor (totalDuration / entriesSeen)
end

---------------------------------------------------------------------------

-- This outputs the play information for the detailed game viewer.
--
-- The play information is the name of the game, the list of designers
-- and list of publishers, along with the number of players supported
-- and the playtime for the game.
local genPlayInfo = function (db, gameInfo, designerList, publisherList)
    -- For game play information, we need to output the designers, publishers,
    -- number of players and the play time required. We're going to use a subtable
    -- for this.
    html.table:open (nil, nil, nil, nil, "5")

    -- First thing displayed is the list of designers.
    html.table:heading (true , nil, nil, "LEFT" , "TOP")
    html:addText ("Designer:")
    html.table:cell (true, nil, nil, "LEFT" , "TOP")

    -- The designer list
    for index,designer in designerList do
        html:link (urls.gamesByDesigner .. designer.id, designer.name)
        html:addText ("<BR>")
    end

    -- Start a new row. The next thing to display is the list of publishers.
    html.table:newRow ()
    html.table:heading (true , nil, nil, "LEFT" , "TOP")
    html:addText ("Publisher:")
    html.table:cell (true, nil, nil, "LEFT" , "TOP")

    -- The publisher list
    for index,publisher in publisherList do
        html:link (urls.gamesByPublisher .. publisher.id, publisher.name)
        html:addText ("<BR>")
    end

    -- Start a new row. Now we need to output the number of players, which is a lot
    -- easier. No subtable here.
    html.table:newRow ()
    html.table:heading (true , nil, nil, "LEFT" , "TOP")
    html:addText ("Players:")
    html.table:cell (nil, nil, nil, "LEFT" , "TOP")
    if gameInfo.minPlayers == gameInfo.maxPlayers then
        html:addText (gameInfo.minPlayers)
    else
        html:addText (gameInfo.minPlayers , " - " , gameInfo.maxPlayers)
    end

    -- Lastly, display the play time in a row of it's own
    html.table:newRow ()
    html.table:heading (true , nil, nil, "LEFT" , "TOP")
    html:addText ("Play Time:")
    html.table:cell (true, nil, nil, "LEFT" , "TOP")
    html:addText (gameInfo.playTimeMinutes , " minutes")

    -- Now, if there is any information in the database regarding the average playtime
    -- of the game, display that. This is taken by counting all session reports for this
    -- game that provide a duration, and determining the average duration from that
    local sessionReports = getSessionReportsForGameID (db, gameInfo.gameID)
    local sessionCount = table.getn (sessionReports)
    local averagePlaytime = calculateAveragePlayTime (sessionReports)

    -- If we calculated an average play time, then go into here. We don't do this if
    -- we couldn't get a play time, since this means that there are either no recorded
    -- plays, or there are plays, but all of them have no recorded duration.
    if averagePlaytime > 0 then
        html.table:newRow ()
        html.table:heading (true , nil, nil, "LEFT" , "TOP")
        html:addText ("Average Play Time:")
        html.table:cell (true, nil, nil, "LEFT" , "TOP")
        html:addText (averagePlaytime, " minutes<br>")
        html:addText ("(")
        html:link (urls.gamesBySession .. gameInfo.gameID, sessionCount)
        if sessionCount == 1 then html:addText (" game recorded)") else html:addText (" games recorded)") end
    end

    -- Close the main table now
    html.table:close ()
end

---------------------------------------------------------------------------

-- This takes the list of categories for the given game and outputs them.
-- This is used by the detailed game viewer.
local genCategoryInfo = function  (gameInfo , categoryList)
    -- For category information, we're going to generate a simple 2 column
    -- sub-table. The first column will have the text "Category:", and the
    -- second column will be a list of all categories, with the first one
    -- listed being the primary.
    html.table:open (nil , nil, nil, nil, "5")
    html.table:heading (true , nil, nil, "LEFT" , "TOP")
    html:addText ("Category:")

    html.table:cell (true, nil, nil, "LEFT" , "TOP")
    for index,category in categoryList do
        html:link (urls.gamesByCategory .. category.id, category.name)
        html:addText ("<BR>")
    end

    -- Close the table now
    html.table:close ()
end

---------------------------------------------------------------------------

-- Generate some navigation information for the provided game list. We generate
-- a row that has a bunch of links to skip to the part of the page where the games
-- with that character start. This will only generate links for letters that exist,
-- although all letters are listed.
--
-- The gameList will be modified such that each entry stored in it may have a key
-- added to indicate that this entry should have a anchor point.
--
-- The key added is "namedPointName"
local generateNavInformation = function (db, gameList, span)
    -- To start out, scan over every game in the game list. If we find
    -- an item that starts with a letter or number not already seen, then
    -- add it to the table so that we know that letter exists, and then edit
    -- this item so that it knows it's the item that should start out items
    -- with that character.
    local foundChars = {}
    for index,gameInfo in gameList do
        -- Get the first byte off of the beginning of the string. It will come back
        -- as it's ascii value.
        local firstChar = string.byte (string.trimleading (gameInfo.name))
        local key = nil

        -- Now, Convert firstChar a bit. First of all, if it's lowercase, then
        -- make it uppercase.
        if firstChar >= 97 and firstChar <= 122 then
            firstChar = firstChar - 32
        end

        -- If the character is uppercase, then it's the key. Otherwise, if it's a
        -- number, then the key will be "num"
        if firstChar >= 65 and firstChar <= 90 then
            key = string.char (firstChar)
        elseif firstChar >= 48 and firstChar <= 57 then
            key = "num"
        end

        -- Now, key is either a character, the word "num", or nil. If it's not nil,
        -- then continue.
        if key then
            -- If the foundChars doesn't have a key for this key already, then we
            -- have work to do.
            if foundChars[key] == nil then
                -- Tell foundChars that we have used this key already, so we don't
                -- need to use it any longer.
                foundChars[key] = true

                -- Now tell this gameInfo that this should have a tag.
                gameInfo.namedPointName = key
            end
        end
    end

    -- Start out the row that will contain our nav information.
    html.table:heading (true, nil, nil, "LEFT", "TOP", nil, span)

    -- Start the text of the bar.
    html:addText ("[ ")

    -- If the key has a number, then generate a link, otherwise just output text.
    if foundChars.num then
        html:link ("#num", "#")
        html:addText (" ")
    else
        html:addText ("# ")
    end

    -- Now loop over all letters of the alphabet.
    for val = 65,90 do
        -- Convert the ascii code to a character.
        local char = string.char (val)

        -- If there is a key for this item, then output
        -- a link to go to the point in the file with this
        -- name. Othewise, just output the character.
        if foundChars[char] then
            html:link ("#" .. char, char)
            html:addText (" ")
        else
            html:addText (char, " ")
        end
    end

    -- Lastly, finish up the bar and go on to the next row.
    html:addText (" ]")
    html.table:newRow ()

    -- Now just use the previous span to make an empty row, so that
    -- the nav information is seperated a bit.
    html.table:heading (true, nil, nil, "LEFT", "TOP", nil, span)
    html:addText ("&nbsp;")
    html.table:newRow ()
end

---------------------------------------------------------------------------

-- Given a list of games (an array of records from the GameList table)
-- generate a view of all of those games.
function generateGameList (db, gameList)
    -- Say how many games there are in the list. We go to pains here to
    -- make sure that the grammar is right regardless of how many games are
    -- listed.
    local label1 = "are "
    local label2 = " games"
    if table.getn (gameList) == 1 then label1 = "is " label2 = " game" end
    html:addText ('<FONT SIZE="-1">')
    html:addText ("There " , label1, table.getn (gameList) , label2, " in this list.<P>")
    html:addText ("</FONT>")

    -- Determine if we want to add nav information to the list. This is a list of letters at
    -- the top of the output that lets you quickly jump to the first game that starts with that
    -- letter. Only do this for more than 1 item.
    local navInfo = true
    if table.getn (gameList) <= 1 then navInfo = false end

    -- Track if this user can delete or edit games
    local canDelete = userInfo.canEditGame
    local canEdit   = userInfo.canEditGame

    -- If the user has javascript enabled, this confirmation will be used to get them to
    -- confirm what they are about to do.
    local deleteConfirm = "return confirm ('Any selected games will be completely removed from the system, " ..
                          "along with all session reports and other related information. " .. 
                          " Are you sure this is what you want to do?');"

    -- Start the form that will allow for the deletion of games.
    html.form:open (env.scriptName , "POST", nil, nil, deleteConfirm)

    -- Display the games in a table
    html.table:open (nil, "100%", nil, nil, "5", "gameList")

    -- Should we output navigation information?
    if navInfo then
        -- The nav output will be spit out as a row that spans the whole table.
        -- Figure out how many columns that will be. Base of 5, plus one if this
        -- user can edit things, plus one if the user can delete.
        local span = 5
        if canEdit then span = span + 1 end
        if canDelete then span = span + 1 end

        -- Now generate the navigation information. This is going to output some nav information
        -- as the first row in the table. It will also modify the games listed by adding an extra
        -- field that will tell the output code below that it should output a special name tag to
        -- allow jumping to that point in the page.
        generateNavInformation (db, gameList, span)
    end

    -- Output some column headers for the games. Not all columns will be used,
    -- depending on if this person is allowed to edit/delete games or not.
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Game Name")
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Category")
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Designer")
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Publisher")
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Players")

    -- Can the user edit properties? If so, include this column
    if canEdit then
        html.table:heading (true, nil, nil, "LEFT", "TOP")
        html:addText ("&nbsp;")
    end
    
    -- Can the user delete? If so, use this column.
    if canDelete then
        html.table:heading (true, nil, nil, "LEFT", "TOP")
        html:addText ("Delete")
    end

    -- Start the first actual row.
    html.table:newRow ()

    -- Scan over all of the games in the list.
    for index,row in gameList do
        -- Name of the game. We clip this value to 30 characters
        -- if it's longer to stop the table from getting too huge.
        --
        -- If this game has a "namedPointName" key, then use it to
        -- generate a named point link at the start of this game.
        --
        -- The namedSpot code does nothing if it's paramter is nil,
        -- so this saves us a test here.
        html.table:cell (true, nil, nil, "LEFT" , "TOP")
        html:namedSpot (row.namedPointName)
        local gameName = row.name
        if string.len (gameName) > 30 then
            gameName = string.sub (gameName, 1, 30) .. "..."
        end
        html:link (urls.gameDetails .. row.gameID , gameName)

        -- Primary category information for the game.
        html.table:cell (true, nil, nil, "LEFT" , "TOP")
        local category = getCategoryByID (db, row.categoryID)
        if category == nil then
            html:addText ("Unknown")
        else
            html:addText (category.categoryName)
        end

        -- Primary designer information for the game.
        html.table:cell (true, nil, nil, "LEFT" , "TOP")
        local designer = getDesignerByID (db, row.designerID)
        if designer == nil then
            html:addText ("Unknown")
        else
            html:addText (designer.designerName)
        end

        -- Primary publisher information for the game
        html.table:cell (true, nil, nil, "LEFT" , "TOP")
        local publisher = getPublisherByID (db, row.publisherID)
        if publisher == nil then
            html:addText ("Unknown")
        else
            html:addText (publisher.publisherName)
        end

        -- Number of players supported. If the min and max are the same,
        -- then only a single number is displayed.
        html.table:cell (true, nil, nil, "LEFT" , "TOP")
        if row.minPlayers == row.maxPlayers then
            html:addText (row.minPlayers)
        else
            html:addText (row.minPlayers .. " - " .. row.maxPlayers)
        end

        -- If the user can edit properties, then include a
        -- propperties link as well.
        if canEdit then
            html.table:cell (true, nil, nil, "LEFT" , "TOP")
            html:link (urls.gameProperties .. row.gameID , "Properties")
        end

        -- If the user can delete, then add the delete row.
        if canDelete then
            html.table:cell (true, nil, nil, "LEFT" , "TOP")
            html.form:checkbox ("delGameID" , row.gameID)
        end

        -- Get ready for the next row
        html.table:newRow ()
    end

    -- If the list of games has at least one entry and this user is allowed
    -- to delete, then we need to insert a submit button for the form. This
    -- isn't needed if this user can't delete or if there was nothing TO
    -- delete.
    if table.getn (gameList) > 0 and canDelete then
        -- The table has at least 5 columns in it. It will be 6 if this user
        -- can edit properties, 5 otherwise.
        local span = "6";
        if canEdit then span = "7" end

        -- Finish up with a submit button
        html.table:cell (true, nil, nil, "CENTER", "TOP", nil, nil, nil, span)
        html.form:submit ("Button", "Delete selected games", "game")
        html.form:hidden ("command", "GameDelete")
    end

    -- Close up the table and the form now
    html.table:close ()
    html.form:close ()
end

---------------------------------------------------------------------------

-- Called by html.form:select. Items is a list of items to display. userInfo
-- should contain the following keys:
-- ["name"] -> The name of the key in the item table to use as the name for the select.
-- ["key"] --> The name of the key in the item table to use as the value for the select.
-- ["selected"] -> If this key is present, it selects which value should be selected by default
--
-- Used to generate the dropdowns for publishers, designers and categories in
-- the genMultiFields function.
local genSelectWithDataList = function (items, userInfo)
    local selectedItem = tonumber (userInfo["selected"]) or -1
    local isSelected

    -- First, output a selection for "None". This will always happen, no matter what,
    -- since for all dropdowns we create we always want a none entry, even if the list
    -- data we've been given don't contain one.
    if selectedItem == 0 then isSelected = true else isSelected = nil end
    html.form:selectOption ("None" , 0 , isSelected)

    -- Now dump all of the games.
    for index,info in items do
        -- Don't output items that are 0 (None), since that was done above.
        if tonumber (info[userInfo.key]) ~= 0 then
            if selectedItem == info[userInfo.key] then isSelected = true else isSelected = nil end
            html.form:selectOption (info[userInfo.name] , info[userInfo.key], isSelected)
        end
    end
end

---------------------------------------------------------------------------

-- Used to generate "multi-fields", which is a dropdown list and a static text
-- field, so that the user can select something from the dropdown list and
-- also specify static text, or just specify static text, as needed.
--
-- This is called by genMultiFieldsExt to actually do it's work.
local genMultiFields = function (db, gameInfo, captionText, dropName, fieldName, itemList, keyList, canGenField)
    -- To start with, set up the heading. If the captionText is nil, we will insert
    -- a nonbreaking space instead.
    html.table:heading (true, nil, nil, "LEFT" , "TOP")
    html:addText (captionText or "&nbsp;")

    -- Now the cell for the dropdown list.
    html.table:cell (true , nil, nil, "LEFT" , "TOP")
    html.form:select (dropName, "1", nil, "gameDrop", nil, genSelectWithDataList , itemList , keyList)

    -- Now the cell for the text field
    html.table:cell (true , nil, nil, "LEFT" , "TOP" , nil, nil, nil, "2")

    -- If we can't generate the field, use a nonbreaking space instead.
    -- This is done when the user in question doesn't have permissions to add
    -- new things of whatever these fields are for.
    if canGenField then
        html.form:text (fieldName , gameInfo[fieldName] , "40" , "60", "gameText")
    else
        html:addText ("&nbsp;")
    end

    -- New row
    html.table:newRow ()
end

---------------------------------------------------------------------------

-- Uses the game form generator to generate a bunch of extended fields.
--
-- When gameInfo.gameID is nil, three fields are generated. When non-nil,
-- One field is generated for each of the selected items in the game, plus
-- a couple of empty fields are generated.
--
-- gameFieldList contains the fields that we care about
local genMultiFieldsExt = function (db, gameInfo, gameFieldList, fieldLabel , baseDropID , baseTextID, itemList, keys, canGenField)
    -- Count how many fields (total) have been output
    local totalFieldsOutput = 1

    -- If there is a gameID, then handle it first. This only happens when the form we're generating
    -- is for a property edit.
    if gameInfo.gameID ~= nil then
        -- Handle all of the items in the gameFieldList.
        for i = 1, table.getn (gameFieldList) do
            local dropName = baseDropID .. totalFieldsOutput
            local textName = baseTextID .. totalFieldsOutput

            -- Make sure that we select the right field before we select this item.
            -- If there is an item in the gameInfo for this field, then use that. This happens
            -- when the form is partially submitted but something has gone wrong. Failing that,
            -- use the item that comes from the actual game information instead. This happens when
            -- the form is initially generated.
            keys["selected"] = gameInfo[dropName] or gameFieldList[i].id

            -- Generate
            genMultiFields (db, gameInfo, fieldLabel, dropName, textName, itemList, keys, canGenField)

            -- Count a field, and make the fieldLabel nil
            totalFieldsOutput = totalFieldsOutput + 1
            fieldLabel = nil
        end
    end

    -- Now, generate some more fields. All of these will default to none.
    -- This will generate 3 blank fields. If you change this, you need to
    -- make sure that you also alter the form verifier, since it uses the 
    -- total number of fields in the existing game (if any) plus 3 to know
    -- how many category, publisher and designer fields it should collect
    -- from the form.
    for i = totalFieldsOutput,totalFieldsOutput + 2 do
        local dropName = baseDropID .. i
        local textName = baseTextID .. i

        -- Choose the item to be selected. If the gameInfo has the name that the form has,
        -- then select that (in this case, the dropList is being populated by a previous
        -- aborted submission. Otherwise, select 0, which is the "none".
        keys["selected"] = gameInfo[dropName] or "0"

        genMultiFields (db, gameInfo, fieldLabel , dropName , textName , itemList, keys, canGenField)
        fieldLabel = nil
    end
end

---------------------------------------------------------------------------

-- Generate a game form. If gameInfo is nil, then an empty form is generated,
-- which directs to "Add New Game". When gameInfo is not nil, then it is a
-- database record of a game whose properties are being edited. In that case,
-- a "Modify Properties" form is generated instead.
local generateGameForm = function (db, gameInfo)
    -- By default, pretend that we're updating a game
    local creatingGame = false

    local designerList = nil
    local publisherList = nil
    local categoryList = nil

    -- If it turns out that we ARE creating a game, then set the create flag
    -- to true, and pretend that gameInfo is an empty table, so that we can
    -- query from it as needed.
    if gameInfo == nil or gameInfo.gameID == nil then
        -- We're creating a game, so flag that and make sure that gameInfo
        -- is a valid table.
        creatingGame = true
        gameInfo = gameInfo or {}
    else
        -- We're editing an existing game. Get all of the designers,
        -- publishers and categories for it.
        designerList  = getAllDesignersForGameInfo (db, gameInfo)
        publisherList = getAllPublishersForGameInfo (db, gameInfo)
        categoryList  = getAllCategoriesForGameInfo (db, gameInfo)
    end

    -- Open the form and table that will contain our form elements
    html.form:open (env.scriptName , "POST")
    html.table:open (nil, nil, nil, nil, "5", "gameList")

    -- If we're not creating a game, then we're editing properties, so include
    -- a control bar for the game.
    if creatingGame == false then
        html.table:heading (true, nil, nil, "LEFT", "TOP", nil,  "4")
        generateGameControlBar (db, gameInfo.gameID, true)
        html.table:newRow ()
    end

    -- Start off asking for the name of the game.
    html.table:heading (true , nil, nil, "LEFT" , "TOP")
    html:addText ("Game Name:")
    html.table:cell (true, nil, nil, "LEFT" , "TOP")
    html.form:text ("gameName" , gameInfo.name, 40, 255, "gameText")

    -- Next, Say what the game is going to expand
    html.table:heading (true , nil, nil, "LEFT" , "TOP")
    html:addText ("Expansion for:")
    html.table:cell (true , nil, nil, "LEFT" , "TOP")
    local games = getGameList (db, true, gameInfo.gameID)
    local keys = {["name"] = "name", ["key"] = "gameID", ["selected"] = gameInfo.expandsGameID or "0"}
    html.form:select ("gameExpandsID", "1" , nil, "gameDrop", nil, genSelectWithDataList , games , keys)

    -- New row
    html.table:newRow ()

    -- Ask for the boardGameGeek ID number.
    html.table:heading (true , nil, nil, "LEFT" , "TOP", nil, nil, nil, "3")
    html:addText ("BGG Game ID")
    html.table:cell (true , nil, nil, "LEFT" , "TOP", nil, nil, nil, "3")
    html.form:text ("gameBGGID" , gameInfo.gamegeekID or "0" , "10" , "10", "gameText")

    -- New row
    html.table:newRow ()

    -- Query the number of players
    html.table:heading (true , nil, nil, "LEFT" , "TOP")
    html:addText ("Players:")
    html.table:cell (true , nil, nil, "LEFT" , "TOP")
    html.form:text ("gameMinPlay" , gameInfo.minPlayers or "2", "5" , "5")
    html:addText (" to ")
    html.form:text ("gameMaxPlay" , gameInfo.maxPlayers or "2" , "5" , "5")

    -- Ask for the playtime
    html.table:heading (true , nil, nil, "LEFT" , "TOP")
    html:addText ("Play Time (minutes):")
    html.table:cell (true , nil, nil, "LEFT" , "TOP")
    html.form:text ("gamePlayTime" , gameInfo.playTimeMinutes or "60" , "5" , "5", "gameText")

    -- New row
    html.table:newRow ()

    -- Set up categories.
    keys = {["name"] = "categoryName", ["key"] = "categoryID", ["selected"] = "0"}
    local categories = getCategoryList (db, true)
    genMultiFieldsExt (db, gameInfo, categoryList, "Categories:" , "gameCategory" , "gameNewCategory", categories, keys, userInfo.canEditCategory)

    -- Set up designers.
    keys = {["name"] = "designerName", ["key"] = "designerID", ["selected"] = "0"}
    local designers = getDesignerList (db, true)
    genMultiFieldsExt (db, gameInfo , designerList, "Designers:" , "gameDesigner" , "gameNewDesigner" , designers, keys, userInfo.canEditDesigner)

    -- Set up publishers
    keys = {["name"] = "publisherName", ["key"] = "publisherID", ["selected"] = "0"}
    local publishers = getPublisherList (db, true)
    genMultiFieldsExt (db, gameInfo , publisherList, "Publishers:" , "gamePublisher" , "gameNewPublisher" , publishers, keys, userInfo.canEditPublisher)

    -- Description.
    html.table:heading (true , nil, nil, "LEFT" , "TOP")
    html:addText ("Description:")
    html.table:cell (false, nil, nil, "LEFT" , "TOP", nil, nil, nil, "3")
    html.form:textArea ("gameDescription" , gameInfo.description, "80" , "10", nil, "gameTextArea")

    -- New row
    html.table:newRow ()

    -- Last row will hold the submit button
    html.table:cell (true, nil, nil, "CENTER", "TOP", nil, nil, nil, "4")

    -- Now, we are going to generate a button and some hidden fields, based on
    -- if we are creating an "add a game" form or a "modify game" form.
    if creatingGame == true then
        -- Creating a game, so submit to the new game insertion command
        html.form:submit ("Button", "Add this game to the Database", "game")
        html.form:hidden ("command", "GameNew")
    else
        -- Editing a game. Make sure we propogate the gameID
        html.form:submit ("Button", "Change Game Information", "game")
        html.form:hidden ("command", "GameProperties")
        html.form:hidden ("submitted", "yes")
        html.form:hidden ("gameID", gameInfo.gameID)
    end

    -- We're all done now
    html.table:close ()
    html.form:close ()

end

---------------------------------------------------------------------------

-- Figure out what form fields will be needed for the game in question.
-- if gameInfo is nil, then we assume that a new game is being added. otherwise,
-- we assume that the property form is going to be used. 
--
-- New game entries use a fixed number of fields, while property forms use that
-- same fixed number, plus one field for each data item that is already
-- contained (so for example if the game has 2 publishers, then you get the
-- normal number of fields, plus 2). This makes sure that the existing information
-- can be listed and at the same time property edits allow for addition of new
-- items.
local formFieldsNeeded = function (db, gameInfo)
    -- This is the list of fields that we know always should exist.
    local retVal = {"gameName" , "gameDescription", "gameExpandsID", "gameBGGID" ,
                    "gameMinPlay", "gameMaxPlay", "gamePlayTime"}

    -- By default, there are 3 design,publish and category field in the form.
    -- There will be more if this is a property form, as there may be more
    -- properties than these set.
    --
    -- If you change the value 3 here, you also need to update the form generator,
    -- as it only generates 3 blank fields.
    local designCount = 3
    local publishCount = 3
    local categoryCount = 3

    -- If we have a valid gameInfo, then we are a property field, so we have some
    -- extras to add.
    if gameInfo ~= nil then
        -- Insert some other values into the table for a property form. We need the
        -- gameID in this case, since that tells us what game's properties are being
        -- edited.
        table.insert (retVal, "gameID")
        table.insert (retVal, "submitted")

        -- Get the list of designers, publishers, and categories for this game 
        local designerList  = getAllDesignersForGameInfo (db, gameInfo)
        local publisherList = getAllPublishersForGameInfo (db, gameInfo)
        local categoryList  = getAllCategoriesForGameInfo (db, gameInfo)

        -- Using those lists, set up the design, publish and category counts to
        -- have that many fields, plus 3 more.
        designCount = table.getn (designerList) + 3
        publishCount = table.getn (publisherList) + 3
        categoryCount = table.getn (categoryList) + 3
    end

    -- Insert fields for the design, publish and category counts into the required
    -- field list.
    for i = 1, designCount do
        table.insert (retVal , "gameDesigner" .. i)
        table.insert (retVal , "gameNewDesigner" .. i)
    end

    for i = 1, publishCount do
        table.insert (retVal , "gamePublisher" .. i)
        table.insert (retVal , "gameNewPublisher" .. i)
    end

    for i = 1, categoryCount do
        table.insert (retVal , "gameCategory" .. i)
        table.insert (retVal , "gameNewCategory" .. i)
    end

    -- Return the list now
    return retVal, designCount, publishCount, categoryCount
end

---------------------------------------------------------------------------

-- Given a game info (which might be nil, if this is an Add form), and a
-- list of form fields (which shouldn't be nil), validate to make sure that
-- all of the information contained in the fields is sensical.
--
-- The return value is nil on success, or some error text on failure.
local validateFormFields = function (db, gameInfo, fields, designCount, publishCount, categoryCount)
    -- Go through all of the easy form fields and give them a good value.
    fields.gameName        = string.trimleadingtrailing (fields.gameName);
    fields.gameDescription = string.trimleadingtrailing (fields.gameDescription);
    fields.gameExpandsID   = tonumber (fields.gameExpandsID) or -1
    fields.gameBGGID       = tonumber (fields.gameBGGID)     or -1
    fields.gameMinPlay     = tonumber (fields.gameMinPlay)   or -1
    fields.gameMaxPlay     = tonumber (fields.gameMaxPlay)   or -1
    fields.gamePlayTime    = tonumber (fields.gamePlayTime)  or -1

    -- Name is mandatory, although description is not.
    if fields.gameName == "" then return ERR_GAME_NOGAMENAME end

    -- Check out the numeric stuff. All must be >= 0 (hence the values above).
    if fields.gameExpandsID < 0 then return ERR_GAME_BADEXPANSION end
    if fields.gameBGGID     < 0 then return ERR_GAME_BADBGGID end
    if fields.gameMinPlay   < 0 then return ERR_GAME_BADMINPLAY end
    if fields.gameMaxPlay   < 0 then return ERR_GAME_BADMAXPLAY end
    if fields.gamePlayTime  < 0 then return ERR_GAME_BADPLAYTIME end

    -- Make sure that the expansion ID is a valid game. An expansion of 0 is always ok,
    -- it means that there is no expansion mapping.
    if fields.gameExpandsID ~= 0 and getGameByID (db, fields.gameExpandsID) == nil then
        return ERR_GAME_BADEXPANSION
    end

    -- Validate all designers
    for i = 1, designCount do
        local existName = "gameDesigner" .. i
        local newName   = "gameNewDesigner" .. i

        -- Fix fields
        fields[existName] = tonumber (fields[existName]) or -1
        fields[newName]   = string.trimleadingtrailing (fields[newName])

        -- Bad if the numeric value of the existing ID is less than 0, or if it
        -- does not specify a valid entry.
        if fields[existName] < 0 or getDesignerByID (db, fields[existName]) == nil then
            return ERR_GAME_BADDESIGNER
        end
    end

    -- Validate all publishers
    for i = 1, publishCount do
        local existName = "gamePublisher" .. i
        local newName   = "gameNewPublisher" .. i

        -- Fix fields
        fields[existName] = tonumber (fields[existName]) or -1
        fields[newName]   = string.trimleadingtrailing (fields[newName])

        -- Bad if the numeric value of the existing ID is less than 0, or if it
        -- does not specify a valid entry.
        if fields[existName] < 0 or getPublisherByID (db, fields[existName]) == nil then
            return ERR_GAME_BADPUBLISHER
        end
    end

    -- Validate all categories
    for i = 1, categoryCount do
        local existName = "gameCategory" .. i
        local newName   = "gameNewCategory" .. i

        -- Fix fields
        fields[existName] = tonumber (fields[existName]) or -1
        fields[newName]   = string.trimleadingtrailing (fields[newName])

        -- Bad if the numeric value of the existing ID is less than 0, or if it
        -- does not specify a valid entry.
        if fields[existName] < 0 or getCategoryByID (db, fields[existName]) == nil then
            return ERR_GAME_BADCATEGORY
        end
    end

    -- If the minimum players is greater than the maximum players, then swap them around
    if fields.gameMinPlay > fields.gameMaxPlay then
        fields.gameMinPlay, fields.gameMaxPlay = fields.gameMaxPlay, fields.gameMinPlay
    end

    -- All good if we get here
    return nil
end

---------------------------------------------------------------------------

-- Given the array list, insert newValue IF it doesn't exist in the list and
-- it isn't the same as primaryValue.
local insertInList = function (list, newValue, primaryValue)
    -- If the value is the same as primaryValue, then we're done.
    if newValue == primaryValue then return end

    -- Scan the list. If we find a match, return
    for index,oldVal in list do
        if oldVal == newVal then return end
    end

    -- Must be ok to insert it
    table.insert (list, newValue)
end

---------------------------------------------------------------------------

-- Generate a list of things (Designer, Category, Publisher) from the
-- list of dropdowns and any entered text.
--
-- args is the form arguments
--
-- count and listType control what we're processing. listType is one of
-- "Designer", "Category" or "Publisher", and count says how many form elements
-- of that type to expect. These two items are used to synthesize field names in
-- the form arguments.
--
-- Functor is a function that takes a db connection and a name of something. It should
-- return back the ID of that item, or create it and return the new ID. This is how the
-- text fields on the form get turned into new designers/publishers/categories on the fly.
--
-- canMakeNew should be a boolean that is true when the current user can add new items, 
-- and false when they can't. This makes the code ignore the form fields if needed.
--
-- Return values are the primary item and the list of values, in that order. The primary
-- is not listed in the list of items, and the list is unique. 
--
-- During processing, the primary and the list will never be filled with the value "0",
-- so if the primary comes back 0, that means that the list is empty.
local generateList = function (db, args, count, listType, functor, canMakeNew)
    local retVal = {}
    local primary = nil

    -- Loop over all of the fields
    for i = 1,count do
        -- Determine the field names for the text field and the field that is in
        -- the dropdown. 
        local newNameKey      = "gameNew" .. listType .. i
        local existingNameKey = "game" .. listType .. i

        -- Get the string that is associated with the write in entry, and convert the 
        -- existing value to a number.
        local newName      = string.trimleadingtrailing (args[newNameKey])
        local existingName = tonumber (args[existingNameKey])

        -- If we're not allowed to add new items of this type, then pretend that
        -- newName is empty.
        if canMakeNew == false then newName = "" end

        -- First, handle the existing name key (no need to add this to the database).
        -- Second, handle the possible new name, if it's set. In both cases we do
        -- the following:
        --  o If the ID is 0, skip it
        --  o If the primary isn't set, set it now
        --  o If this ID isn't the primary and isn't in the list, then add it
        
        -- Check to see if this is valid or not.
        if existingName ~= 0 then
            -- If the primary isn't set, then set it now. Otherwise, isert this value
            -- into the list.
            if primary == nil then
                primary = existingName
            else
                insertInList (retVal, existingName, primary)
            end
        end

        -- Now, did we get a new name that had something in it?
        if newName ~= "" then
            local addedName = functor (db, newName)

            if addedName ~= 0 then
                -- Now, set the primary or add things into the list, as approprite.
                if primary == nil then
                    primary = addedName
                else
                    insertInList (retVal, addedName, primary)
                end
            end
        end
    end

    -- If the primary is still nil, set it to be 0.
    if primary == nil then primary = 0 end

    -- Return now
    return primary, retVal
end

---------------------------------------------------------------------------

-- This takes validated form arguments and uses them to update the properties
-- for a game. The return value is the gameID of the game that is currently
-- being edited.
local updateGameProperties = function (db, args, designCount, publishCount, categoryCount)
    local primaryDesigner = nil
    local designList = nil
    local primaryPublisher = nil
    local publishList = nil
    local primaryCategory = nil
    local categoryList = nil

    -- For each of designer, publisher and category, do the following:
    --
    -- Use a combination of the dropdown fields and all of the text fields
    -- that allow for new additions to determine the complete list of items.
    --
    -- The function will return back the item that should be primary (first
    -- one found), and then a list of all other items, which might be empty.
    primaryDesigner, designList = generateList (db, args , designCount, "Designer", selectOrCreateDesigner, userInfo.canEditDesigner)
    primaryPublisher, publishList = generateList (db, args , publishCount, "Publisher", selectOrCreatePublisher, userInfo.canEditPublisher)
    primaryCategory, categoryList = generateList (db, args , categoryCount, "Category", selectOrCreateCategory, userInfo.canEditCategory)

    -- Update the core game information now.
    changeGamePropertiesForGameID (db, args.gameID, args.gameBGGID, args.gameName, args.gameDescription,
                primaryDesigner, primaryPublisher, primaryCategory,
                args.gameMinPlay, args.gameMaxPlay, args.gamePlayTime, args.gameExpandsID)

    -- Insert all of the extra designers.
    deleteDesignerPlacementByGame (db, args.gameID)
    for i = 1,table.getn (designList) do
        addDesignerPlacement (db, args.gameID, designList[i])
    end

    -- Insert all of the extra publishers.
    deletePublisherPlacementByGame (db, args.gameID)
    for i = 1,table.getn (publishList) do
        addPublisherPlacement (db, args.gameID, publishList[i])
    end

    -- Insert all of the extra categories.
    deleteCategoryPlacementByGame (db, args.gameID)
    for i = 1,table.getn (categoryList) do
        addCategoryPlacement (db, args.gameID, categoryList[i])
    end

    -- Say how happy we are
    addLog (db, "Modified game: " .. args.gameName)
    html:addText (successText ("Modified properties for '" .. args.gameName .. "'"), "<P>")

    -- Return the gameID of the game we're editing
    return args.gameID
end

---------------------------------------------------------------------------

-- This takes validated form arguments, and adds a new game. The return value
-- is the gameID of the new game.
local addNewGame = function (db, args, designCount, publishCount, categoryCount)
    local primaryDesigner = nil
    local designList = nil
    local primaryPublisher = nil
    local publishList = nil
    local primaryCategory = nil
    local categoryList = nil

    -- For each of designer, publisher and category, do the following:
    --
    -- Use a combination of the dropdown fields and all of the text fields
    -- that allow for new additions to determine the complete list of items.
    --
    -- The function will return back the item that should be primary (first
    -- one found), and then a list of all other items, which might be empty.
    primaryDesigner, designList = generateList (db, args , designCount, "Designer", selectOrCreateDesigner, userInfo.canEditDesigner)
    primaryPublisher, publishList = generateList (db, args , publishCount, "Publisher", selectOrCreatePublisher, userInfo.canEditPublisher)
    primaryCategory, categoryList = generateList (db, args , categoryCount, "Category", selectOrCreateCategory, userInfo.canEditCategory)

    -- Insert the core information now and save the game ID that we got from it
    local newGameID = addNewGame (db, args.gameBGGID, args.gameName, args.gameDescription,
                        primaryDesigner, primaryPublisher, primaryCategory,
                        args.gameMinPlay, args.gameMaxPlay,
                        args.gamePlayTime, args.gameExpandsID)
    newGameID = tonumber (newGameID)

    -- Now, insert all of the other designers for the game.
    for i = 1,table.getn (designList) do
        addDesignerPlacement (db, newGameID, designList[i])
    end

    -- Now, insert all of the other publishers for the game.
    for i = 1,table.getn (publishList) do
        addPublisherPlacement (db, newGameID, publishList[i])
    end

    -- Now, insert all of the other categories for the game.
    for i = 1,table.getn (categoryList) do
        addCategoryPlacement (db, newGameID, categoryList[i])
    end

    -- Say how happy we are
    addLog (db, "Added game: " .. args.gameName)
    html:addText (successText ("Added the new game '" .. args.gameName .. "'"), "<P>")

    -- Return the gameID back.
    return newGameID
end

---------------------------------------------------------------------------

-- This takes a table that may be form arguments or may be game information
-- from a database. Any form fields that have a value are copied back into
-- this same table with the names that they would have when they come out 
-- of the database.
--
-- This allows us to treat form arguments as a regular game query result.
local fixFormArgs = function (db , info)
    -- If there is a gameID, then just leave.
    if info.gameID ~= nil then return end

    -- Copy any existing form fields back using the format that we would get them from
    -- the database.
    if info.gameName        ~= nil then info.name            = info.gameName end
    if info.gameDescription ~= nil then info.description     = info.gameDescription end
    if info.gameExpandsID   ~= nil then info.expandsGameID   = info.gameExpandsID end
    if info.gameBGGID       ~= nil then info.gamegeekID      = info.gameBGGID end
    if info.gameMinPlay     ~= nil then info.minPlayers      = info.gameMinPlay end
    if info.gameMaxPlay     ~= nil then info.maxPlayers      = info.gameMaxPlay end
    if info.gamePlayTime    ~= nil then info.playTimeMinutes = info.gamePlayTime end
end

---------------------------------------------------------------------------

-- This command will find and view all games by a particular designer.
function commandGameViewDesigner (db)
    -- Collect form arguments. We should have a designerID to view.
    -- If we get one, then try to find the designer.
    local designer = nil
    local args = collectFormArguments ({"designerID"})
    if args ~= nil then
        designer = getDesignerByID (db, args.designerID)
    end

    -- Header for the page, which is based on the designer.
    if designer == nil then
        generatePageHeader (thisPageSection, "Games by Designer", "gamesByDesigner")
    else
        generatePageHeader (thisPageSection, "Games Designed by '" .. designer.designerName .. "'", "gamesByDesigner")
    end

    -- Generate a button to go back to the main Game page
    html.form:singleButtonForm ("Return to Designers Page" , "Designer" , env.scriptName , "POST" , nil, nil, "game")
    html.form:singleButtonForm ("Return to Games Page" , "Game" , env.scriptName , "POST" , nil, nil, "game")
    html:addText ("<P>")

    -- If we have no designer info, then we can't look up any games by designer.
    if designer == nil then html:addText (errorText (ERR_GAME_NODESIGNERID) , "<P>") return end

    -- Find all games that are designed by this designer.
    local gameList = getAllGamesByDesigner (db, designer.designerID)

    -- Generate the list of games now
    generateGameList (db, gameList)
end

---------------------------------------------------------------------------

-- View games by a particular publisher
function commandGameViewPublisher (db)
    -- Collect form arguments. We should have a publisherID to view.
    -- If we get one, then try to find the publisher.
    local publisher = nil
    local args = collectFormArguments ({"publisherID"})
    if args ~= nil then
        publisher = getPublisherByID (db, args.publisherID)
    end

    -- Header for the page, which is based on the publisher.
    if publisher == nil then
        generatePageHeader (thisPageSection, "Games by Publisher", "gamesByPublisher")
    else
        generatePageHeader (thisPageSection, "Games Published by '" .. publisher.publisherName .. "'", "gamesByPublisher")
    end

    -- Generate a button to go back to the main Game page
    html.form:singleButtonForm ("Return to Publishers Page" , "Publisher" , env.scriptName , "POST" , nil, nil, "game")
    html.form:singleButtonForm ("Return to Games Page" , "Game" , env.scriptName , "POST" , nil, nil, "game")
    html:addText ("<P>")

    -- If we have no Publisher info, then we can't look up any games by publisher.
    if publisher == nil then html:addText (errorText (ERR_GAME_NOPUBLISHERID) , "<P>") return end

    -- Find all games published by this publisher.
    local gameList = getAllGamesByPublisher (db, publisher.publisherID)

    -- Generate the list of games now
    generateGameList (db, gameList)
end

---------------------------------------------------------------------------

-- View games by a particular category
function commandGameViewCategory (db)
    -- Collect form arguments. We should have a categoryID to view.
    -- If we get one, then try to find the category.
    local category = nil
    local args = collectFormArguments ({"categoryID"})
    if args ~= nil then
        category = getCategoryByID (db, args.categoryID)
    end

    -- Header for the page, which is based on the category.
    if category == nil then
        generatePageHeader (thisPageSection, "Games by Category", "gamesByCategory")
    else
        generatePageHeader (thisPageSection, "Games in the Category '" .. category.categoryName .. "'", "gamesByCategory")
    end

    -- Generate a button to go back to the main Game page
    html.form:singleButtonForm ("Return to Categories Page" , "Category" , env.scriptName , "POST" , nil, nil, "game")
    html.form:singleButtonForm ("Return to Games Page" , "Game" , env.scriptName , "POST" , nil, nil, "game")
    html:addText ("<P>")

    -- If we have no Category info, then we can't look up any games by category.
    if category == nil then html:addText (errorText (ERR_GAME_NOCATEGORYID) , "<P>") return end

    -- Get all games in this category.
    local gameList = getAllGamesByCategory (db, category.categoryID)

    -- Generate the list of games now
    generateGameList (db, gameList)
end

---------------------------------------------------------------------------

-- View expansions for a given game.
function commandGameViewExpansion (db)
    -- Collect form arguments. We should have a gameID to view.
    local gameInfo = nil
    local args = collectFormArguments ({"gameID"})
    if args ~= nil then
        gameInfo = getGameByID (db, args.gameID)
    end

    -- Header for the page.
    if gameInfo == nil then
        generatePageHeader (thisPageSection, "Expansions for Game", "gamesExpansions")
    else
        generatePageHeader (thisPageSection, "Expansions for '" .. gameInfo.name .. "'", "gamesExpansions")
    end

    -- Generate a button to go back to the games page.
    html.form:singleButtonForm ("Return to Games Page" , "Game" , env.scriptName , "POST" , nil, nil, "game")
    html:addText ("<P>")

    -- With no args, leave now
    if args == nil then html:addText (errorText (ERR_GAME_NOGAMEID) , "<P>") return end

    -- Get the information for this game.
    local gameList = getExpansionsForGameID (db, gameInfo.gameID, true)

    -- Generate the list of games now
    generateGameList (db, gameList)
end

---------------------------------------------------------------------------

-- View a detailed game listing.
function commandGameView (db)
    -- Header for the page
    generatePageHeader (thisPageSection, "Detailed View", "gamesView")

    -- Generate a button to go back to the main Game page
    html.form:singleButtonForm ("Return to Games Page" , "Game" , env.scriptName , "POST" , nil, nil, "game")
    html:addText ("<P>")

    -- Collect form arguments. We should have a gameID to view.
    local args = collectFormArguments ({"gameID"})
    if args == nil then html:addText (errorText (ERR_GAME_NOGAMEID) , "<P>") return end

    -- Get the information for this game. Bail if THAT didn't work.
    local gameInfo = getGameByID (db, args.gameID)
    if gameInfo == nil then html:addText (errorText (ERR_GAME_NOGAMEID) , "<P>") return end

    -- If the game information provided expands a game, then figure out the name of that
    -- game.
    local expandedName = "Unknown"
    if gameInfo.expandsGameID ~= 0 then
        local info = getGameByID (db, gameInfo.expandsGameID)
        if info then
            expandedName = info.name
        end
    end

    -- Get the complete list of designers, publishers and categories. This will always return an array
    -- with at least one element, even if it's the string "Unknown".
    --
    -- In all cases, the element at the first location is the primary, and the rest (if any) are the
    -- secondaries, sorted alphabetically.
    local designerList  = getAllDesignersForGameInfo (db, gameInfo)
    local publisherList = getAllPublishersForGameInfo (db, gameInfo)
    local categoryList  = getAllCategoriesForGameInfo (db, gameInfo)

    -- Open a table for the game. 
    html.table:open (nil, "70%", nil, nil, "5", "gameList")

    -- Insert a control bar for the game
    html.table:heading (true, nil, nil, "LEFT" , "TOP", nil,  "2")
    generateGameControlBar (db, gameInfo.gameID , false)
    html.table:newRow ()

    -- Open up a doublewide header cell that names the game. It needs to
    -- span 2 columns.
    html.table:heading (true, nil, nil, "LEFT" , "TOP", nil, "2")
    html:addText (gameInfo.name)

    -- If this game is an expansion for another game, then list it here
    if gameInfo.expandsGameID ~= 0 then
        html:addText (" (Expansion for '")
        html:link (urls.gameDetails .. gameInfo.expandsGameID , expandedName)
        html:addText ("')")
    end

    -- Check how many expansions this game has. If there are any, then generate text
    -- to say how many, and link to viewing all games that are expansions of this game.
    local expansionCount = tonumber (expansionCountForGameID (db, gameInfo.gameID) or "0")
    if expansionCount ~= 0 then
        html:addText (" (has ")
        html:link (urls.gamesByExpansion .. gameInfo.gameID , expansionCount)
        if expansionCount == 1 then html:addText (" expansion)") else html:addText (" expansions)") end
    end
    
    -- New row.
    html.table:newRow ()

    -- Open up the first cell in the table, which is going to hold the
    -- designer, publisher and play information.
    html.table:cell (true, "1%", nil, "LEFT" , "TOP")
    genPlayInfo (db, gameInfo, designerList, publisherList)

    -- Now, create a second cell, which will contain the category list.
    html.table:cell (true, "99%", nil, "LEFT" , "TOP")
    genCategoryInfo (gameInfo , categoryList)

    -- New row now. We will create a cell that will span both columns,
    -- and it will contain the description information for this game.
    html.table:newRow ()
    html.table:cell (nil, nil, nil, "JUSTIFY" , "TOP", nil, nil, nil, "2") 
    local fixedDescription = string.gsub (gameInfo.description or "", "\n", "<p>")
    html:addText (fixedDescription)

    -- Close up the table, done now
    html.table:close ()
end
---------------------------------------------------------------------------

-- Handles the deletion of games from the system.
function commandGameDelete (db)
    -- Header for the page
    generatePageHeader (thisPageSection, "Delete Games")

    -- Link back to the main game page.
    html.form:singleButtonForm ("Return to Games Page" , "Game" , env.scriptName , "POST" , nil, nil, "game")

    -- Leave with an error if this user can't delete games from the
    -- system.
    if userInfo.canDeleteGame == false then
        addLog (db , "Attempted to delete a game")
        html:addText (errorText (ERR_GAME_NODELPRIV) , "<P>")
        return
    end
    
    -- Collect the form arguments, which is a list of the games to
    -- be deleted, according to their ID's. If we don't get any
    -- games to delete, then complain and leave.
    local formArgs = collectFormArguments ({"delGameID"})
    if formArgs == nil then 
        addLog (db, "Attempted delete with no games selected");
        html:addText (errorText (ERR_DEL_NOGAMES) , "<P>");
        return
    end

    -- The value of delGameID can be a single string, or an array, depending on
    -- how many games are to be deleted. Make sure it's a table in case it's not.
    if type (formArgs.delGameID) ~= "table" then
        formArgs.delGameID = {formArgs.delGameID}
    end

    -- We got confirmation on deletion prior to getting here, when the browser asked if
    -- they were sure. So, loop over all games and remove them.
    for index,game in formArgs.delGameID do
        -- Get the information on the game to delete.
        local delGameInfo = getGameByID (db, game)

        -- If we couldn't find the information for the game to delete, then complain about
        -- it now.
        if delGameInfo == nil then
            -- Error.
            html:addText (errorText (ERR_DEL_NOFINDGAME) , "<BR>");
        else
            -- The delete is working. First, delete this game from the game list, and then
            -- remove it from all of the placement tables that keep track of publishers,
            -- designers and categories.
            deleteGameByID (db, game)
            deleteCategoryPlacementByGame (db, game)
            deleteDesignerPlacementByGame (db, game)
            deletePublisherPlacementByGame (db, game)

            -- Dump all session reports for this game. To do that, we must first select
            -- all session reports for this game.
            local reports = getSessionReportsForGameID (db, game)

            -- For each saved session report, we must first delete the session report with the
            -- sessionID, and then all sessionReportEntry fields with that sessionID
            for index,report in reports do
                deleteSessionReportByID (db, report.sessionID)
                deleteSessionReportEntriesForSessionID (db, report.sessionID)
            end

            -- We're done.
            html:addText (successText ("Deleted game '" .. delGameInfo.name .. "'") , "<BR>")
            addLog (db , "Deleted game '" .. delGameInfo.name .. "'")
        end
    end
end

---------------------------------------------------------------------------

-- Command to handle the properties editing of a game that is already in the system.
function commandGameProperties (db)
    -- Header for the page
    generatePageHeader (thisPageSection, "Edit Game Properties", "gamesEdit")

    -- Generate a link back to the main page.
    html.form:singleButtonForm ("Return to Games Page" , "Game" , env.scriptName , "POST" , nil, nil, "game")

    -- Stop now with an error if this user isn't allowed to edit
    -- game properties.
    if userInfo.canEditGame == false then
        addLog (db, "Attempted to edit a game")
        html:addText (errorText (ERR_GAME_NOEDITPRIV) , "<P>")
        return
    end

    -- Collect initial form arguments. There should be a gameID of the game to be
    -- editing. Leave on error if there isn't.
    local args = collectFormArguments ({"gameID"})
    if args == nil then html:addText (errorText (ERR_GAME_NOGAMEID) , "<P>") return end

    -- Try to get the information for this game. If that doesn't work, then 
    -- we need to return with an error.
    local gameInfo = getGameByID (db, args.gameID)
    if gameInfo == nil then html:addText (errorText (ERR_GAME_NOGAMEID) , "<P>") return end

    -- Collect the list of needed arguments for this game. The number of fields needed for
    -- categories, designers and publishers is dependant on how many records are currently
    -- in place for this game.
    local neededArguments, designCount, publishCount, categoryCount = formFieldsNeeded (db, gameInfo)
    args = collectFormArguments (neededArguments)

    -- If we got arguments, then we can verify them and possibly update the game information.
    -- If the args are nil, that's ok. It just means we're at the properties page but nothing
    -- has been submitted yet.
    if args then
        -- Validate the firm fields. This returns error text on an error or nil on success.
        local result = validateFormFields (db, gameInfo, args, designCount, publishCount, categoryCount)

        -- Was there a problem validating the form?
        if result then
            -- Say what the error was.
            html:addText (errorText (result) , "<P>")

            -- Set the gameInfo to be the args, and then fix them up. This way the form will regenerate
            -- with the values it was submitted with instead of the game's actual properties. For this to
            -- work, we need to temporarily save the gameID, then remove it from the dict while we fix the
            -- arguments, since the argument fixer won't do anything if there is a gameID present.
            local gameID = args.gameID
            args.gameID = nil
            gameInfo = args
            fixFormArgs (db, gameInfo)
            gameInfo.gameID = gameID
        else
            -- Things worked, so update the game's properties using the arguments given.
            local gameID = updateGameProperties (db, args, designCount, publishCount, categoryCount)
            
            -- Get rid of the arguments now that we've been added.
            args = nil

            -- Try to retrieve the new information for the game. If this doesn't work, then the 
            -- the form will regenerate with an empty form, but that's ok.
            if gameID then
                gameInfo = getGameByID (db, gameID)
            end
        end

    end

    -- Generate the form now. This might use the actual game properties, or it might use
    -- previously submitted values, depending on the state of things.
    generateGameForm (db, gameInfo)
end

---------------------------------------------------------------------------

-- This handles the addition of new games to the database.
function commandGameNew (db)
    -- Header for the page
    generatePageHeader (thisPageSection, "Add a Game", "gamesAdd")

    -- Generate a button to get back to the main game page.
    html.form:singleButtonForm ("Return to Games Page" , "Game" , env.scriptName , "POST" , nil, nil, "game")

    -- Leave now with an error if this user isn't allowed to add new games
    -- to the system.
    if userInfo.canAddGame == false then
        addLog (db, "Attempted to add a new game")
        html:addText (errorText (ERR_GAME_NOADDPRIV) , "<P>")
        return
    end

    -- Collect the list of needed arguments, then try to get them from the form.
    -- By passing nil to formFieldsNeeded, we tell if that we need to generate fields
    -- for a new game. This gets it to give us a default number of elements for
    -- adding in categories, publishers and designers. If this wasn't nil, then we would
    -- get enough fields for all of the current item of each type, plus a few more extras.
    local neededArguments, designCount, publishCount, categoryCount = formFieldsNeeded (db, nil)
    local args = collectFormArguments (neededArguments)

    -- This is nil to start, but might get filled up with game information if 
    -- the form is successfully submitted. In that case the form below will get
    -- generated as if it's a property change, allowing a newly added game to
    -- be edited right away.
    local gameInfo = nil

    -- If we got our form arguments, then the form has been submitted, so try to handle it.
    if args then
        -- Validate the form fields. On success, this returns nil, while on error it gives us
        -- the reason for the error.
        local result = validateFormFields (db, nil, args, designCount, publishCount, categoryCount)

        -- Was there a problem validating the form?
        if result then
            -- Yes, complain.
            html:addText (errorText (result) , "<P>")
        else
            -- Add the game into the system now, and save it's gameID.
            local gameID = addNewGame (db, args, designCount, publishCount, categoryCount)
            
            -- Now that we know that the add is complete, throw away the form arguments,
            -- so that they won't be used to re-generate the form below.
            args = nil

            -- Try to get the information for the newly added game. If this doesn't work, then
            -- the form will regenerate with an empty form, but that's ok.
            if gameID then
                gameInfo = getGameByID (db, gameID)
            end
        end
    end

    -- If we don't have game information, try to use form arguments instead. This may still come
    -- up as nil, but that's ok. The idea is that on the first invocation, this will be nil.
    -- If the form is submitted but otherwise broken during validation, the form values submitted
    -- will be used to fill things out. If the form is submitted successfully, then the form will
    -- be generated as properties.
    gameInfo = gameInfo or args
    if gameInfo then fixFormArgs (db, gameInfo) end

    -- Generate the game form now. We need to pass in the counts of fields that should be
    -- used so that the generator can generate all of the needed fields.
    generateGameForm (db, gameInfo, designCount, publishCount, categoryCount)
end

---------------------------------------------------------------------------

-- The main game command, displays the complete current game list.
function commandGame (db)
    -- Header for the page
    generatePageHeader (thisPageSection, nil, "gamesBase")

    -- Add a button for creating a new game, if the user is allowed to do so.
    if userInfo.canAddGame then
        html.form:singleButtonForm ("Add a Game" , "GameNew" , env.scriptName , "POST" , nil, nil, "game")
        html:addText ("<P>")
    end

    -- Collect all games in the system, and then
    -- display them.
    local gameList = getGameList (db, true)
    generateGameList (db, gameList)
end

---------------------------------------------------------------------------
--- MAIN CODE -------------------------------------------------------------
---------------------------------------------------------------------------

