---------------------------------------------------------------------------
--- LOCAL VARS ------------------------------------------------------------
---------------------------------------------------------------------------

-- What section to use in page headers in this file.
local thisPageSection = "UNKNOWN COMMAND"                         

-- Error messages.
local ERR_CMD_NOGOOD = [[The command given is unknown. Please 
                         select an item from the menu and try again. 
                         If the problem persists, contact your administrator.]]

---------------------------------------------------------------------------
--- FUNCTIONS -------------------------------------------------------------
---------------------------------------------------------------------------

-- Takes a time interval that represents some number of minutes, and returns
-- back a string that represents that number of minutes in HH:MM format.
--
-- An interval of 0 (00:00) is returned back as the string ??:??. This is
-- assumed to be an invalid interval, since every game should take at least
-- 1 minute to finish.
function toTimeValue (interval)
    interval = tonumber (interval) or 0
    if interval == 0 then return "??:??" end

    -- The duration is in minutes, so split it into hours and minutes. math.floor
    -- rounds it's value down to the lowest integral value, to get the hours, while
    -- math.mod calculates the remainder.
    local hours = math.floor (interval / 60)
    local minutes = math.mod (interval, 60)

    -- Fix it up
    return string.format ("%d:%02d", hours, minutes)
end

---------------------------------------------------------------------------

-- Takes a string in the format of HH:MM OR MMMM (that is, if there is no :, it
-- is assumed to be minutes) and returns back the time as an actual duration.
--
-- For an input of HH:MM, this returns HH * 60 + MM.
--
-- For an input of MMMM, this returns MMMM.
--
-- The number of digits before and after the optional colon can be anything, and there can be
-- any leading or trailing space as desired.
-- 
-- For HH:MM format, if MM is larger than 60, this will roll over into a new hour.
--
-- If the format of the string is not recognized, then 0 is returned.
function fromTimeValue (timeStr)
    -- Fix up the incoming time, just in case.
    timeStr = string.trimleadingtrailing (timeStr or "")

    -- If we can convert this string to a number directly, then
    -- just return that value. This would be the MMMM format.
    local interval = tonumber (timeStr)
    if interval then return interval end

    -- Split the string into hour and minute portions.
    local start,stop,hourStr,minuteStr = string.find (timeStr, "(%d+):(%d+)")

    -- If we didn't get a match, then return back 0.
    if start == nil or hourStr == nil or minuteStr == nil then return 0 end

    -- Try to convert both parts to numbers. If either fails, return 0.
    local hours   = tonumber (hourStr)
    local minutes = tonumber (minuteStr)

    if hours == nil or minutes == nil then return 0 end

    -- All good
    return hours * 60 + minutes
end

---------------------------------------------------------------------------


-- This special command is invoked when the url submitted has a command that
-- is not known.
function commandNotKnown (db)
    -- Header for the page
    generatePageHeader (thisPageSection, nil, "badCommand")

    html:addText (errorText (ERR_CMD_NOGOOD) , "<P>")
end

---------------------------------------------------------------------------

-- This is a special link generator which outputs links wrapped in square
-- brackets and nonbreaking spaces, for use ine control bar generation.
--
-- If the url passed in is nil, then no link is generated and just regular
-- text is added instead.
function wrappedLink (url, text, target)
    html:addText ("&nbsp; [ &nbsp;")
    if url then html:link (url, text, target) else html:addText (text) end
    html:addText ("&nbsp; ]")
end

---------------------------------------------------------------------------

-- This helper function will return you the authToken needed for the given
-- username and password. This will always return the same value for the
-- same username/password combination.
function getAuthToken (username, password)
    return ext.crypt (username .. ":" .. password , "69")
end

---------------------------------------------------------------------------

-- This function should be called when page generation is finished. It makes
-- sure that the database is closed, and then flushes all HTML to finish
-- things up.
function donePage ()
    if db ~= nil then db:close () end
    html:endPage ()
end

---------------------------------------------------------------------------

-- Checks to see if the database opened properly. If it did, then true is
-- returned. If not, then an error message is added to the page and it is
-- closed up, and false is returned.
function checkDB (db, errStr)
    -- If the Database wasn't opened, then complain and return.
    --
    -- The only time this actually happens is when the database file doesn't
    -- exist yet, and the current user doesn't have permissions to create
    -- a new file. Otherwise, trying to open a nonexistant database just
    -- creates a new database.
    if db == nil then
        html:addText (errorText ("Unable to open db: " .. errStr) , "<P>")
        donePage ()
        return false
    end

    -- Worked
    return true
end

---------------------------------------------------------------------------

-- Return back the text required as the HREF part of a link so that when the
-- link is clicked javascript is invoked to open up a new window to display
-- the URL provided.
local getJavascriptWindowCodeForURL = function (url)
    return "javascript:window.open ('" .. url .. "','gamedbHelpWin','scrollbars=yes,menubar=no,height=500,width=600,resizable=no,toolbar=no,location=no,status=no'); return false;"
end

---------------------------------------------------------------------------

-- Generate a page header. This says what section of the site is being
-- viewed, optionally appending some subsection information. Also, the
-- current user is listed, and if that user has a homepage, a link is
-- made to it.
--
-- When helpTopic is non-nil, it is used to determine the help topic to
-- use for the help button on the right hand end of the header. When
-- that parameter is nil, no help is generated.
function generatePageHeader (section, subsection, helpTopic)
    -- Main section
    section = section or "FORGOT TO PROVIDE SECTION HEADER"

    -- Optional subsection
    if subsection == nil then 
        subsection = "" 
    else
        subsection = " : " .. subsection
    end

    -- This outputs the main header information
    html.table:open (nil, "100%", nil, nil, "5", "pageHeader")
    html.table:cell (true, "99%")
    html:addText (section, subsection)

    -- Generate the cell for the help.
    html.table:cell (true, "1%", nil, nil, nil, "help")

    -- If we were given a help topic, then go ahead and generate a help link 
    if helpTopic then
        -- Figure out the URL to use.
        local helpURL = gamedbConfig.baseURL .. "gamedbhelp.lua?topic=" .. helpTopic

        -- The help link uses onClick to activate javascript that will open up a new window
        -- (or use an existing help window from a previous click) and load the help URL in it.
        -- 
        -- For browsers that understand javascript, the HREF field will be ignored entirely.
        -- If the browser does not understand javascript, then it will load the help URL directly,
        -- replacing the current window.
        html:link (helpURL, "Help", nil, nil, nil, getJavascriptWindowCodeForURL (helpURL))
    else
        html:addText ("No help available for this page")
    end

    html.table:close ()
    html:addText ("\nCurrent User: <B>" , userInfo.name , " (" , userInfo.nickname , ") &nbsp; &nbsp;")

    if userInfo.homepage ~= "" then
        html:link (userInfo.homepage , "Jump to your homepage" , "_new")
    end
    
    html:addText ("</B><P>")
end

---------------------------------------------------------------------------

-- Color the output of text. This will return an array that consists of all
-- the strings passed in wrapped in a font tag that will set the given color
-- to the text.
function colorText (color,...)
    return {'<FONT COLOR="' .. color .. '">', unpack (arg), "</FONT>"}
end

---------------------------------------------------------------------------

-- Return the given text formatted to be red, as a table
--
-- Used to display error messages.
function errorText (...)
    return colorText ("Red" , unpack (arg))
end

---------------------------------------------------------------------------

-- Return the given text formatted to be green, as a table
--
-- Used to display information messages that indicate that something was
-- successful.
function successText (...)
    return colorText ("Green" , unpack (arg))
end

---------------------------------------------------------------------------

-- This function collects form arguments together into a table and returns them
-- back to you. 
--
-- The first argument, if not nil, is an array of field names that must be present
-- in the form data. If any of these fields is not present, then the return value of
-- the function is nil. This ensures that all fields required for proper operation
-- all exist.
--
-- The second parameter works similarly, except that the elements contained in it are
-- optional, so if they do not exist in the form data, things are still ok.
--
-- The third paramter describes what should be placed into the return table for
-- any optional variable that doesn't exist. This can be nil, in which case the
-- variable will not be set, or any other default value.
--
-- The return value on success is a table containing the values of the required
-- fields and any optional fields, and nil if any of the required fields are
-- missing.
function collectFormArguments (required, optional, optionalDefaultValue)
    -- Our return value
    local retVal = {}

    -- Is there a required list?
    if required ~= nil then
        -- Yes, iterate over all of them
        for index,var in ipairs (required) do
            -- If there is a form data with this name, then
            -- add it to the retVal. Otherwise, return nil.
            local value = CGI.formData[var]
            if value ~= nil then 
                retVal[var] = value
            else
                return nil
            end
        end
    end

    -- Is there an optional list?
    if optional ~= nil then
        -- Yes, iterate over all of them
        for index,var in ipairs (optional) do
            -- Get the value, possibly replacing it with the default
            -- if it doesn't exist.
            local value = CGI.formData[var] or optionalDefaultValue
            retVal[var] = value
        end
    end

    -- Return
    return retVal
end

---------------------------------------------------------------------------

-- Given an playerID, this returns the player nickname or nil. The playerID
-- 0 is the special ID that represents the system
function nicknameByID (db, id)
    id = tonumber (id) or 0

    -- The special ID 0 indicates the system
    if id == 0 then return "System" end

    -- Try to find the nickname for the player with this ID
    local info = getPlayerByID (db, id)

    -- If it didn't work, return the unknown user, otherwise return the nickname
    if info == nil then return nil else return info.nickname end
end

---------------------------------------------------------------------------

-- Tells you if the date given is valid or not.
-- This just validates that for the year and month given, the day is valid.
function isValidDate (year, month, day)
    -- Is this feb?
    if month == 2 then
        -- Figure out how many days february can have.
        local size = 28

        -- Leap years are years divisible by 4 but not by 100, except for years that
        -- are divisible by 400.
        if ((math.mod (year, 4) == 0 and math.mod (year, 100) ~= 0) or math.mod (year, 400) == 0) then
            size = 29
        end

        -- If it's more than that, bad, otherwise good.
        if day > size then return false else return true end
    end

    -- Is this a month that has 30 days?
    if month == 4 or month == 6 or month == 9 or month == 11 then
        -- If the day is greater than 30, that's bad. Otherwise,
        -- it's ok.
        if day > 30 then return false else return true end
    end

    -- If we get here, all good
    return true
end

---------------------------------------------------------------------------
--- MAIN CODE -------------------------------------------------------------
---------------------------------------------------------------------------

