---------------------------------------------------------------------------
--- FUNCTIONS -------------------------------------------------------------
---------------------------------------------------------------------------

-- This will flush all rows from the session list where the last access is
-- too old. This requires that the configuration information be loaded,
-- since that is where the timeout comes from.
local flushOldSessions =  function (db)
    -- Find out the last valid access time that should be kept. Any time
    -- older than this should get flushed out. The config timeout information
    -- is in minutes, so convert it to seconds. Subtract that from the current
    -- time to get what the last valid time is where the user should not be
    -- timed out. 
    local lastValidTime = os.time () - (configInfo.loginTimeout * 60)

    -- Using the above value, delete all items from the session list where 
    -- the last access date is lower than the calculated value (too old).
    deleteOldSessions (db, lastValidTime)
end

---------------------------------------------------------------------------

-- Removes a current authToken by clobbering it out of the cookie list and
-- removing the cookie.
local removeAuthToken = function ()
    html:removeCookie ("authToken")
    cookies.authToken = nil
end

---------------------------------------------------------------------------

-- Check to see if the authToken given has a session attached, and if so,
-- that it references a user that actually exists, and if it does, that the
-- user it references and the user that the authToken is for match. If all
-- of that is true, then a table of user information is returned back.
-- Otherwise, nil is returned.
local verifyAuthToken = function (db , authToken)
    -- Find a session that matches this authToken.
    local sessionInfo = getSessionForAuthToken (db, authToken)

    -- If there was no session, then this authToken is useless. The user will need to
    -- log in again. This happens if someone changes their password (which changes
    -- the authToken), or if they log out in one browser and try to use the session in
    -- another.
    if sessionInfo == nil then removeAuthToken () return nil end

    -- There was a session. Look up the player information of the user that logged in
    -- to that session.
    local playerInfo = getPlayerByID (db, sessionInfo.playerID)

    -- If there was no such player, we're unhappy.
    if playerInfo == nil then removeAuthToken () return nil end

    -- There was player information. Make an authToken and compare it to this
    -- authToken. If the tokens don't match, that's bad.
    local localToken = getAuthToken (playerInfo.nickname , playerInfo.password)
    if localToken ~= authToken then removeAuthToken () return nil end

    -- The tokens match. That means that the authToken that we got was for
    -- a user that had a valid session (was the same user with the same
    -- password). Update the session to set up the new access time, then
    -- return the player information.
    updateAccessTimeForSession (db, sessionInfo.sessionID, os.time ())
    return playerInfo 
end

---------------------------------------------------------------------------

-- Set up a session for the current authToken, assuming that it has just
-- been logged in properly. This will first delete out any sessions that
-- use this authToken, and then it will introduce a new session.
local setupSession = function (db, authToken, playerID)
    -- Delete any existing sessions for this authToken. This is not really
    -- neccesary because if there already is an entry for this authToken, it must
    -- be for the same user with the same password but in a different browser. But
    -- this is easier than checking for a session and then alternatively updating
    -- or inserting as appopriate.
    deleteSessionForAuthToken (db, authToken)

    -- Now insert a new one
    addSessionListEntry (db , authToken , playerID, os.time ())
end

---------------------------------------------------------------------------

-- This assumes that there is a login form. If the login works, then 
-- set up the userInfo table and return "loggedIn". If there was an
-- error with the login (invalid username or password), then return
-- "error". If the form was empty, then return "needLogin".
local attemptLogin = function (db)
    -- If username or password is empty, then return "needLogin",
    -- so that the user will be prompted to login again and not get an
    -- error. We do this because an empty password is never valid, and
    -- an empty username is just stupid.
    local username = string.trimleadingtrailing (CGI.formData.username or "")
    local password = string.trimleadingtrailing (CGI.formData.password or "")
    if username == "" or password == "" then
        return "needLogin"
    end

    -- Check to see if there is a user with this username and password.
    -- If there isn't, then return "error".
    local playerInfo = getPlayerByNicknamePassword (db, username, password)
    if playerInfo == nil then return "error" end

    -- We have the user. Generate the authToken that this user will use.
    local authToken = getAuthToken (username , password)

    -- Add a cookie for this authToken. We will also pretend that this
    -- cookie existed all along for this run, in case any other code
    -- wants to look it up.
    html:addCookie ("authToken" , authToken)
    cookies.authToken = authToken

    -- Check to see if there is a session associated with this authToken.
    -- If there is, then update it. Otherwise, add it in.
    setupSession (db, authToken , playerInfo.playerID)

    -- All logged in, set up the userInfo table and then return.
    userInfo = fixLoginData (playerInfo)

    addLog (db, "Logged in")
    return "loggedIn"
end

---------------------------------------------------------------------------

-- This will generate a login form for the user to log in with.
function generateLoginForm ()
    html.form:open (env.scriptName, "POST")
    html.table:open (nil, nil, nil, nil, "5", "login")

    html.table:cell (true)
    html:addText ("<B>Username:</B>")
    html.table:cell ()
    html.form:text ("username" , CGI.formData.username, "20" , "60")
    html.table:newRow ()

    html.table:cell (true)
    html:addText ("<B>Password:</B>")
    html.table:cell ()
    html.form:password ("password" , nil, "20", "60")
    html.table:newRow ()

    html.table:cell (true, nil, nil, "CENTER", nil, nil, nil, nil, "2")
    html.form:submit ("command" , "Login", "login")

    html.table:close ()
    html.form:close ()
    html:addText ("<P>")
end

---------------------------------------------------------------------------

-- Checks to see if there is a user logged in or not. In particular:
--
-- If there is a current login form submitted, then check the login
-- information to see if it is valid. If it is, then set an authToken
-- cookie, add a session entry, and return the string "loggedIn".
--
-- If there isn't a login, but there is a valid authToken, and there is
-- a session associated with it, then return the string "loggedIn".
--
-- If there is an authToken but it is not valid because there is no
-- session or it doesn't identify a currently valid user, then
-- return the string "error".
--
-- If there is no login and no authToken, then return the string
-- "needLogin".
--
-- In the case of the return value "loggedIn", a global variable
-- named userInfo will be created, which is a table containing
-- all of the information from the player that is logged in.
function getLoginUser (db)
    -- Before we do anything, flush any old sessions that may exist.
    flushOldSessions (db)

    -- If there is currently a login form submitted, then try to do
    -- a login. If this works, then return the string "loggedIn".
    if CGI.formData.command == "Login" then
        CGI.formData.command = nil
        return attemptLogin (db)
    end

    -- If there is no authToken cookie, then there isn't anybody logged
    -- in, so we need to log in.
    if cookies.authToken == nil then return "needLogin" end

    -- Check the session data to see if there is someone associated with
    -- this authToken. If there is, then get the user information for that
    -- player. Otherwise, return an error.
    -- 
    -- The verifier returns the verified player information on success
    -- and nil on error. If it returns player information, it sets up
    -- a session for it as well.
    local playerInfo = verifyAuthToken (db , cookies.authToken)
    if playerInfo then userInfo = fixLoginData (playerInfo) return "loggedIn" else return "error" end
end

---------------------------------------------------------------------------

-- This is called when we know that the global userInfo that represents the currently
-- logged in user is set up properly and valid. It fixes all of the keys that are
-- boolean values to be true or false based on their textual value, so that they can
-- be used naturally in expressions.
function fixLoginData (oldConfig)
    local newLoginInfo = {}

    -- Loop over all keys that exist in the dictionary
    for key,value in oldConfig do
        -- Does this key start with the regular expression can[A-Z]? If so,
        -- then do further processing.
        if string.find (key, "^can%u") ~= nil then
            -- yes, this is a can key. If the value is a true value, then set the new key to be
            -- true, otherise false.
            if string.upper (value) == "YES" or string.upper (value) == "ON" or string.upper (value) == "TRUE" then
                newLoginInfo[key] = true
            else
                newLoginInfo[key] = false
            end
        else
            -- This isn't a boolean canBlah key, so just copy it's value over.
            newLoginInfo[key] = value
        end
    end

    -- Return the new login info now
    return newLoginInfo
end

---------------------------------------------------------------------------

-- Logs out a logged in user. This destroys the authToken cookie and all
-- entries that reference it in the session list.
function logoutUser (db)
    -- First of all, remove the form command.
    CGI.formData.command = nil

    -- Get the authToken, if there is one.
    local authToken = cookies.authToken

    -- If there wasn't an auth token, then just leave (though this should
    -- never happen, it could if someone tried to log the logout URL).
    if authToken == nil then return end

    -- Delete all entries from the session list that use this authToken.
    deleteSessionForAuthToken (db, authToken)

    -- Remove the cookie, and then pretend it was never set.
    cookies.authToken = nil
    html:removeCookie ("authToken")
end

---------------------------------------------------------------------------
--- MAIN CODE -------------------------------------------------------------
---------------------------------------------------------------------------

