---------------------------------------------------------------------------
--- LOCAL VARS ------------------------------------------------------------
---------------------------------------------------------------------------

-- What section to use in page headers in this file.
local thisPageSection = "PLAYERS"                         

-- Error messages
local ERR_PLAY_NOPLAYERID = [[An error has occured. I don't know what player you're trying to view. Please try again.]]
local ERR_PLAY_NOFINDPLAYER = [[Unable to find the player to view.]]

-- Error messages generated by the profile editor
local ERR_PROF_BADNAME = [[You have not specified a name.]]
local ERR_PROF_BADNICK = [[You have not specified a nickname. This is needed because it is part of your login.]]
local ERR_PROF_BADPASS = [[The password you have entered is not correct.]]
local ERR_PROF_PASSNOMATCH = [[You did not enter your new password the same both times.]]
local ERR_PROF_NOCHANGEPROF = [[You cannot change your profile unless you also enter your current password.]]
local ERR_PROF_NICKINUSE = [[The nickname you have selected is already being used. Please choose another.]]

-- Error messages generated by the add new player section
local ERR_PLAY_NOADDPRIV = [[You are not allowed to add new players to the system. Please check with your administrator.]]
local ERR_PLAY_NONAME = [[You need to specify a name for the player (it does not have to be the player's real name).]]
local ERR_PLAY_NONICK = [[You need to specify a nickname for the player. This will be their username in the system.]]
local ERR_PLAY_BADNICK = [[The nickname you have selected is already in use. Please choose another one.]]
local ERR_PLAY_EMPTYPASS = [[You need to specify a password for the new player.]]
local ERR_PLAY_BADPASS = [[You did not enter the new player password the same both times.]]

-- Error messages generated by the delete player section
local ERR_PLAY_NODELPRIV = [[You are not allowed to delete players from the system. Please check with your administrator.]]
local ERR_DEL_NOUSERS = [[You did not select any players to delete.]]
local ERR_DEL_NODELSELF = [[You are not allowed to delete your own user. Check with your administrator if you wish to have your account removed.]]
local ERR_DEL_NOFINDUSER = [[Unable to find the user to delete. Possibly deleted from another session.]]

-- Error messages generated by the edit properties section
local ERR_PROP_NOPRIV = [[You are not allowed to modify player permissions in the system. Please check with your administrator.]]
local ERR_PROP_NOID = [[An error has occured. I do not know which player you want to modify the properties for. Please try again.]]
local ERR_PROP_BADAUTH = [[Unable to locate the properties for the player requested. Please try again.]]

-- This array names all of the properties for an player, in no particular order.
local playerPropertyFields = {"canEditCategory", "canAddGame" , "canEditGame", "canCreatePlayer", 
                              "canViewLog", "canDeletePlayer", "canEditPlayer" , "canConfigSystem", 
                              "canEditLog" , "canRawDB" , "canEditDesigner" , "canEditPublisher",
                              "canAddSession", "canEditSession"}

---------------------------------------------------------------------------
--- FUNCTIONS -------------------------------------------------------------
---------------------------------------------------------------------------

-- Given a number of wins and a number of losses, display the totals, using
-- proper grammar depending on the numbers.
local displayWinLoss = function (wins, losses)
    -- Get the total number of games played, which is the total number of games
    -- won and the total number of games lost.
    local totalGames = wins + losses

    -- Figure out how many of the games have been won, as a percentage.
    local pct = wins / totalGames * 100

    -- Figure out what label to use for the wins and losses. Starts out as
    -- plurals.
    local label1 = " plays, "
    local label2 = " wins, "
    local label3 = " losses "

    -- If either is one, then change the label to be more grammatically correct.
    if totalGames == 1 then label1 = " play, " end
    if wins == 1 then label2 = " win, " end
    if losses == 1 then label3 = " loss " end

    -- Replace the number 0 with the word no, for readability.
    if wins == 0 then wins = "no" end
    if losses == 0 then losses = "no" end


    -- Output now
    html:addText ("<b>", totalGames, "</b>", label1, "<b>" , wins , "</b>", label2, "<b>", losses , "</b>" , label3, string.format (" (won %2.0f%%)", pct))
end

---------------------------------------------------------------------------

-- This takes some information regarding a win or losss at a game as returned
-- from getLatestWinLossForPlayerID (and so nil means no information available).
-- 
-- If information is provided, then a link to the game won or lost will be generated,
-- as well as a link to the session report for the win/loss of record.
local displayLatestVictoryLoss = function (winlossInfo)
    -- We only want to display things if we have some information.
    if winlossInfo then
        -- Format the session date enclosed to be human readable.
        local sessionDate = os.date ("%A %B %e, %Y @ %l:%M %P", winlossInfo.sessionDate)

        -- Generate a link to the game in the information provided.
        html:link (urls.gameDetails .. winlossInfo.gameID, winlossInfo.name)

        -- Generate a link directly to the session report that this information came from.
        html:addText ("&nbsp; [ ")
        html:link (urls.sessionDetails .. winlossInfo.sessionID, sessionDate)
        html:addText ("&nbsp; ]")
    else
        -- There was no information available.
        html:addText ("None")
    end
end

---------------------------------------------------------------------------

-- Display how many players the system has registered, and how many of
-- them are currently logged in.
local displayPlayerCount = function (db)
    -- Count the number of rows in the player list and the session list
    -- tables. The first tells us how many players are in the system, 
    -- and the second tells us how many of them are logged in right now.
    local totalPlayers = rowsInTable (db , "PlayerList")
    local loggedIn = rowsInTable (db, "SessionList")

    -- If either one failed then return now without doing anything else.
    if totalPlayers == nil or loggedIn == nil then return end

    -- What we output will be slightly different, depending on how many
    -- players are in the system.
    html:addText ('<FONT SIZE="-1">')
    if totalPlayers == 1 then
        html:addText ("You are the only user currently registered with the system.")
    else
        html:addText ("Currently, " , loggedIn , " of " , totalPlayers , " users are logged into the system.")
    end
    html:addText ("</FONT><P>")
end

---------------------------------------------------------------------------

-- Returns true if the playerID given is currently logged in, or false if
-- not (or on error).
--
-- A user is considered logged in if there is currently a row in the SessionList
-- table referencing their playerID. If it's there, they must be logged in. This
-- works because for every page loaded the first thing that happens is all 
-- expired sessions are removed. Also when someone manually logs out, the entry
-- is removed from the session table.
local playerIsLoggedIn = function (db , playerID)
    -- Try to find a session in SessionList for this playerID.
    if getSessionForPlayer (db, playerID) then
        return true
    else
        return false
    end
end

---------------------------------------------------------------------------

-- This generates the form that allows a user to alter their profile
-- information.
local generateProfileForm = function ()
    -- Open a form for editing the profile information
    html.form:open (env.scriptName , "POST", nil, nil)

    -- Open a table for the values
    html.table:open (nil, nil, nil, nil, "5", "playerList")

    -- name
    html.table:heading (true, nil, nil, "RIGHT", "CENTER")
    html:addText ("Player Name")
    html.table:cell (true, nil, nil, "LEFT" , "CENTER")
    html.form:text ("name", userInfo.name, "40" , "60")
    html.table:newRow ()

    -- nickname
    html.table:heading (true, nil, nil, "RIGHT", "CENTER")
    html:addText ("Nickname")
    html.table:cell (true, nil, nil, "LEFT" , "CENTER")
    html.form:text ("nickname", userInfo.nickname, "40" , "60")
    html.table:newRow ()

    -- currentPassword
    html.table:heading (true, nil, nil, "RIGHT", "CENTER")
    html:addText ("Current Password")
    html.table:cell (true, nil, nil, "LEFT" , "CENTER")
    html.form:password ("currentPassword", nil, "40" , "60")
    html.table:newRow ()

    -- newPass1
    html.table:heading (true, nil, nil, "RIGHT", "CENTER")
    html:addText ("New Password")
    html.table:cell (true, nil, nil, "LEFT" , "CENTER")
    html.form:password ("newPassword1", nil, "40" , "60")
    html.table:newRow ()

    -- newPass2
    html.table:heading (true, nil, nil, "RIGHT", "CENTER")
    html:addText ("Repeat New Password")
    html.table:cell (true, nil, nil, "LEFT" , "CENTER")
    html.form:password ("newPassword2", nil, "40" , "60")
    html.table:newRow ()

    -- email
    html.table:heading (true, nil, nil, "RIGHT", "CENTER")
    html:addText ("E-Mail Address")
    html.table:cell (true, nil, nil, "LEFT" , "CENTER")
    html.form:text ("emailAddress", userInfo.emailAddress, "40" , "60")
    html.table:newRow ()

    -- homepage
    html.table:heading (true, nil, nil, "RIGHT", "CENTER")
    html:addText ("Homepage")
    html.table:cell (true, nil, nil, "LEFT" , "CENTER")
    html.form:text ("homepage", userInfo.homepage, "40" , "60")
    html.table:newRow ()

    -- Finish up with some submit action
    html.table:cell (true, nil, nil, "CENTER", "TOP", nil, nil, nil, "2")
    html.form:submit ("Button", "Update your Profile", "player")
    html.form:hidden ("command", "PlayerProfile")
    html.form:hidden ("submitted", "yes")

    -- Close up the table and the form now
    html.table:close ()
    html.form:close ()
end

---------------------------------------------------------------------------

-- Takes a new nickname and password and changes them for the current
-- user. If args.nickname is the same as the current nickname, then
-- the username isn't changing. If the newPassword is nil, then the
-- password isn't changing.
--
-- The return value is nil on success, or the reason for the failure on
-- error (usually because a nickname is already in use or something).
local changePlayerNicknamePassword = function (db, newNickname, newPassword)
    -- Set up what the new password field is going to be.
    newPassword = newPassword or userInfo.password

    -- Check to see if the nickname given is the nickname for the current
    -- user or not.
    if newNickname ~= userInfo.nickname then
        -- The nickname is changing. In that case, try to find a player
        -- that is using the new nickname. If we find something, then
        -- this nickname is already in use (and thus invalid)
        if getPlayerByName (db, newNickname) then
            return ERR_PROF_NICKINUSE
        end
    end

    -- Now we know the nickname is valid. Create a new authToken based on
    -- the new nickname and password (either of which may be the original,
    -- untouched version). Also take a copy of the old auth token.
    local newAuthToken = getAuthToken (newNickname , newPassword)
    local oldAuthToken = cookies.authToken

    -- If both authTokens are identical, then the username and password given are
    -- identical. In that case, nothing needs to change. The authToken, session
    -- information and player list already contain the correct data. So, just 
    -- return nil in this case to save time.
    if newAuthToken == oldAuthToken then return nil end

    -- Set the authToken cookie to use the new authToken, and pretend that's
    -- always how it's been.
    cookies.authToken = newAuthToken
    html:addCookie ("authToken" , newAuthToken)

    -- Now, do an update to change the username/password field of the given
    -- player.
    changeNickPassForPlayerID (db, userInfo.playerID, newNickname, newPassword)

    -- Lastly, for the current login to remain logged in as a result of the
    -- authToken change, we need to update the session information, so that
    -- the correct auth token is associated with this player.
    switchSessionAuthTokens (db, oldAuthToken, newAuthToken)

    -- Change the current user information so that the rest of the page has
    -- the new values.
    userInfo.nickname = newNickname
    userInfo.password = newPassword

    -- All done successfully.
    return nil
end

---------------------------------------------------------------------------

-- This handles the submission of the form that allows players to change
-- their own profile information.
-- 
-- If the fields entered in the form are valid, then the current user settings
-- and the database are both updated and nil is returned. nil is also returned
-- if the form hasn't been submitted.
--
-- If some forms are not correct, then the return value is a string that explains
-- the error,
local handleProfileForm = function (db)
    -- Get the form arguments that we'll need.
    local args = collectFormArguments ({"name" , "nickname", "emailAddress", "homepage", "submitted", "currentPassword" , "newPassword1", "newPassword2"})
    local newPassword = nil

    -- If the form submission is incomplete because one of the form arguments is
    -- missing, then just return nil now. No errors, because this means someone is
    -- manually entering a URL or just came to the page and hasn't submitted anything
    -- yet. We don't care either way.
    if args == nil then return nil end

    -- Fix up the arguments a bit.
    args.name            = string.trimleadingtrailing (args.name            or "")
    args.nickname        = string.trimleadingtrailing (args.nickname        or "") 
    args.currentPassword = string.trimleadingtrailing (args.currentPassword or "") 
    args.newPassword1    = string.trimleadingtrailing (args.newPassword1    or "") 
    args.newPassword2    = string.trimleadingtrailing (args.newPassword2    or "") 
    args.emailAddress    = string.trimleadingtrailing (args.emailAddress    or "") 
    args.homepage        = string.trimleadingtrailing (args.homepage        or "") 

    -- The user's current password must be provided, or no profile information can be
    -- edited. 
    if args.currentPassword == "" then return errorText (ERR_PROF_NOCHANGEPROF) end

    -- The password entered has to be an actual valid password for things to change.
    if args.currentPassword ~= userInfo.password then return errorText (ERR_PROF_BADPASS) end

    -- The name and nickname fields are required, so leave if either of them is empty.
    if args.name == "" then return errorText (ERR_PROF_BADNAME) end
    if args.nickname == "" then return errorText (ERR_PROF_BADNICK) end

    -- If either of the two new password fields has a value in it, then someone is trying to
    -- change the password.
    if args.newPassword1 ~= "" or args.newPassword2 ~= "" then 
        -- Both of the new passwords must match for us to want to change things.
        if args.newPassword1 ~= args.newPassword2 then return errorText (ERR_PROF_PASSNOMATCH) end

        -- In here we know that a) one of the two passwords given is not blank, and b) both
        -- passwords are identical. That means that both of them have to be non-blank. Since they
        -- are non-blank and identical, then this is a valid new password, so set that up to
        -- change later on, along with possibly other information.
        newPassword = args.newPassword1
    end

    -- Start by changing the username and password. Note that here one or none of nickname and password
    -- may be changing, but even if they are both identical, that is ok. We know that the password given
    -- above is valid, so it is safe to make this change.
    --
    -- The information in the database and the login credentials for the user will be changed,
    -- so that the system won't throw them out on the next page load for having the wrong
    -- authToken cookie.
    --
    -- Note also that here newPassword may be nil if the password isn't changing. That is ok.
    --
    -- This function will return nil on success, or an error message that can be used for error text
    -- on failure (usually because a new nickname is identical to an existing nickname).
    local result = changePlayerNicknamePassword (db, args.nickname, newPassword)
    if result then return errorText (result) end

    -- Save the changes to name, email address and homepage. 
    changePlayerProfileForPlayerID (db, userInfo.playerID, args.name, args.emailAddress, args.homepage)

    -- We need to put the changes made just now into the userInfo so that they propogate for the rest
    -- of the page that may be generated after this form submission. Note that possibly the nickname
    -- and password changed. If that's the case, then changePlayerNicknamePassword has already updated userInfo
    -- with those changes for us.
    userInfo.name         = args.name
    userInfo.emailAddress = args.emailAddress
    userInfo.homepage     = args.homepage

    addLog (db , "Altered profile information")

    -- Now change any other fields that have changed.
    return successText ("Your changes have been saved")
end

---------------------------------------------------------------------------

-- This generates the form for adding a new player to the system. Args can
-- be nil, or it can be be the results of previous form submission, in which
-- case the values it stores will be used to fill out the form again
-- (except for password fields).
local generateNewPlayerForm = function (args)
    -- Make args be an empty table if none are specified.
    args = args or {}

    -- Open a form for editing the profile information
    html.form:open (env.scriptName , "POST", nil, nil)

    -- Open a table for the values
    html.table:open (nil, nil, nil, nil, "5", "playerList")

    -- name
    html.table:heading (true, nil, nil, "RIGHT", "CENTER")
    html:addText ("Player Name")
    html.table:cell (true, nil, nil, "LEFT" , "CENTER")
    html.form:text ("name", args.name, "40" , "60")
    html.table:newRow ()

    -- nickname
    html.table:heading (true, nil, nil, "RIGHT", "CENTER")
    html:addText ("Nickname")
    html.table:cell (true, nil, nil, "LEFT" , "CENTER")
    html.form:text ("nickname", args.nickname, "40" , "60")
    html.table:newRow ()

    -- newPass1
    html.table:heading (true, nil, nil, "RIGHT", "CENTER")
    html:addText ("New Password")
    html.table:cell (true, nil, nil, "LEFT" , "CENTER")
    html.form:password ("newPassword1", nil, "40" , "60")
    html.table:newRow ()

    -- newPass2
    html.table:heading (true, nil, nil, "RIGHT", "CENTER")
    html:addText ("Repeat New Password")
    html.table:cell (true, nil, nil, "LEFT" , "CENTER")
    html.form:password ("newPassword2", nil, "40" , "60")
    html.table:newRow ()

    -- email
    html.table:heading (true, nil, nil, "RIGHT", "CENTER")
    html:addText ("E-Mail Address")
    html.table:cell (true, nil, nil, "LEFT" , "CENTER")
    html.form:text ("emailAddress", args.emailAddress, "40" , "60")
    html.table:newRow ()

    -- homepage
    html.table:heading (true, nil, nil, "RIGHT", "CENTER")
    html:addText ("Homepage")
    html.table:cell (true, nil, nil, "LEFT" , "CENTER")
    html.form:text ("homepage", args.homepage, "40" , "60")
    html.table:newRow ()

    -- Finish up with some submit action
    html.table:cell (true, nil, nil, "CENTER", "TOP", nil, nil, nil, "2")
    html.form:submit ("Button", "Add New Player", "player")
    html.form:hidden ("command", "PlayerNew")
    html.form:hidden ("submitted", "yes")

    -- Close up the table and the form now
    html.table:close ()
    html.form:close ()
end

---------------------------------------------------------------------------

-- This tables the results of a form submission for the addition of a new
-- player to the system, and checks to see if everything is ok or not.
-- 
-- On success, the user is added and nil is returned.
--
-- On error, error text is returned back.
local addPlayer = function(db, formArgs)
    -- Fix up the arguments a bit.
    formArgs.name         = string.trimleadingtrailing (formArgs.name         or "")
    formArgs.nickname     = string.trimleadingtrailing (formArgs.nickname     or "") 
    formArgs.newPassword1 = string.trimleadingtrailing (formArgs.newPassword1 or "") 
    formArgs.newPassword2 = string.trimleadingtrailing (formArgs.newPassword2 or "") 
    formArgs.emailAddress = string.trimleadingtrailing (formArgs.emailAddress or "") 
    formArgs.homepage     = string.trimleadingtrailing (formArgs.homepage     or "") 

    -- If the name of the user is blank, then that's bad.
    if formArgs.name == "" then return errorText (ERR_PLAY_NONAME) end

    -- If the nickname is blank, that's also bad. The nickname is the login
    -- id of this user.
    if formArgs.nickname == "" then return errorText (ERR_PLAY_NONICK) end

    -- If there is a user in the system that is using this nickname already, then
    -- that's bad. 
    if getPlayerByName (db, formArgs.nickname) then return errorText (ERR_PLAY_BADNICK) end

    -- Make sure that at least one of the two password fields has been filled out properly.
    if formArgs.newPassword1 == "" and formArgs.newPassword2 == "" then return errorText (ERR_PLAY_EMPTYPASS) end

    -- The user can't be added if both of the passwords given are different from each
    -- other.
    if formArgs.newPassword1 ~= formArgs.newPassword2 then return errorText (ERR_PLAY_BADPASS) end

    -- All of the other fields are not vital, so it's ok if they're blank.
    -- Go ahead and add the user now.
    addNewPlayer (db, formArgs.name, formArgs.nickname, formArgs.newPassword1, formArgs.emailAddress, formArgs.homepage , userInfo.playerID)

    -- Worked, so log it and return nil to signify that
    addLog (db , "Added player " .. formArgs.nickname)
    return nil
end

---------------------------------------------------------------------------

-- This is a helper for the code that generates the player property form.
-- We take a textual description of the property, it's database field in
-- the player list, and some player information. 
--
-- We will generate a row in the table which lists the description, 
-- followed by a checkbox which is either checked or not based on the
-- information we've been given.
--
-- The checkbox will have a name identical to the propertyName, to make
-- form submissions easy as well.
local checkboxForUserProperty = function (description, propertyName, infoTable)
    -- Figure out if this property should be checked or not.
    local isChecked = false
    if infoTable[propertyName] then isChecked = true end

    -- Open a cell for the description of the property.
    html.table:heading (true, nil, nil, "RIGHT", "CENTER")
    html:addText (description)

    -- Now open a cell for the checkbox, and check it as needed.
    html.table:cell (true, nil, nil, "LEFT" , "CENTER")
    html.form:checkbox (propertyName, "YES", isChecked)

    -- Start a new row before we return.
    html.table:newRow ()
end

---------------------------------------------------------------------------

-- This will generate a property form for the player with the provided playerID.
-- The checkboxes for the properties are either checked or not based on the current
-- properties for that user in the database.
local generatePlayerPropertyForm = function (db, playerID)
    -- Get the player information we're generating for. If this doesn't work, then
    -- generate an error message and leave now.
    local playerInfo = getPlayerByID (db, playerID)
    if playerInfo == nil then
        html:addText (errorText (ERR_PROP_BADAUTH) , "<P>")
        return
    end

    -- Every time a page is loaded, the setup code will get the player information for the
    -- currently logged in user, and then "fix" it so that the values in the boolean fields
    -- are true booleans.
    --
    -- Here, we do the same thing with the just loaded data, which makes the other code
    -- easier to write, since there are no comparisons to strings to be done.
    playerInfo = fixLoginData (playerInfo)

    -- Open a form for editing the profile information
    html.form:open (env.scriptName , "POST", nil, nil)

    -- Open a table for the values
    html.table:open (nil, nil, nil, nil, "5", "playerList")
    html.table:caption ("Properties for : <B>" .. playerInfo.name .. "</B>", "CENTER")

    -- Use the helper function to generate a checkbox for each player property.
    checkboxForUserProperty ("Add new players"                  , "canCreatePlayer"  , playerInfo)
    checkboxForUserProperty ("Delete existing players"          , "canDeletePlayer"  , playerInfo)
    checkboxForUserProperty ("Edit player properties"           , "canEditPlayer"    , playerInfo)
    checkboxForUserProperty ("Add/Edit game Designers"          , "canEditDesigner"  , playerInfo)
    checkboxForUserProperty ("Add/Edit game Publishers"         , "canEditPublisher" , playerInfo)
    checkboxForUserProperty ("Add/Edit game Categories"         , "canEditCategory"  , playerInfo)
    checkboxForUserProperty ("Add new games"                    , "canAddGame"       , playerInfo)
    checkboxForUserProperty ("Edit existing games"              , "canEditGame"      , playerInfo)
    checkboxForUserProperty ("Add new Session Reports"          , "canAddSession"    , playerInfo)
    checkboxForUserProperty ("Edit existing Session Reports"    , "canEditSession"   , playerInfo)
    checkboxForUserProperty ("Configure the system"             , "canConfigSystem"  , playerInfo)
    checkboxForUserProperty ("View the event log"               , "canViewLog"       , playerInfo)
    checkboxForUserProperty ("Empty the event log"              , "canEditLog"       , playerInfo)
    checkboxForUserProperty ("Raw Database access (dangerous!)" , "canRawDB"         , playerInfo)
    
    -- Finish the form with a submit button
    html.table:cell (true, nil, nil, "CENTER" , "CENTER", nil, nil, nil, "2")
    html.form:submit ("Button", "Change Player Properties", "player")

    -- Close the table
    html.table:close ()

    -- Extra form fields to carry over who the form is being submitted for, so that the form
    -- comes back to us and keeps track of what the playerID was. 
    html.form:hidden ("submitted" , "yes")
    html.form:hidden ("command" , "PlayerProperties")
    html.form:hidden ("playerID" , playerID)

    -- Close the form now
    html.form:close ()
end

---------------------------------------------------------------------------

-- Handles a form submission for player properties on the given player.
-- This only needs a playerID to work. All of the other fields are optional,
-- since they're checkboxes; if they're not checked, they're not submitted
-- with the rest of the form.
local handlePropertiesSubmission = function (db, playerID, args)
    -- The variable playerPropertyFields contains a list of property
    -- fields that belong both in the database and also in the form.
    -- Loop over each one. 
    for key,value in pairs (playerPropertyFields) do
        -- Default to the property not being present
        local newValue = "NO"

        -- If the form arguments supplies the property field, then it
        -- was checked, so change the key to yes.
        if args[value] then newValue = "YES" end

        -- Store the key now
        args[value] = newValue
    end

    -- Now, do DB stuff to store the properties for the user with the given
    -- playerID.
    changePlayerPropertiesForPlayerID (db, playerID, 
                args.canEditCategory, args.canAddGame, args.canEditGame, args.canCreatePlayer,
                args.canViewLog, args.canDeletePlayer, args.canEditPlayer, args.canConfigSystem, 
                args.canEditLog, args.canRawDB, args.canEditDesigner , args.canEditPublisher, args.canAddSession, args.canEditSession)

    -- Log what we did.
    addLog (db , "Altered profile information for user " .. nicknameByID (db, playerID))
end

---------------------------------------------------------------------------

-- Handles editing of player properties
function commandPlayerProperties (db)
    -- Header for the page
    generatePageHeader (thisPageSection, "Edit Player Properties", "playerProperties")

    -- Generate a button to go back to the main player page
    html.form:singleButtonForm ("Return to Players Page" , "Player" , env.scriptName , "POST" , nil, nil, "player")
    html:addText ("<P>")

    -- If this user isn't allowed to edit player properties, tell them that and then
    -- leave. This happens first so that even if the user tries to craft a URL
    -- that will add a player, it will still stop them.
    if userInfo.canEditPlayer == false then
        addLog (db , "Attempted to modify properties of a player")
        html:addText (errorText (ERR_PROP_NOPRIV) , "<P>")
        return
    end

    -- Check to make sure that we have a player ID
    local args = collectFormArguments ({"playerID"})

    -- If we don't, generate some bogus text and leave.
    if args == nil then
        html:addText (errorText (ERR_PROP_NOID) , "<P>")
        return
    end

    -- Get the player ID
    local playerID = args.playerID

    -- Get other form arguments. Note that the only required parts here are the submitted part.
    -- The rest are optional (if they are not present, they were not checked in the form
    -- submission).
    args = collectFormArguments ({"submitted"} , playerPropertyFields)

    -- If we got any arguments, then at least we know the form was submitted. So handle the submission.
    if args ~= nil then 
        handlePropertiesSubmission (db, playerID, args)
        html:addText (successText ("Saved your property changes") , "<P>")
    end

    -- Generate a property form for this player, which may use the new
    -- properties. Note here that if we are editing properties for ourselves, we 
    -- may have just removed our own ability to edit further properties. We
    -- didn't update the player information. Since we are just going to generate
    -- this form and stop, that's ok.
    generatePlayerPropertyForm (db, playerID)
end

---------------------------------------------------------------------------

-- This handles form submissions from the main player page that allow for
-- the deletion of players from the system.
function commandPlayerDelete (db)
    -- Header for the page
    generatePageHeader (thisPageSection, "Delete Players")

    -- Generate a button to go back to the main player page
    html.form:singleButtonForm ("Return to Players Page" , "Player" , env.scriptName , "POST" , nil, nil, "player")

    -- If this user isn't allowed to delete a player, tell them that and then
    -- leave. This happens first so that even if the user tries to craft a URL
    -- that will delete a player, it will still stop them.
    if userInfo.canDeletePlayer == false then
        addLog (db , "Attempted to delete a player")
        html:addText (errorText (ERR_PLAY_NODELPRIV) , "<P>")
        return
    end

    -- Collect the delete form arguments. The result could be nil if there
    -- were no players checked, a single string if only one player is being
    -- deleted, or a table for many players. The only thing that we need
    -- to worry about here is nil. Complain if someone submits the form without
    -- saying who to delete.    
    local formArgs = collectFormArguments ({"delPlayerID"})
    if formArgs == nil then 
        addLog (db, "Attempted delete with no players selected");
        html:addText (errorText (ERR_DEL_NOUSERS) , "<P>");
        return
    end

    -- Make sure that the list of players is a table and not a string.
    if type (formArgs.delPlayerID) ~= "table" then
        formArgs.delPlayerID = {formArgs.delPlayerID}
    end

    -- We got confirmation on deletion prior to getting here, when the browser asked if
    -- they were sure. So, loop over all users and remove them.
    for index,auth in formArgs.delPlayerID do
        -- Convert the playerID into a number. If it's the same as the currently logged in
        -- user, then display an error instead of deleting. Players are not allowed to delete
        -- themselves.
        if tonumber (auth) == userInfo.playerID then
            -- Can't delete yourself.
            html:addText (errorText (ERR_DEL_NODELSELF) , "<BR>");
        else
            -- Get the info on the player to delete
            local delPlayerInfo = getPlayerByID (db, auth)

            -- if we can't find the information for this user, complain.
            if delPlayerInfo == nil then
                html:addText (errorText (ERR_DEL_NOFINDUSER) , "<BR>");
            else
                -- We're deleting. To start with, remove all entries for this player
                -- from SessionList (the table that holds login sessions) and PlayerList.
                deleteSessionForPlayerID (db, auth)
                deletePlayerByID (db, auth)

                -- Now, go through the system log and replace all instances of this userid
                -- with the current nickname.
                renameLogEntryWhois (db, auth, delPlayerInfo.nickname)

                -- Do the same in the SessionReportEntry table, but replace with the name instead of the
                -- nickname.
                renameSessionReportEntryWhois (db, auth, delPlayerInfo.name)

                -- We're done.
                html:addText (successText ("Deleted player '" .. delPlayerInfo.name .. "'") , "<BR>");
                addLog (db , "Deleted player '" .. delPlayerInfo.name .. "'")
            end
        end
    end
end

---------------------------------------------------------------------------

-- This handles the addition of a new player into the system by providing a
-- form for new player information.
function commandPlayerNew (db)
    -- Header for the page
    generatePageHeader (thisPageSection, "Add a Player", "playerAdd")

    -- Generate a button to go back to the main player page
    html.form:singleButtonForm ("Return to Players Page" , "Player" , env.scriptName , "POST" , nil, nil, "player")
    html:addText ("<P>")

    -- Stop now if this user is not allowed to add new players. This should only
    -- ever trigger if someone manually enters a url, since the system will not
    -- display the button to get to here if the user doesn't already have the
    -- ability to add players.
    if userInfo.canCreatePlayer == false then
        addLog (db , "Attempted to add a new player")
        html:addText (errorText (ERR_PLAY_NOADDPRIV) , "<P>")
        return
    end

    -- Collect form arguments to see if the form has been submitted or not.
    local formArgs = collectFormArguments ({"name" , "nickname" , "newPassword1" , "newPassword2" , "emailAddress" , "homepage", "submitted"})

    -- If we got arguments, then attempt to add the new player to the system.
    if formArgs then
        -- Try to add the player to the system using these form arguments.
        -- On success, this returns nil. On error, it returns the reason
        -- why the form could not be submitted.
        local result = addPlayer (db, formArgs)

        -- Check to see if the addition worked.
        if result == nil then
            -- It works. Remove the form arguments that we have so that the form will
            -- be generated empty, ready for another player.
            html:addText (successText ("The new player has been added to the system.") , "<P>")
            formArgs = nil
        else
            -- The addition failed, so explain why. We leave form arguments alone in
            -- this case.
            html:addText (result , "<P>")
        end
    end

    -- Generate a player form. If a form has been submitted but there was
    -- something wrong, then formArgs will keep the previously entered
    -- values (except for the password), so that the form will be filled out
    -- again with similar values and the user only has to fix the problem.
    --
    -- If a player was successfully added or no form has ever been submitted,
    -- the form will be generated empty, allowing a new addition.
    generateNewPlayerForm (formArgs)
end

---------------------------------------------------------------------------

-- This function handles the editing of this player's profile information.
function commandPlayerProfile (db)
    -- Try to handle a property form submission first. The return value is
    -- some text that says what happened with the submission (success or
    -- failure), or nil if nothing happened (no submission).
    --
    -- We need to do this before anything else, since the page header contains
    -- profile information, and if the profile has changed, we want to make
    -- sure that gets reflected.
    --
    -- If the form is submitted, then the database is updated and the currently
    -- loaded copy of the user information is also updated.
    local result = handleProfileForm (db)

    -- Header for the page
    generatePageHeader (thisPageSection, "Edit your Profile", "playerProfile")

    -- Generate a button to go back to the main player page
    html.form:singleButtonForm ("Return to Players Page" , "Player" , env.scriptName , "POST" , nil, nil, "player")
    html:addText ("<P>")

    -- If the form was submitted, then add any text it spat out onto the
    -- page now.
    if result then html:addText (result , "<P>") end

    -- Lastly, generate a profile form. This uses the information on the
    -- current user to fill out the fields.
    generateProfileForm ()
end

---------------------------------------------------------------------------

-- This handles the viewing of the specific profile information of any
-- particular player, allowing other players to view/interact with other
-- players in the system.
function commandPlayerView (db)
    -- Header for the page
    generatePageHeader (thisPageSection, "View Player Profile", "playerView")

    -- Generate a button to go back to the main player page
    html.form:singleButtonForm ("Return to Players Page" , "Player" , env.scriptName , "POST" , nil, nil, "player")
    html:addText ("<P>")

    -- Check to find the playerID of the player that we're supposed to view.
    local args = collectFormArguments ({"playerID"})
    if args == nil then html:addText (errorText (ERR_PLAY_NOPLAYERID), "<P>") return end

    -- Get the information for the player
    local playerInfo = getPlayerByID (db, args.playerID)
    if playerInfo == nil then html:addText (errorText (ERR_PLAY_NOFINDPLAYER), "<P>") return end

    -- Display the profile information for this player
    html.table:open (nil, nil, nil, nil, "5", "playerList")

    -- Display the name of this player.
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Name:")
    html.table:cell (true, nil, nil, "LEFT", "TOP")
    html:addText (playerInfo.name)

    -- Display the login name
    html.table:newRow ()
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Login:")
    html.table:cell (true, nil, nil, "LEFT", "TOP")
    html:addText (playerInfo.nickname)

    -- If the user has a homepage listed, then generate a link to it.
    if playerInfo.homepage ~= "" then
        html.table:newRow ()
        html.table:heading (true, nil, nil, "LEFT", "TOP")
        html:addText ("Homepage:")
        html.table:cell (true, nil, nil, "LEFT", "TOP")
        html:link (playerInfo.homepage , playerInfo.homepage , "_new")
    end

    -- If the user has an email address listed, then generate a link to
    -- that as well.
    if playerInfo.emailAddress ~= "" then
        html.table:newRow ()
        html.table:heading (true, nil, nil, "LEFT", "TOP")
        html:addText ("Email:")
        html.table:cell (true, nil, nil, "LEFT", "TOP")
        html:link ("mailto:" .. playerInfo.emailAddress , playerInfo.emailAddress)
    end

    -- New row containing a single cell with a nonbreaking space, to act as a spacer.
    html.table:newRow ()
    html.table:cell (true, nil, nil, "LEFT", "TOP", nil, nil, nil, "2")
    html:addText ("&nbsp;")

    -- Get the information for the last game that this player won, and the last game that this player
    -- lost. If either of these is nil, then that means "None".
    -- 
    -- The result is similar to what we get from gamesPlayedInformationForPlayerID below, except that
    -- it's a single entry or nil, so it doesn't contain the number of times that game has been won
    -- or lost.
    local lastWin = getLatestWinLossForPlayerID (db, true, playerInfo.playerID) 
    local lastLoss = getLatestWinLossForPlayerID (db, false, playerInfo.playerID)

    -- Say what game was Won most recently.
    html.table:newRow ()
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Most Recent Win:")
    html.table:cell (true, nil, nil, "LEFT", "TOP")
    displayLatestVictoryLoss (lastWin)

    -- Say what game was lost most recently.
    html.table:newRow ()
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Most Recent Loss:")
    html.table:cell (true, nil, nil, "LEFT", "TOP")
    displayLatestVictoryLoss (lastLoss)

    -- Get the list of games this player has played, sorted according to name.
    local playList = gamesPlayedInformationForPlayerID (db, playerInfo.playerID)
    local totalWins = 0
    local totalLosses = 0

    -- If the list has any entries in it, then list them all, including some
    -- final output totals. If not, then don't display anything. This way, for
    -- players who have played no games, there is not a bunch of extra stuff
    -- in the table.
    if table.getn (playList) ~= 0 then
        -- New row containing a single cell with a nonbreaking space, to act as a spacer.
        html.table:newRow ()
        html.table:cell (true, nil, nil, "LEFT", "TOP", nil, nil, nil, "2")
        html:addText ("&nbsp;")

        -- Now headings for the game list to follow
        html.table:newRow ()
        html.table:heading (true, nil, nil, "LEFT", "TOP")
        html:addText ("Game")
        html.table:heading (true, nil, nil, "LEFT", "TOP")
        html:addText ("Won/Lost")

        -- Iterate over each entry in the win list and spit some information out.   
        for index, playInfo in playList do
            -- Start a new row
            html.table:newRow ()

            -- Game name
            html.table:cell (true, nil, nil, "LEFT", "TOP")
            html:addText (playInfo.name)

            -- Times won
            html.table:cell (true, nil, nil, "LEFT", "TOP")
            displayWinLoss (playInfo.timesWon, playInfo.timesLost)

            -- Add in these wins and losses.
            totalWins = totalWins + playInfo.timesWon
            totalLosses = totalLosses + playInfo.timesLost
        end
     
        -- New row containing a single cell with a nonbreaking space, to act as a spacer.
        html.table:newRow ()
        html.table:cell (true, nil, nil, "LEFT", "TOP", nil, nil, nil, "2")
        html:addText ("&nbsp;")

        -- Finally, totals.
        html.table:newRow ()
        html.table:heading (true, nil, nil, "LEFT", "TOP")
        html:addText ("Totals:")
        html.table:cell (true, nil, nil, "LEFT", "TOP")
        displayWinLoss (totalWins, totalLosses)
    end

    -- Done now
    html.table:close ()
end

---------------------------------------------------------------------------

-- This handles the base "Player" command that you get when you click the
-- player menu button. This is just a gateway to other player sections. 
function commandPlayer (db)
    -- Header for the page
    generatePageHeader (thisPageSection, nil, "player")

    -- Generate a single button form to take the user to the section that allows the editing
    -- of their profile information.
    html.form:singleButtonForm ("Edit your Player profile" , "PlayerProfile" , env.scriptName , "POST" , nil, nil, "player")
    html:addText ("<P>")

    -- If this user is allowed to create players, then output a button to allow that to happen as well.
    if userInfo.canCreatePlayer then
        html.form:singleButtonForm ("Add a Player" , "PlayerNew" , env.scriptName , "POST" , nil, nil, "player")
        html:addText ("<P>")
    end

    -- Display the number of users in the system and logged in.
    displayPlayerCount (db)

    -- If the user has javascript enabled, this confirmation will be used to get them to
    -- confirm what they are about to do.
    local deleteConfirm = "return confirm ('Any selected players will be completely removed from the system. Are you sure this is what you want to do?');"

    -- Now display the list of players
    html.form:open (env.scriptName , "POST", nil, nil, deleteConfirm)

    -- Display the playersa in a table
    html.table:open (nil, nil, nil, nil, "5", "playerList")

    -- Track if this user can delete or edit properties of players.
    local canDelete = userInfo.canDeletePlayer
    local canEdit   = userInfo.canEditPlayer

    -- The table we use generally have 4 columns:
    -- Name, Properties, Added By, and Delete. The properties
    -- cell is only included if this user can edit people. The
    -- Delete column is only included if this user can delete people.
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Player Name")

    -- A cell to hold how many games this player has played, lost and won
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Played")
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Lost")
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Won")

    -- Can the user edit properties? If so, include this column
    if canEdit then
        html.table:heading (true, nil, nil, "LEFT", "TOP")
        html:addText ("&nbsp;")
    end
    
    -- Column for who added this user.
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Added by")

    -- Can the user delete? If so, use this column.
    if canDelete then
        html.table:heading (true, nil, nil, "LEFT", "TOP")
        html:addText ("Delete")
    end

    -- Start the first actual row.
    html.table:newRow ()

    -- Get the complete list of players in the system, sorted by name.
    local playerList = getPlayerList (db, true)
    for unusedIndex, row in playerList do
        -- Include the name of this player
        html.table:cell (true, nil, nil, "LEFT" , "TOP")
        html:link (urls.playerProfile .. row.playerID, row.name)
        html:addText (" (" , row.nickname , ")")

        -- If this person is currently logged in, then add extra text to
        -- reflect that.
        if playerIsLoggedIn (db, row.playerID) then
            html:addText ('&nbsp; <FONT SIZE="-2">(logged in)</FONT>')
        end

        -- Get the count of games played, won and lost, and the win percentage.
        local playCount = tonumber (playCountForPlayerID (db, row.playerID)) or 0
        local winCount = tonumber (winCountForPlayerID (db, row.playerID)) or 0
        local lostCount = playCount - winCount
        local winPct = ""
        if playCount ~= 0 then 
            winPct = string.format (" (%2.0f%%)", (winCount / playCount) * 100)
        end

        -- Now put each item in
        html.table:cell (true, nil, nil, "LEFT" , "TOP")
        html:addText (playCount)
        html.table:cell (true, nil, nil, "LEFT" , "TOP")
        if lostCount == 0 then
            html:addText (lostCount)
        else
            html:link (urls.sessionsByPlayerLoss .. row.playerID, lostCount)
        end
        html.table:cell (true, nil, nil, "LEFT" , "TOP")
        if winCount == 0 then
            html:addText (winCount)
        else
            html:link (urls.sessionsByPlayerWin .. row.playerID, winCount)
        end
        html:addText (winPct)

        -- If the user can edit properties, then include a
        -- propperties link as well.
        if canEdit then
            html.table:cell (true, nil, nil, "LEFT" , "TOP")
            html:link (urls.playerProperties .. row.playerID , "Properties")
        end

        -- Say who added this person.
        html.table:cell (true, nil, nil, "LEFT" , "TOP")
        html:addText (nicknameByID (db , row.createdBy) or row.createdBy)

        -- If the user can delete, then add the delete column.
        if canDelete then
            html.table:cell (true, nil, nil, "LEFT" , "TOP")

            -- If this player isn't the current playerID, then add a checkbox. Otherwise,
            -- just insert a nonbreaking space. We don't want to allow the player to delete
            -- him/herself.
            if row.playerID == userInfo.playerID then 
                html:addText ("&nbsp;")
            else
                html.form:checkbox ("delPlayerID" , row.playerID)
            end
        end

        -- Get ready for the next row
        html.table:newRow ()
    end

    -- Now we need to know, should we add a submit button for the delete form? 
    -- If this user can't delete, we don't need to. If the user can delete, but
    -- the player list size is only 1, then they also can't delete (because people
    -- can't delete themselves).
    --
    -- If we can delete and the number of players is 2 or more, then include the
    -- submit button.
    if table.getn (playerList) >= 2 and canDelete then
        -- The table has at least 6 columns in it. It will be 6 if this user
        -- can edit properties, 7 otherwise.
        local span = "6";
        if canEdit then span = "7" end

        -- Create a cell to span the whole table and put a submit button in it.
        html.table:cell (true, nil, nil, "CENTER", "TOP", nil, nil, nil, span)
        html.form:submit ("Button", "Delete selected players", "player")
        html.form:hidden ("command", "PlayerDelete")
    end

    -- Close up the table and the form now
    html.table:close ()
    html.form:close ()
end

---------------------------------------------------------------------------
--- MAIN CODE -------------------------------------------------------------
---------------------------------------------------------------------------

