---------------------------------------------------------------------------
--- LOCAL VARS ------------------------------------------------------------
---------------------------------------------------------------------------

-- What section to use in page headers in this file.
local thisPageSection = "PUBLISHERS"                         

-- Error Messages
local ERR_AUTH_NOADDPRIV = [[You are not allowed to add new publishers to the system. Please check with your administrator.]]
local ERR_AUTH_NODELPRIV = [[You are not allowed to delete publishers from the system. Please check with your administrator.]]
local ERR_AUTH_NOEDITPRIV = [[You are not allowed to edit publishers in the system. Please check with your administrator.]]
local ERR_PUB_DUPPUBLISH = [[A publisher by this name already exists.]]
local ERR_DEL_NOPUBLISHERS = [[You did not select any publishers to delete.]]
local ERR_DEL_NOFINDPUBLISH = [[Unable to find the publishers to delete. It may have been removed in another session.]]
local ERR_PUB_NOPUBID = [[An error has occured. I don't know what publisher you're trying to modify. Please try again.]]
local ERR_PUB_BADPROPNAME = [[The publisher name you have specified is invalid. Please try again.]]
local SUCC_PUB_NOCHANGE = [[No changes have been made.]]

-- The number of publisher fields to include on the add new publisher page.
local ADD_PUBLISHER_FIELD_COUNT = 5

---------------------------------------------------------------------------
--- FUNCTIONS -------------------------------------------------------------
---------------------------------------------------------------------------

-- Display information on the number of publishers that are currently entered
-- in the city. This discounts the entry in the table with the ID of 0, which
-- is the "None" entry, since that doesn't show up in any lists.
local displayPublisherCount = function (db)
    -- Count the number of rows in the publisher table. if this doesn't work,
    -- then just return without doing anything.
    local totalPublishers = rowsInTable (db , "PublisherList")
    if totalPublishers == nil then return end

    -- Decrement the number of publishers by 1, since the count counts the
    -- item with ID 0, "None".
    totalPublishers = totalPublishers - 1

    -- Figure out the proper grammar for the output.
    local label1 = "are "
    local label2 = " publishers"
    if totalPublishers == 1 then label1 = "is " label2 = " publisher" end

    -- Do it
    html:addText ('<FONT SIZE="-1">')
    html:addText ("There " , label1, totalPublishers , label2, " registered with the system.")
    html:addText ("</FONT><P>")
end

---------------------------------------------------------------------------

-- Generate the form that lists current publishers and the ability to delete them
local generatePublisherList = function (db)
    -- Say how many publishers there are.
    displayPublisherCount (db)

    -- If the user has javascript enabled, this confirmation will be used to get them to
    -- confirm what they are about to do.
    local deleteConfirm = "return confirm ('Any selected publishers will be completely removed from the system. Are you sure this is what you want to do?');"

    -- Open the form that will allow for the deletion of publishers
    html.form:open (env.scriptName , "POST", nil, nil, deleteConfirm)

    -- Open a table for the values
    html.table:open (nil, nil, nil, nil, "5", "publisherList")

    -- Make a short alias for wether or not the current user can edit publisher
    -- information.
    local canEdit = userInfo.canEditPublisher 

    -- Display a heading for all of the rows, starting
    -- with the name of this publisher.
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Publisher")

    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Games")

    -- If this user can edit publishers, then include extra columns.
    if canEdit then
        -- Properties column, has no heading
        html.table:heading (true, nil, nil, "LEFT", "TOP")
        html:addText ("&nbsp;")

        -- The column with the delete checkbox.
        html.table:heading (true, nil, nil, "LEFT", "TOP")
        html:addText ("Delete")
    end

    -- Get ready for the next row
    html.table:newRow ()

    -- Obtain a complete list of publishers, sorted by name, and excluding
    -- the default "None" publisher.
    --
    -- Once we have it, loop over everything displaying information on them.
    local pubList = getPublisherList (db, true, true)
    for index,row in pubList do
        -- Include the name of this publisher
        html.table:cell (true, nil, nil, "LEFT" , "TOP")
        html:addText (row.publisherName)

        -- Include how many games there are by this publisher
        html.table:cell (true, nil, nil, "LEFT" , "TOP")

        -- If we can determine how many games are handled by this publisher, then
        -- output a link to display all games by this publisher.
        local count = countGamesByPublisher (db, row.publisherID)
        if count ~= nil then
            html:link (urls.gamesByPublisher .. row.publisherID , count)
        else
            html:addText ("N/A")
        end

        -- If the user can edit,then we need to include properties and delete information
        if canEdit then
            -- A properties link
            html.table:cell (true, nil, nil, "LEFT" , "TOP")
            html:link (urls.publisherProperties .. row.publisherID , "Properties")

            -- A delete checkbox
            html.table:cell (true, nil, nil, "LEFT" , "TOP")
            html.form:checkbox ("delPublisherID" , row.publisherID)
        end

        -- Get ready for the next row
        html.table:newRow ()
    end

    -- If the publisher list has at least 1 item and this user can edit, then output the
    -- form button that will allow deletions. This isn't neccesary if there are no publishers
    -- or if the current user has no permissions.
    if table.getn (pubList) > 0 and canEdit then
        html.table:cell (true, nil, nil, "CENTER", "TOP", nil, nil, nil, "4")
        html.form:submit ("Button", "Delete selected Publishers", "publisher")
        html.form:hidden ("command", "PublisherDelete")
    end

    -- Close the table and form now
    html.table:close ()
    html.form:close ()
end

---------------------------------------------------------------------------

-- Generate the form that allows you to add new publishers to the system.
local generateAddForm = function (db)
     -- Output a blank line of the same size as the one that the publisher
     -- list uses to show how many publishers that there are. 
     html:addText ('<FONT SIZE="-1">&nbsp;</FONT><P>')

    -- Open the form that will allow for the addition of publishers
    html.form:open (env.scriptName , "POST")

    -- Open a table for the values
    html.table:open (nil, nil, nil, nil, "5", "publisherList")

    -- Display a heading for all of the rows
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Publisher Name")

    -- Get ready for the next row
    html.table:newRow ()

    -- Output a row for each publisher field we're configured for.
    for i = 1,ADD_PUBLISHER_FIELD_COUNT do
        html.table:cell (true, nil, nil, "LEFT" , "TOP")
        html.form:text ("newPublisher" .. i , "" , "40" , "60")
        html.table:newRow ()
    end

    -- Close up the table with a submit button.
    html.table:cell (true, nil, nil, "CENTER", "TOP")
    html.form:submit ("Button", "Add Publishers", "publisher")
    html.form:hidden ("command", "PublisherAdd")

    -- Close the table and form now
    html.table:close ()
    html.form:close ()
end

---------------------------------------------------------------------------

-- Takes some publisherInfo (ID and name, as in the PublistList table), and
-- then finds all games that have that publisher as the primary publisher. For
-- any that do, the primary publisher is replaced with the first match in
-- the PublisherPlacement table. If the game has only one publisher, then
-- the publisher is set to None.
local fixPublishedGames = function (db , publishInfo)
    -- We are removing the publisher in publishInfo. So, find all games for which
    -- that publisher is the primary publisher. 
    local foundGames = getGamesWithPrimaryPublisher (db, publishInfo.publisherID)

    -- Now, fix all of the found games.
    for index,row in foundGames do
        -- For this game, find the first row in PublisherPlacement for this game.
        local secondaryPublisher = getFirstSecondaryPublisherForGame (db, row.gameID)

        -- Now, if secondaryPublisher is nil, then this game had only a single publisher. In that
        -- case, set the publisher in this game to be 0 (None).
        if secondaryPublisher == nil then
            -- No secondary publisher, so just update this game to have a publisher that is empty.
            alterGamePrimaryPublisher (db, row.gameID, 0)
            html:addText ("Primary publisher of " , row.name , " has been set to 'None'<BR>")
        else
            -- Here there WAS a secondary publisher. First of all, delete this row from the
            -- PublisherPlacement table. Then, update the GameList entry for this game to have
            -- the secondary publisher as primary.
            deletePublisherPlacementByPlacement (db, secondaryPublisher.placementID)
            alterGamePrimaryPublisher (db, row.gameID, secondaryPublisher.publisherID)

            -- Figure out the new primary and say what happend
            local newPrimary = getPublisherByID (db, secondaryPublisher.publisherID)
            if newPrimary then
                html:addText ("Primary publisher of " , row.name , " has been set to '" , newPrimary.publisherName , "'<BR>")
            end
        end
    end
end

---------------------------------------------------------------------------

-- Called to process the deletion of one or more publishers from the system.
local processDeleteForm = function (db)
    -- Make sure that this user is allowed to delete publishers from the
    -- system.
    if userInfo.canEditPublisher == false then
        addLog (db, "Attempted to delete publishers")
        html:addText (errorText (ERR_AUTH_NODELPRIV) , "<P>")
        return 
    end

    -- Get the arguments that we need. The value could be nil, a string,
    -- or a table, depending on how many checkboxes on the form were checked.
    local args = collectFormArguments ({"delPublisherID"})

    -- We need at least one checked item in order to delete.
    if args == nil then
        addLog (db, "Attempted delete with no publishers selected")
        html:addText (errorText (ERR_DEL_NOPUBLISHERS) , "<P>")
        return
    end

    -- In case we got only a single publisher to delete, make sure that it's
    -- a list anyway. This makes the code below much cleaner.
    if type (args.delPublisherID) ~= "table" then
        args.delPublisherID = {args.delPublisherID}
    end

    -- Loop over and remove all publishers given. Confirmation was received
    -- before the form submitted.
    for index,auth in args.delPublisherID do
        -- The publisher with ID 0 is the "None" publisher, which is never 
        -- placed in the list to be deleted, but make sure anyway.
        if tonumber (auth) ~= 0 then
            -- Try to get information on the publisher that is being removed.
            local delPublishInfo = getPublisherByID (db, auth)

            -- If we can't find information on this publisher, then we're unhappy.
            if delPublishInfo == nil then
                html:addText (errorText (ERR_DEL_NOFINDPUBLISH) , "<BR>")
            else
                -- We're deleting this publisher. Firstly, remove all mention of this publisher
                -- from the PublisherList and the PublisherPlacement tables.
                deletePublisherByID (db, auth)
                deletePublisherPlacementsByPublisher (db, auth)

                -- Say what publisher was deleted
                html:addText (successText ("Deleted publisher '" .. delPublishInfo.publisherName .. "'") , "<BR>")
                addLog (db , "Deleted publisher '" .. delPublishInfo.publisherName .. "'")

                -- Now fix all games which have this publisher as the primary publisher. This will
                -- give those games a new primary publisher from one of the other publishers listed.
                -- Once this is done, output a paragraph separator to group things nicely.
                fixPublishedGames (db , delPublishInfo)
                html:addText ("<P>")
            end
        end
    end
end

---------------------------------------------------------------------------

-- Called to process the possible submission of an Add Publisher form
-- submission.
local processAddForm = function (db)
    -- Before we do anything, make sure that the current user has permission
    -- to add new publishers. We have to do this to stop people from faking
    -- URL's to do stuff they otherwise shouldn't be able to do.
    if userInfo.canEditPublisher == false then
        addLog (db, "Attempted to add publishers")
        html:addText (errorText (ERR_AUTH_NOADDPRIV) , "<P>")
        return 
    end

    -- Figure our what form arguments we're going to need. We configure
    -- this value at the top of this file, so that it's easily changeable.
    local argsNeeded = {}
    for i = 1,ADD_PUBLISHER_FIELD_COUNT do
        table.insert (argsNeeded , "newPublisher" .. i)
    end

    -- Get the form arguments. If we can't, then just leave. All this means
    -- is that no form has been submitted.
    local args = collectFormArguments (argsNeeded)
    if args == nil then return end

    -- Scan all of the fields to see if any publishers are being added.
    -- Check over all 5 of the publishers and see if we should add them.
    -- We keep track of each one so we can log and whatnot.
    local newPublish = {}
    for i = 1,ADD_PUBLISHER_FIELD_COUNT do
        -- Get this publisher from the form.
        local thisPublisher = string.trimleadingtrailing (args["newPublisher" .. i] or "")

        -- If this publisher isn't blank, then do something with it.
        if thisPublisher ~= "" then
            -- Try to get some information for this publisher
            local publisherInfo = getPublisherByName (db, thisPublisher)

            -- If we got nil, then someone is trying to add a new publisher, so do that. Otherwise,
            -- complain about a duplicate.
            if publisherInfo == nil then
                -- Create the new publisher.
                selectOrCreatePublisher (db, thisPublisher)

                -- Add some text to the page to say what publisher was added.
                local text = "Added publisher '" .. thisPublisher .. "'"
                html:addText (successText (text) , "<BR>")

                -- Add this publisher to the list.
                table.insert (newPublish, thisPublisher)
            else
                -- This publisher was duplicated.
                html:addText (errorText (ERR_PUB_DUPPUBLISH) , "<BR>")
            end
        end
    end

    -- Seperate output from above from stuff that will follow.
    html:addText ("<P>")

    -- If any publishers were added, then add a log.
    if table.getn (newPublish) ~= 0 then
        addLog (db, "Added publishers: " .. table.concat (newPublish, ", "))
    end
end

---------------------------------------------------------------------------

-- The main publisher commands for viewing the publisher list, adding
-- publishers and deleting publishers all route through the same command
-- handler. This checks the command argument to see what's what, and then
-- processes the proper form as needed.
local processForm = function (db)
    -- First, collect form arguments. What we really need is a command, 
    -- at the moment.
    local args = collectFormArguments ({"command"})

    -- If there are no arguments, then return. This can't really happen,
    -- since if there was no command, we wouldn't be here. But what the
    -- heck.
    if args == nil then return end

    -- If the command is just "Publisher", then the page is being viewed
    -- and nothing else is happening, so allow it.
    if args.command == "Publisher" then
        return 
    end

    -- Based on the command, process the form. We know the command has to
    -- be one of the two commands that we expect, since control wouldn't
    -- have reached us otherwise.
    if args.command == "PublisherAdd" then 
        processAddForm (db)
    else
        processDeleteForm (db) 
    end
end

---------------------------------------------------------------------------

-- Take the database, the database row for the publisher to be modified,
-- and form args {"publisherID" , "submitted", and "newPublisherName"}, where
-- the last two fields are optional. 
--
-- If this is a valid submission, then update and alter publisherInfo on the
-- way out.
local handlePropertyFormSubmission = function (db , publisherInfo , args)
    -- We only need to do something if the form has been submitted. 
    if args.submitted == nil then return end

    -- The form has been submitted. If the new publisher name is empty or
    -- nil, then complain.
    local newName = string.trimleadingtrailing (args.newPublisherName or "")

    -- If the new name is empty, then that's bad.
    if newName == "" then
        html:addText (errorText (ERR_PUB_BADPROPNAME) , "<P>")
        return
    end

    -- If the new name matches a publisher that's already in the system, that
    -- also is bad.
    local otherPublisher = getPublisherByName (db, newName)
    if otherPublisher then
        -- If this happens, it just means that someone submitted the property form
        -- without changing anything. The name is redundant, but the redundant
        -- ID is us, so it's ok. Just leave.
        -- If this is us, then the form was submitted with no changes made.
        if tonumber (otherPublisher.publisherID) == tonumber (publisherInfo.publisherID) then 
            html:addText (successText (SUCC_PUB_NOCHANGE), "<P>")
            return
        end

        -- At this point, the dupicate name is bad.
        html:addText (errorText (ERR_PUB_DUPPUBLISH) , "<P>")
        return
    end

    -- Rename the publisher, since the name seems to be ok.
    renamePublisherWithID (db, publisherInfo.publisherID, newName)

    -- Add to the system log and to the page information regarding the change.
    addLog (db, "Renamed publisher '" .. publisherInfo.publisherName .. "' to '" .. newName .. "'")
    html:addText (successText ("Renamed publisher '" .. publisherInfo.publisherName .. "' to '" .. newName .. "'") , "<P>")

    -- Before returning, make sure the publisher name is updated, so that the generated page
    -- is correct.
    publisherInfo.publisherName = newName
end

---------------------------------------------------------------------------

-- This handles publisher property editing commands.
-- Handle publisher property editing
function commandPublisherProperties (db)
    -- Header for the page
    generatePageHeader (thisPageSection, "Edit Publisher Properties", "publisherProperties")

    -- Generate a button to take us back to the main publisher page
    html.form:singleButtonForm ("Return to Publishers Page" , "Publisher" , env.scriptName , "POST" , nil , nil , "publisher")

    -- Check before we do anything to make sure that this user is allowed to 
    -- edit properties. 
    if userInfo.canEditPublisher == false then
        addLog (db, "Attempted to edit publishers properties")
        html:addText (errorText (ERR_AUTH_NOEDITPRIV) , "<P>")
        return 
    end

    -- Get some form arguments. At the very least, we need to have the publisherID
    -- of the publisher that we're editing. Optionally, we care about the submitted
    -- field and the new publisher ID field.
    local args = collectFormArguments ({"publisherID"} , {"submitted" , "newPublisherName"})
    if args == nil then 
        html:addText (errorText (ERR_PUB_NOPUBID) , "<P>")
        return
    end

    -- Get information on the publisher to be edited. Leave if we can't get
    -- the information.
    local publisherInfo = getPublisherByID (db, args.publisherID)
    if publisherInfo == nil then
        html:addText (errorText (ERR_PUB_NOPUBID) , "<P>")
        return
    end

    -- Handle form submission, if needed.
    handlePropertyFormSubmission (db , publisherInfo , args)

    -- Now display the form. 
    html.form:open (env.scriptName , "POST");
    html.table:open (nil, nil, nil, nil, "5", "publisherList")
    html.table:caption ("Properties for: <B>" .. publisherInfo.publisherName .. "</B>", "CENTER")

    -- The only property we support at the moment is changing the name of the publisher.
    html.table:heading (true , nil, nil, "RIGHT" , "CENTER")
    html:addText ("Name")
    html.table:cell (true, nil, nil, "LEFT", "CENTER")
    html.form:text ("newPublisherName" , publisherInfo.publisherName , "40" , "60")
    html.table:newRow ()

    -- The submit button.
    html.table:cell (true, nil, nil, "CENTER" , "TOP", nil, nil, nil, "2")
    html.form:submit ("Button" , "Update Properties" , "publisher")
 
    -- Table is complete now
    html.table:close ();

    -- Finish up the form now. Hidden form elements tell us if things have been
    -- published or not.
    html.form:hidden ("submitted" , "yes")
    html.form:hidden ("command" , "PublisherProperties")
    html.form:hidden ("publisherID" , publisherInfo.publisherID)
    html.form:close ();
end

---------------------------------------------------------------------------

-- This is the main function to view and edit game publishers
function commandPublisher (db)
    -- Header for the page
    generatePageHeader (thisPageSection, nil, "publisherBase")

    -- Handle the form submission, if there has been one. 
    processForm (db)

    -- If this user is allowed to edit publishers, then we are going to generate
    -- a special add form to allow for the addition. We want that to appear to 
    -- the right of the publisher list, so we will need a table to wrap everything.
    if userInfo.canEditPublisher then
        html.table:open (nil, nil, nil, nil, "10")
        html.table:cell (false, "1%", nil, "LEFT", "TOP")
    end

    -- Now, generate the delete form. This is the form that lists all of
    -- the publishers in the system, along with a delete checkbox next to
    -- them. If the current user can't delete publishers, then the delete
    -- checkboxes aren't listed. 
    generatePublisherList (db)

    -- If this user is allowed to edit publishers, then generate a form that
    -- allows for the addition of new publishers.
    if userInfo.canEditPublisher then
        html.table:cell (true, "99%", nil, "LEFT", "TOP")
        generateAddForm (db)
        html.table:close ()
    end
end

---------------------------------------------------------------------------
--- MAIN CODE -------------------------------------------------------------
---------------------------------------------------------------------------

