---------------------------------------------------------------------------
--- VARS ------------------------------------------------------------------
---------------------------------------------------------------------------


---------------------------------------------------------------------------
--- FUNCTIONS -------------------------------------------------------------
---------------------------------------------------------------------------

-- Return the contents of the activiy log, ordered in reverse order from the
-- order that the items were inserted. This makes the last thing that happened
-- appear first in the list, which is a boon if there are a lot of things
-- listed in the log.
function getActivityLog (db)
    local retVal = {}

    for row in db:SQLrows ("SELECT * FROM Logs ORDER BY logID DESC") do
        table.insert (retVal, row)
    end

    return retVal
end

---------------------------------------------------------------------------

-- Try to load the configuration information from the database. If this
-- works, a global named configInfo is created, which is a table in which
-- the keys and values map directly to those in the database. On success,
-- true is returned. Otherwise, false is returned.
function getConfigInfo (db)
    -- Set up the config information
    configInfo = {}

    -- Copy all values over. Note that the script will bomb if the select
    -- can't be executed (and the error log will say why)
    for data in db:SQLirows ("SELECT * FROM ConfigInfo") do
        configInfo[data[1]] = data[2]
    end

    -- Worked
    return true
end

---------------------------------------------------------------------------

-- Return the number of rows being stored in the passed in table, or nil on
-- error.
function rowsInTable (db, tableName)
    local info = db:SQLfirstRow ("SELECT count (*) as rowSize from " .. tableName)
    if info == nil then return nil end

    return info.rowSize
end

---------------------------------------------------------------------------

-- Given a game ID, return back the total count of games that are listed
-- as expansions for this game.
function expansionCountForGameID (db, gameID)
    local info = db:SQLfirstRow ("SELECT count (*) AS count FROM (SELECT * FROM GameList WHERE expandsGameID = ?)" , gameID)
    if info == nil then return nil end
    
    return info.count
end

---------------------------------------------------------------------------

-- Given a player ID, returns back the number of rows in the SessionReportEntry
-- where the player listed is listed as a winner.
function winCountForPlayerID (db, playerID)
    local info = db:SQLfirstRow ("SELECT count (*) AS count FROM (SELECT * FROM SessionReportEntry WHERE playerID = ? and winner = 'YES')", playerID)
    if info == nil then return nil end

    return info.count
end

---------------------------------------------------------------------------

-- Given a player ID, return back the number of rows in the SessionReportEntry
-- table where the player is listed as participating (winner or loser). This
-- will be the total number of games that this user has played.
function playCountForPlayerID (db, playerID)
    local info = db:SQLfirstRow ("SELECT count (*) AS count FROM (SELECT * FROM SessionReportEntry WHERE playerID = ?)", playerID)
    if info == nil then return nil end

    return info.count
end

---------------------------------------------------------------------------

-- Sort function that sorts according to player nickname
local sortByNickName = function (first, second)
    -- Return true if the first is less than the second.
    if string.strcasecmp (first.nickname, second.nickname) < 0 then
        return true
    else
        return false
    end
end

---------------------------------------------------------------------------

-- Return an array of player information, or nil on error.
--
-- If sortedByName is given, then the players returned will be sorted
-- by nickname (case insensitively). Otherwise, they're in the order that they
-- come out of the database.
function getPlayerList (db, sortedByName)
    local retVal = {}

    -- Get all the data
    for row in db:SQLrows ("SELECT * FROM PlayerList") do
        table.insert (retVal, row)
    end

    -- Sort if needed before returning
    if sortedByName then table.sort (retVal, sortByNickName) end
    return retVal
end

---------------------------------------------------------------------------

-- Get information about a player based on their player ID
function getPlayerByID (db, id)
    return db:SQLfirstRow ("SELECT * FROM PlayerList WHERE playerID = ?" , id)
end

---------------------------------------------------------------------------

-- Look up a player by name
function getPlayerByName (db, name)
    return db:SQLfirstRow ("SELECT * FROM PlayerList WHERE nickName = ?" , name)
end

---------------------------------------------------------------------------

-- Get information about a player based on a nickname and password. This is
-- used during login.
--
-- nil is returned if there is no such nickname/password combination.
function getPlayerByNicknamePassword (db, nickname, password)
    return db:SQLfirstRow ("SELECT * FROM PlayerList WHERE nickname = ? and password = ?" , nickname , password) 
end

---------------------------------------------------------------------------

-- If there is an entry in the session list for this authorization token, then
-- return it. Otherwise, return nil.
function getSessionForAuthToken (db, token)
    return db:SQLfirstRow ("SELECT * FROM SessionList WHERE authToken = ?" , token)
end

---------------------------------------------------------------------------

-- Find the session attached to this player ID, or nil if there isn't one.
function getSessionForPlayer (db, id)
    return db:SQLfirstRow ("SELECT * FROM SessionList WHERE playerID = ?" , id)
end

---------------------------------------------------------------------------

-- Sort function that sorts according to game name.
local sortByGameName = function (first, second)
    -- Return true if the first is less than the second.
    if string.strcasecmp (first.name, second.name) < 0 then
        return true
    else
        return false
    end
end

---------------------------------------------------------------------------

-- Takes an array of tables which have a key "gameID", and returns back
-- a table in which the gameID is the key and the associated entry is the
-- value.
local invertTableWithGameID = function (table)
    local new = {}

    -- Invert
    for index,content in table do
        new[content.gameID] = content
    end

    return new
end

---------------------------------------------------------------------------

-- This function takes 2 arguments. The first is a table of full game
-- information where the key is the gameID of the game in question, and
-- the value is a row of information direct from the DB.
--
-- The second paramter is just a simple array of results from a secondary
-- database query. The rows may differ (could be designer, category or
-- publisher), but there is always a key of gameID.
--
-- This scans otherList, and for every gameID mentioned that doesn't already
-- exist in gameList, this will look up the information for that game and
-- insert it into the gameList table.
--
-- After the merge operation is complete, this will create a new array that
-- is just game information from gameList (essentially a copy of gameList no
-- longer indexed by gameID), sort it according to game name, and then return
-- it back.
local mergeGameInfo = function (db, gameList , otherList)
    -- Scan over all of the entries in otherList
    for index,row in otherList do
        -- If the gameID mentioned here isn't already in gameList,
        -- then try to add it.
        if gameList[row.gameID] == nil then
            -- Try to get info for this game
            local info = getGameByID (db, row.gameID)

            -- If we did, then add it into the game list
            if info ~= nil then
                gameList[row.gameID] = info
            end
        end
    end

    -- Now create a new list
    local fixedList = {}

    -- Insert all of the data from the first gameList into it
    for gameID,gameInfo in gameList do
        table.insert (fixedList , gameInfo)
    end

    table.sort (fixedList, sortByGameName)

    -- Return the new gameList
    return fixedList
end

---------------------------------------------------------------------------

-- Sort function that sorts according to category name.
local sortByCategoryName = function (first, second)
    -- Return true if the first is less than the second.
    if string.strcasecmp (first.categoryName, second.categoryName) < 0 then
        return true
    else
        return false
    end
end

---------------------------------------------------------------------------

-- Get category information for the category with the given name, or nil 
-- if there is no such name.
function getCategoryByName (db, name)
    return db:SQLfirstRow ("SELECT * FROM CategoryList WHERE categoryName = ?" , name)
end

---------------------------------------------------------------------------

-- Get category information for the category with the given id, or nil 
-- if there is no such id.
function getCategoryByID (db, id)
    return db:SQLfirstRow ("SELECT * FROM CategoryList WHERE categoryID = ?" , id)
end

---------------------------------------------------------------------------

-- Return an array of category information, or nil on error.
--
-- If sortedByName is given, then the categories returned will be sorted
-- by name (case insensitively). Otherwise, they're in the order that they
-- come out of the database.
--
-- If excludeDefault is true, then the category with id 0 will not be
-- in the result set.
function getCategoryList (db, sortedByName, excludeDefault)
    local retVal = {}

    -- Get all the data
    for row in db:SQLrows ("SELECT * FROM CategoryList") do
        -- Check to see if we want this one or not
        local wantThis = true
        if excludeDefault and tonumber (row.categoryID) == 0 then wantThis = false end

        -- If we want it, insert it
        if wantThis then table.insert (retVal, row) end
    end

    -- Sort if needed before returning
    if sortedByName then table.sort (retVal, sortByCategoryName) end
    return retVal
end

---------------------------------------------------------------------------

-- Return the total number of games attached to this category, or nil on
-- error.
function countGamesInCategory (db, categoryID)
    -- First, determine how many rows there are in the GameList that have
    -- this designer as the primary.
    local primInfo = db:SQLfirstRow ("SELECT count (*) AS primCount FROM (SELECT * FROM GameList where categoryID = ?)" , categoryID)
    if primInfo == nil then return nil end

    -- Secondly, find all rows in DesignerPlacement that reference this designer
    local secInfo = db:SQLfirstRow ("SELECT count (*) AS secCount FROM (SELECT * FROM CategoryPlacement where categoryID = ?)" , categoryID)
    if secInfo == nil then return nil end

    -- Return the total count.
    return primInfo.primCount + secInfo.secCount
end

---------------------------------------------------------------------------

-- Sort function that sorts according to designer name.
local sortByDesignerName = function (first, second)
    -- Return true if the first is less than the second.
    if string.strcasecmp (first.designerName, second.designerName) < 0 then
        return true
    else
        return false
    end
end

---------------------------------------------------------------------------

-- Get designer information for the designer with the given name, or nil 
-- if there is no such name.
function getDesignerByName (db, name)
    return db:SQLfirstRow ("SELECT * FROM DesignerList WHERE designerName = ?" , name)
end

---------------------------------------------------------------------------

-- Get designer information for the designer with the given id, or nil 
-- if there is no such id.
function getDesignerByID (db, id)
    return db:SQLfirstRow ("SELECT * FROM DesignerList WHERE designerID = ?" , id)
end

---------------------------------------------------------------------------

-- Return an array of designer information, or nil on error.
--
-- If sortedByName is given, then the designers returned will be sorted
-- by name (case insensitively). Otherwise, they're in the order that they
-- come out of the database.
--
-- If excludeDefault is true, then the designer with id 0 will not be
-- in the result set.
function getDesignerList (db, sortedByName, excludeDefault)
    local retVal = {}

    -- Get all the data
    for row in db:SQLrows ("SELECT * FROM DesignerList") do
        -- Check to see if we want this one or not
        local wantThis = true
        if excludeDefault and tonumber (row.designerID) == 0 then wantThis = false end

        -- If we want it, insert it
        if wantThis then table.insert (retVal, row) end
    end

    -- Sort if needed before returning
    if sortedByName then table.sort (retVal, sortByDesignerName) end
    return retVal
end

---------------------------------------------------------------------------

-- Return the total number of games attached to this designer, or nil on
-- error.
function countGamesByDesigner (db, designerID)
    -- First, determine how many rows there are in the GameList that have
    -- this designer as the primary.
    local primInfo = db:SQLfirstRow ("SELECT count (*) AS primCount FROM (SELECT * FROM GameList where designerID = ?)" , designerID)
    if primInfo == nil then return nil end

    -- Secondly, find all rows in DesignerPlacement that reference this designer
    local secInfo = db:SQLfirstRow ("SELECT count (*) AS secCount FROM (SELECT * FROM DesignerPlacement where designerID = ?)" , designerID)
    if secInfo == nil then return nil end

    -- Return the total count.
    return primInfo.primCount + secInfo.secCount
end

---------------------------------------------------------------------------

-- Sort function that sorts according to publisher name.
local sortByPublisherName = function (first, second)
    -- Return true if the first is less than the second.
    if string.strcasecmp (first.publisherName, second.publisherName) < 0 then
        return true
    else
        return false
    end
end

---------------------------------------------------------------------------

-- Get publisher information for the publisher with the given name, or nil 
-- if there is no such name.
function getPublisherByName (db, name)
    return db:SQLfirstRow ("SELECT * FROM PublisherList WHERE publisherName = ?" , name)
end

---------------------------------------------------------------------------

-- Get publisher information for the publisher with the given id, or nil 
-- if there is no such id.
function getPublisherByID (db, id)
    return db:SQLfirstRow ("SELECT * FROM PublisherList WHERE publisherID = ?" , id)
end

---------------------------------------------------------------------------

-- Return an array of publisher information, or nil on error.
--
-- If sortedByName is given, then the publishers returned will be sorted
-- by name (case insensitively). Otherwise, they're in the order that they
-- come out of the database.
--
-- If excludeDefault is true, then the designer with id 0 will not be
-- in the result set.
function getPublisherList (db, sortedByName, excludeDefault)
    local retVal = {}

    -- Get all the data
    for row in db:SQLrows ("SELECT * FROM PublisherList") do
        -- Check to see if we want this one or not
        local wantThis = true
        if excludeDefault and tonumber (row.publisherID) == 0 then wantThis = false end

        -- If we want it, insert it
        if wantThis then table.insert (retVal, row) end
    end

    -- Sort if needed before returning
    if sortedByName then table.sort (retVal, sortByPublisherName) end
    return retVal
end

---------------------------------------------------------------------------

-- Return the total number of games attached to this publisher, or nil on
-- error.
function countGamesByPublisher (db, publisherID)
    -- First, determine how many rows there are in the GameList that have
    -- this designer as the primary.
    local primInfo = db:SQLfirstRow ("SELECT count (*) AS primCount FROM (SELECT * FROM GameList where publisherID = ?)" , publisherID)
    if primInfo == nil then return nil end

    -- Secondly, find all rows in DesignerPlacement that reference this designer
    local secInfo = db:SQLfirstRow ("SELECT count (*) AS secCount FROM (SELECT * FROM PublisherPlacement where publisherID = ?)" , publisherID)
    if secInfo == nil then return nil end

    -- Return the total count.
    return primInfo.primCount + secInfo.secCount
end

---------------------------------------------------------------------------

-- Return a list of all games for whom the given category ID is the primary
-- category.
function getGamesWithPrimaryCategory (db, categoryID)
    local retVal = {}

    for row in db:SQLrows ("SELECT * FROM GameList WHERE categoryID = ?" , categoryID) do
        table.insert (retVal, row)
    end

    return retVal
end

---------------------------------------------------------------------------

-- If this game has at least one secondary category, then return back the
-- category information for the first one found. Otherwise, return nil.
function getFirstSecondaryCategoryForGame (db, gameID)
    return db:SQLfirstRow ("SELECT * FROM CategoryPlacement WHERE gameID = ?" , gameID)
end

---------------------------------------------------------------------------

-- Return a list of CategoryPlacement information for all rows that mention
-- this category ID.
function getSecondaryCategoryPlacements (db, categoryID)
    local retVal = {}

    for row in db:SQLrows ("SELECT * FROM CategoryPlacement WHERE categoryID = ?" , categoryID) do
        table.insert (retVal, row)
    end

    return retVal
end

---------------------------------------------------------------------------

-- Return a list of all games for whom the given designer ID is the primary
-- designer.
function getGamesWithPrimaryDesigner (db, designerID)
    local retVal = {}

    for row in db:SQLrows ("SELECT * FROM GameList WHERE designerID = ?" , designerID) do
        table.insert (retVal, row)
    end

    return retVal
end

---------------------------------------------------------------------------

-- If this game has at least one secondary designer, then return back the
-- designer information for the first one found. Otherwise, return nil.
function getFirstSecondaryDesignerForGame (db, gameID)
    return db:SQLfirstRow ("SELECT * FROM DesignerPlacement WHERE gameID = ?" , gameID)
end

---------------------------------------------------------------------------

-- Return a list of DesignerPlacement information for all rows that mention
-- this designer ID.
function getSecondaryDesignerPlacements (db, designerID)
    local retVal = {}

    for row in db:SQLrows ("SELECT * FROM DesignerPlacement WHERE designerID = ?" , designerID) do
        table.insert (retVal, row)
    end

    return retVal
end

---------------------------------------------------------------------------

-- Return a list of all games for whom the given publisher ID is the primary
-- publisher.
function getGamesWithPrimaryPublisher (db, publisherID)
    local retVal = {}

    for row in db:SQLrows ("SELECT * FROM GameList WHERE publisherID = ?" , publisherID) do
        table.insert (retVal, row)
    end

    return retVal
end

---------------------------------------------------------------------------

-- If this game has at least one secondary publisher, then return back the
-- publisher information for the first one found. Otherwise, return nil.
function getFirstSecondaryPublisherForGame (db, gameID)
    return db:SQLfirstRow ("SELECT * FROM PublisherPlacement WHERE gameID = ?" , gameID)
end

---------------------------------------------------------------------------

-- Return a list of PublisherPlacement information for all rows that mention
-- this publisher ID.
function getSecondaryPublisherPlacements (db, publisherID)
    local retVal = {}

    for row in db:SQLrows ("SELECT * FROM PublisherPlacement WHERE publisherID = ?" , publisherID) do
        table.insert (retVal, row)
    end

    return retVal
end

---------------------------------------------------------------------------

-- Get game information for the game with the given id, or nil 
-- if there is no such id.
function getGameByID (db, id)
    return db:SQLfirstRow ("SELECT * FROM GameList WHERE gameID = ?" , id)
end

---------------------------------------------------------------------------

-- Return an array of game information, or nil on error.
--
-- If sortedByName is given, then the games returned will be sorted
-- by name (case insensitively). Otherwise, they're in the order that they
-- come out of the database.
--
-- If excludingID is non-nil, then the game with that ID will not be added
-- to the list.
function getGameList (db, sortedByName, excludingID)
    local retVal = {}

    -- Get all the data
    for row in db:SQLrows ("SELECT * FROM GameList") do
        local wantThis = true

        if excludingID and tonumber (excludingID) == tonumber (row.gameID) then wantThis = false end

        if wantThis then table.insert (retVal, row) end
    end

    -- Sort if needed before returning
    if sortedByName then table.sort (retVal, sortByGameName) end
    return retVal
end

---------------------------------------------------------------------------

-- Given a gameID, return back all of the games that list it as an expansion.
function getExpansionsForGameID (db, gameID, sortedByName)
    local retVal = {}

    for row in db:SQLrows ("SELECT * FROM GameList WHERE expandsGameID = ?", gameID) do
        table.insert (retVal, row)
    end

    -- Sort if needed before returning
    if sortedByName then table.sort (retVal, sortByGameName) end
    return retVal;
end

---------------------------------------------------------------------------

-- Finds all games designed by the designer in question. This includes games
-- that would be returned by getGamesWithPrimaryDesigner, as well as any
-- games for which this designer was a secondary designer as well.
function getAllGamesByDesigner (db, designerID)
    -- First, get the primary and secondary lists of games by this designer.
    local primaryGames   = getGamesWithPrimaryDesigner (db, designerID)
    local secondaryGames = getSecondaryDesignerPlacements (db, designerID)

    -- Invert both tables.
    primaryGames   = invertTableWithGameID (primaryGames)
    secondaryGames = invertTableWithGameID (secondaryGames)

    -- Now merge the items together and return it
    return mergeGameInfo (db, primaryGames, secondaryGames)
end

---------------------------------------------------------------------------

-- Finds all games published by the publisher in question. This includes games
-- that would be returned by getGamesWithPrimaryPublisher, as well as any
-- games for which this publisher was a secondary publisher as well.
function getAllGamesByPublisher (db, publisherID)
    -- First, get the primary and secondary lists of games by this designer.
    local primaryGames   = getGamesWithPrimaryPublisher (db, publisherID)
    local secondaryGames = getSecondaryPublisherPlacements (db, publisherID)

    -- Invert both tables.
    primaryGames   = invertTableWithGameID (primaryGames)
    secondaryGames = invertTableWithGameID (secondaryGames)

    -- Now merge the items together and return it
    return mergeGameInfo (db, primaryGames, secondaryGames)
end

---------------------------------------------------------------------------

-- Finds all games in the category in question. This includes games
-- that would be returned by getGamesWithPrimaryCategory, as well as any
-- games for which this category is a secondary category as well.
function getAllGamesByCategory (db, categoryID)
    -- First, get the primary and secondary lists of games in this category
    local primaryGames   = getGamesWithPrimaryCategory (db, categoryID)
    local secondaryGames = getSecondaryCategoryPlacements (db, categoryID)

    -- Invert both tables.
    primaryGames   = invertTableWithGameID (primaryGames)
    secondaryGames = invertTableWithGameID (secondaryGames)

    -- Now merge the items together and return it
    return mergeGameInfo (db, primaryGames, secondaryGames)
end

---------------------------------------------------------------------------

-- Get all the main session report information for a particular game ID.
function getSessionReportsForGameID (db, gameID)
    local retVal = {}

    for row in db:SQLrows ("SELECT * FROM SessionReport WHERE gameID = ? ORDER BY sessionDate DESC" , gameID) do
        table.insert (retVal , row)
    end

    return retVal
end

---------------------------------------------------------------------------

-- Given a GameID, return the number of session reports for this gameID
function countSessionReportsForGameID (db, gameID)
    local info = db:SQLfirstRow ("SELECT count (*) AS count FROM (SELECT * FROM SessionReport WHERE gameID = ?)" , gameID)
    if info == nil then return nil end
    
    return info.count
end

---------------------------------------------------------------------------

-- This will take some number that is 1+, and then find that many of the
-- most recent session reports. 
--
-- The return value is an array of strings which represent the gameID's of
-- the session reports found. The list is unique, so if multiple session 
-- reports list the same game, it will only be added to the list once.
--
-- The array can have anywhere from 0 to the number passed in values in it,
-- depending on how many session reports there are in the system, and how
-- many of them are unique.
function getGamesFromRecentSessionReports (db, sessionCount)
    -- Make sure that we have been given a value session count
    if sessionCount < 1 then sessionCount = 1 end

    -- Iterate over all of the rows, and for each one, insert the gameID into the temporary table
    -- if it's not already there. We have to do this manually and not with a DISTINCT clause, 
    -- because then it will go too far back in time if there are multiple entries (because it wants
    -- to return the limit provided if it can).
    local temp = {}
    for row in db:SQLrows ("SELECT * FROM SessionReport ORDER BY sessionDate DESC LIMIT ?", sessionCount) do
        -- Put this gameID into the temporary as a key if it's not already there
        if temp[row.gameID] == nil then temp[row.gameID] = true end
    end

    -- Now create the return value, and insert all of the keys from the temp table in it.
    local retVal = {}
    for key,value in temp do table.insert (retVal, key) end

    -- Return it
    return retVal
end

---------------------------------------------------------------------------

-- Get the session report with the provided ID.
function getSessionReportByID (db, sessionID)
    return db:SQLfirstRow ("SELECT * FROM SessionReport WHERE sessionID = ?" , sessionID)
end

---------------------------------------------------------------------------

-- Update the session report provided.
function updateSessionReport (db, sessionID, gameID, date, duration, notes)
    db:execSQL ("UPDATE SessionReport set sessionDate = ?, gameID = ?, sessionDuration = ?, notes = ? WHERE sessionID = ?", 
                date, gameID, duration, notes, sessionID)
end

---------------------------------------------------------------------------

-- Get all players for the provided session ID. 
--
-- if winnersOnly is true, then only people flagged as being winners will
-- be returned.
function getPlayersForSessionID (db, sessionID, winnersOnly)
    local retVal = {}

    for row in db:SQLrows ("SELECT * FROM SessionReportEntry WHERE sessionID = ?" , sessionID) do
        local wantThis = true

        -- If we want winners only and this isn't YES, then we don't want this.
        -- If we want everyone this never happens.
        if winnersOnly and string.upper (row.winner) ~= "YES" then wantThis = false end

        if wantThis then table.insert (retVal , row) end
    end

    return retVal
end

---------------------------------------------------------------------------

-- Insert a new SessionReport entry for the given gameID. The return value
-- is the session report's ID.
function addSessionReport (db, gameID, date, duration, notes)
    db:execSQL ("INSERT INTO SessionReport VALUES (NULL, ?, ?, ?, ?)" , date, duration, gameID, notes)
    return db:last_insert_rowid ()
end

---------------------------------------------------------------------------

-- Add someone to a session report entry for a given session report.
function addSessionReportPlayer (db, sessionID, playerID, score, winner)
    if winner then winner = "YES" else winner = "NO" end

    db:execSQL ("INSERT INTO SessionReportEntry VALUES (NULL, ?, ?, ?, ?)" , sessionID, playerID, score, winner)
end

---------------------------------------------------------------------------

-- Get the complete session list.
function getSessionList (db)
    local retVal = {}

    for row in db:SQLrows ("SELECT * FROM SessionReport ORDER BY sessionDate DESC") do
        table.insert (retVal , row)
    end

    return retVal
end

---------------------------------------------------------------------------

-- Get the designerID for the designer named. If there is no such designer,
-- then add a new one.
function selectOrCreateDesigner (db, name)
    -- Try to get the existing designer, if possible.
    local info = getDesignerByName (db, name)
    if info then return info.designerID end

    -- No good, add a new one and return it's id.
    db:execSQL ("INSERT INTO DesignerList VALUES (NULL, ?)" , name)
    return db:last_insert_rowid ()
end

---------------------------------------------------------------------------

-- Get the publisherID for the publisher named. If there is no such publisher,
-- then add a new one.
function selectOrCreatePublisher (db, name)
    -- Try to get the existing designer, if possible.
    local info = getPublisherByName (db, name)
    if info then return info.publisherID end

    -- No good, add a new one and return it's id.
    db:execSQL ("INSERT INTO PublisherList VALUES (NULL, ?)" , name)
    return db:last_insert_rowid ()
end

---------------------------------------------------------------------------

-- Get the categoryID for the category named. If there is no such category,
-- then add a new one.
function selectOrCreateCategory (db, name)
    -- Try to get the existing category, if possible.
    local info = getCategoryByName (db, name)
    if info then return info.categoryID end

    -- No good, add a new one and return it's id.
    db:execSQL ("INSERT INTO CategoryList VALUES (NULL, ?)" , name)
    return db:last_insert_rowid ()
end

---------------------------------------------------------------------------

-- Look up all information in the given DB for gameInfo.gameID such that:
--
--  o Find all entries in lookupTable that have the gameID given
--    o For each, save the field crossRefKey
--
--  o For each of the saved fields from the first step:
--    o Look up the record in masterTable where crossRefKey is the one saved
--      o For each looked up record, save the value of lookupKey
--
--  o Sort all fields saved by the second step
--
--  o Perform step 2 exclusively for the field crossRefKey in the given gameInfo
--    o This value is inserted into the first position of the data saved in step 3
--
--  o Return the stored table, which will be an array of tables with 2 keys, "name" and "id".
local getCrossRefForFields = function (db , gameInfo , lookupKey, crossRefKey, lookupTable, masterTable)
    -- First of all, collect all of the secondary information. There might not be any at all.
    local secondary = {}
    for row in db:SQLrows ("SELECT * FROM " .. lookupTable .. " WHERE gameID = ?" , gameInfo.gameID) do
        table.insert (secondary , row[crossRefKey])
    end

    -- Loop over all of the saved keys to look up their names in the master table given.
    local retVal = {}
    for index,value in secondary do
        local info = db:SQLfirstRow ("SELECT * FROM " .. masterTable .. " WHERE " .. crossRefKey .. " = ?" , value)
        
        -- Create an item and set up it's value
        local thisItem = {}
        thisItem.id = value;

        -- Set in a name
        if info ~= nil then
            thisItem.name = info[lookupKey]
        else
            thisItem.name = "Unknown"
        end

        -- Insert it now
        table.insert (retVal, thisItem)
    end

    -- Now, sort what we just came up with.
    table.sort (retVal, sortByGameName)

    -- Now, perform a lookup similar to the above, but for the field that we got in the
    -- gameInfo structure.
    local info = db:SQLfirstRow ("SELECT * FROM " .. masterTable .. " WHERE " .. crossRefKey .. " = ?" , gameInfo[crossRefKey])
    local thisItem = {}
    thisItem.id = gameInfo[crossRefKey]

    -- Insert the unknown
    if info ~= nil then
        thisItem.name = info[lookupKey]
    else
        thisItem.name = "Unknown"
    end

    -- Insert the primary
    table.insert (retVal , 1, thisItem)

    -- Now we can return the retVal.
    return retVal
end

---------------------------------------------------------------------------

-- Get all designers for the provided game info, would should be a complete
-- database record.
function getAllDesignersForGameInfo (db, gameInfo)
    return getCrossRefForFields (db, gameInfo, "designerName"  , "designerID"  , "DesignerPlacement"  , "DesignerList")
end

---------------------------------------------------------------------------

-- Get all publishers for the provided game info, would should be a complete
-- database record.
function getAllPublishersForGameInfo (db, gameInfo)
    return getCrossRefForFields (db, gameInfo, "publisherName" , "publisherID" , "PublisherPlacement" , "PublisherList")
end

---------------------------------------------------------------------------

-- Get all categories for the provided game info, would should be a complete
-- database record.
function getAllCategoriesForGameInfo (db, gameInfo)
    return getCrossRefForFields (db, gameInfo, "categoryName"  , "categoryID"  , "CategoryPlacement"  , "CategoryList")
end

---------------------------------------------------------------------------

-- Empty all items in the system activity log.
function emptyActivityLog (db)
    db:execSQL ("DELETE FROM Logs")
end

---------------------------------------------------------------------------

-- Delete all items from SessionList where the last access time is old enough
-- that the user should now be considered logged out.
function deleteOldSessions (db, lastValidTime)
    db:execSQL ("DELETE FROM SessionList WHERE lastAccessDate < ?" , lastValidTime)
end

---------------------------------------------------------------------------

-- Delete all entries from SessionList that have the given authToken. 
function deleteSessionForAuthToken (db, authToken)
    db:execSQL ("DELETE FROM SessionList WHERE authToken = ?" , authToken)
end

---------------------------------------------------------------------------

-- Delete all entries from SessionList for the given playerID
function deleteSessionForPlayerID (db, playerID)
    db:execSQL ("DELETE FROM SessionList WHERE playerID = ?" , playerID)
end

---------------------------------------------------------------------------

-- Delete the player with the given id
function deletePlayerByID (db, playerID)
    db:execSQL ("DELETE FROM PlayerList WHERE playerID = ?" , playerID)
end

---------------------------------------------------------------------------

-- Delete the given categoryID from the database. 
function deleteCategoryByID (db, categoryID)
    db:execSQL ("DELETE FROM CategoryList WHERE categoryID = ?" , categoryID)
end

---------------------------------------------------------------------------

-- Delete a category placement record by it's placement ID
function deleteCategoryPlacementByPlacement (db, placementID)
    db:execSQL ("DELETE FROM CategoryPlacement WHERE placementID = ?" , placementID)
end

---------------------------------------------------------------------------

-- Delete a category placement record by it's game ID
function deleteCategoryPlacementByGame (db, gameID)
    db:execSQL ("DELETE FROM CategoryPlacement WHERE gameID = ?" , gameID)
end

---------------------------------------------------------------------------

-- Delete all category placements for the given category ID
function deleteCategoryPlacementsByCategory (db, categoryID)
    db:execSQL ("DELETE FROM CategoryPlacement WHERE categoryID = ?" , categoryID)
end

---------------------------------------------------------------------------

-- Delete the given designerID from the database. 
function deleteDesignerByID (db, designerID)
    db:execSQL ("DELETE FROM DesignerList WHERE designerID = ?" , designerID)
end

---------------------------------------------------------------------------

-- Delete a designer placement record by it's placement ID
function deleteDesignerPlacementByPlacement (db, placementID)
    db:execSQL ("DELETE FROM DesignerPlacement WHERE placementID = ?" , placementID)
end

---------------------------------------------------------------------------

-- Delete a designer placement record by it's game ID
function deleteDesignerPlacementByGame (db, gameID)
    db:execSQL ("DELETE FROM DesignerPlacement WHERE gameID = ?" , gameID)
end

---------------------------------------------------------------------------

-- Delete all designer placements for the given designer ID
function deleteDesignerPlacementsByDesigner (db, designerID)
    db:execSQL ("DELETE FROM DesignerPlacement WHERE designerID = ?" , designerID)
end

---------------------------------------------------------------------------

-- Delete the given publisherID from the database. 
function deletePublisherByID (db, publisherID)
    db:execSQL ("DELETE FROM PublisherList WHERE publisherID = ?" , publisherID)
end

---------------------------------------------------------------------------

-- Delete a publisher placement record by it's placement ID
function deletePublisherPlacementByPlacement (db, placementID)
    db:execSQL ("DELETE FROM PublisherPlacement WHERE placementID = ?" , placementID)
end

---------------------------------------------------------------------------

-- Delete a publisher placement record by it's game ID
function deletePublisherPlacementByGame (db, gameID)
    db:execSQL ("DELETE FROM PublisherPlacement WHERE gameID = ?" , gameID)
end

---------------------------------------------------------------------------

-- Delete all publisher placements for the given publisher ID
function deletePublisherPlacementsByPublisher (db, publisherID)
    db:execSQL ("DELETE FROM PublisherPlacement WHERE publisherID = ?" , publisherID)
end

---------------------------------------------------------------------------

-- Delete a game by it's ID
function deleteGameByID (db, gameID)
    db:execSQL ("DELETE FROM GameList WHERE gameID = ?" , gameID)
end

---------------------------------------------------------------------------

-- Delete a session report by id
function deleteSessionReportByID (db, sessionID)
    db:execSQL ("DELETE FROM SessionReport WHERE sessionID = ?" , sessionID)
end

---------------------------------------------------------------------------

-- Delete all session report entries for the given session report id
function deleteSessionReportEntriesForSessionID (db, sessionID)
    db:execSQL ("DELETE FROM SessionReportEntry WHERE sessionID = ?" , sessionID)
end

---------------------------------------------------------------------------

-- Add a log to the system. The current IP address and playerID and time
-- will be filled in automatically.
function addLog (db, msg)
    -- Set up the player that is adding the log. If there is a current valid
    -- player, then use that ID. Otherwise, use 0, which is the "system".
    local playerID
    if userInfo == nil then
        playerID = 0
    else
        playerID = userInfo.playerID or 0
    end

    -- Set up the ip address of the user creating the log, and the time that
    -- it is happening.
    local ipAddr = env.remoteAddr or "UNKNOWN"
    local eventTime = tostring (os.time ())

    -- Now, set up the message
    local message = msg or "Empty message"

    -- Add the log now
    db:execSQL ("INSERT INTO Logs VALUES (NULL, ?, ?, ?, ?)" , message, ipAddr, playerID, eventTime)
end

---------------------------------------------------------------------------

-- Add an entry to the SessionList table for this authToken, playerID and
-- mark it with the time given.
function addSessionListEntry (db, authToken, playerID, currentTime)
    db:execSQL ("INSERT INTO SessionList VALUES (NULL, ?, ?, ?)", authToken, playerID, currentTime)
end

---------------------------------------------------------------------------

-- Add a new player to the system.
function addNewPlayer (db, name, nickname, password, email, homepage, creatorID)
    db:execSQL ("INSERT INTO PlayerList VALUES (NULL, ?, ?, ?, ?, ?, 'NO', 'NO', 'NO', 'NO', 'NO', 'NO', 'NO', 'NO', 'NO', 'NO', 'NO', 'NO', 'NO', 'NO', ?)" , 
        name, nickname, password, email, homepage, creatorID)
end

---------------------------------------------------------------------------

-- Add a new designer placement that maps this game id to this designer.
function addDesignerPlacement (db, gameID, designerID)
    db:execSQL ("INSERT INTO DesignerPlacement VALUES (NULL, ?, ?)" , gameID, designerID)
end

---------------------------------------------------------------------------

-- Add a new publisher placement that maps this game id to this publisher.
function addPublisherPlacement (db, gameID, publisherID)
    db:execSQL ("INSERT INTO PublisherPlacement VALUES (NULL, ?, ?)" , gameID, publisherID)
end

---------------------------------------------------------------------------

-- Add a new category placement that maps this game id to this category.
function addCategoryPlacement (db, gameID, categoryID)
    db:execSQL ("INSERT INTO CategoryPlacement VALUES (NULL, ?, ?)" , gameID, categoryID)
end

---------------------------------------------------------------------------

-- Insert a new game into the system, and return it's gameID
function addNewGame (db, bgg, name, description, designer, publisher, category, minPlay, maxPlay, playTime, expands)
    db:execSQL ("INSERT INTO GameList VALUES (NULL, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)" ,
        bgg, name, description, designer, publisher, category, minPlay, maxPlay, playTime, expands)

    return db:last_insert_rowid ()
end

---------------------------------------------------------------------------

-- Update the last access time for the SessionList id given
function updateAccessTimeForSession (db, sessionID, newTime)
    db:execSQL ("UPDATE SessionList SET lastAccessDate = ? WHERE sessionID = ?" , newTime, sessionID)
end

---------------------------------------------------------------------------

-- Set the config value given to the new value.
function updateConfigInfo (db, configItem, newValue)
    db:execSQL ("UPDATE ConfigInfo SET itemValue = ? WHERE itemName = ?" , newValue, configItem)
end

---------------------------------------------------------------------------

-- Rename the category with the given id
function renameCategoryWithID (db, categoryID, newName)
    db:execSQL ("UPDATE CategoryList SET categoryName = ? where categoryID = ?" , newName, categoryID)
end

---------------------------------------------------------------------------

-- Rename the designer with the given id
function renameDesignerWithID (db, designerID, newName)
    db:execSQL ("UPDATE DesignerList SET designerName = ? where designerID = ?" , newName, designerID)
end

---------------------------------------------------------------------------

-- Rename the publisher with the given id
function renamePublisherWithID (db, publisherID, newName)
    db:execSQL ("UPDATE PublisherList SET publisherName = ? where publisherID = ?" , newName, publisherID)
end

---------------------------------------------------------------------------

-- Change the primary category of the given game.
function alterGamePrimaryCategory (db, gameID, categoryID)
    db:execSQL ("UPDATE GameList SET categoryID = ? WHERE gameID = ?" , categoryID , gameID)
end

---------------------------------------------------------------------------

-- Change the primary designer of the given game.
function alterGamePrimaryDesigner (db, gameID, designerID)
    db:execSQL ("UPDATE GameList SET designerID = ? WHERE gameID = ?" , designerID , gameID)
end

---------------------------------------------------------------------------

-- Change the primary publisher of the given game.
function alterGamePrimaryPublisher (db, gameID, publisherID)
    db:execSQL ("UPDATE GameList SET publisherID = ? WHERE gameID = ?" , publisherID , gameID)
end

---------------------------------------------------------------------------

-- Change the nickname and password for the given player ID
function changeNickPassForPlayerID (db, playerID, nickname, password)
    db:execSQL ("UPDATE PlayerList SET nickname = ?, password = ? where playerID = ?", nickname, password, playerID)
end

---------------------------------------------------------------------------

-- Change the non-login fields for a player.
function changePlayerProfileForPlayerID (db, playerID, name, email, homepage)
    db:execSQL ("UPDATE PlayerList SET name = ?, emailAddress = ?, homepage = ? WHERE playerID = ?" ,
       name, email, homepage, playerID)
end

---------------------------------------------------------------------------

-- Change the properties for a player
function changePlayerPropertiesForPlayerID (db, playerID, canEditCategory, canAddGame, canEditGame, canCreatePlayer, canViewLog, canDeletePlayer, canEditPlayer, canConfigSystem, canEditLog, canRawDB, canEditDesigner , canEditPublisher, canAddSession, canEditSession)
    db:execSQL ("UPDATE PlayerList set canEditCategory = ?, canAddGame = ?, canEditGame = ?, canCreatePlayer = ?, canViewLog = ?, canDeletePlayer = ?, canEditPlayer = ?, canConfigSystem = ?, canEditLog = ?, canRawDB = ?, canEditDesigner = ?, canEditPublisher = ?, canAddSession = ?, canEditSession = ?  WHERE playerID = ?" , 
                canEditCategory, canAddGame, canEditGame, canCreatePlayer, canViewLog, 
                canDeletePlayer, canEditPlayer, canConfigSystem, canEditLog, 
                canRawDB, canEditDesigner , canEditPublisher, canAddSession, canEditSession, playerID)
end

---------------------------------------------------------------------------

-- Change the authToken associated with a particular sessionlist entry. This is
-- used when username/password changes on a user, to keep their login credentials
-- updated so that they're still logged in.
function switchSessionAuthTokens (db, oldToken, newToken)
    db:execSQL ("UPDATE SessionList SET authToken = ? where authToken = ?", newToken, oldToken)
end

---------------------------------------------------------------------------

-- This will scan all of the logs, and for each one that was entered by the given
-- playerID, the playerID will be changed to the name provided. This is for when
-- a user is deleted, but still has some actions in the log.
function renameLogEntryWhois (db, playerID, name)
    db:execSQL ("UPDATE Logs SET playerID = ? WHERE playerID = ?", name , playerID)
end

---------------------------------------------------------------------------

-- This will find all SessionReportEntries for the given player ID, and for
-- each, replace their playerID with the name provided.
function renameSessionReportEntryWhois (db, playerID, name)
    db:execSQL ("UPDATE SessionReportEntry SET playerID = ? WHERE playerID = ?", name, playerID)
end

---------------------------------------------------------------------------

-- Change the properties for a game
function changeGamePropertiesForGameID (db, gameID, gameBGGID, gameName, gameDescription, primaryDesigner, primaryPublisher, primaryCategory, gameMinPlay, gameMaxPlay, gamePlayTime, gameExpandsID)
    db:execSQL ("UPDATE GameList SET gamegeekID = ?, name = ?, description = ?, designerID = ?, publisherID = ?, categoryID = ?, minPlayers = ?, maxPlayers = ?, playTimeMinutes = ?, expandsGameID = ? WHERE gameID = ?" ,
                 gameBGGID, gameName, gameDescription, primaryDesigner, primaryPublisher, primaryCategory, gameMinPlay, gameMaxPlay, gamePlayTime, gameExpandsID, gameID)
end

---------------------------------------------------------------------------

-- Take a playerID and return back information on the games that that player
-- has won/lost/played
--
-- You get:
--   SessionID, sessionDate and gameID from SessionReport,
--   The complete information of a SessionReportEntry for this sessionID
--   The gameID and gameName from GameList
-- for EACH SessionReport entry for the player with the provided ID, in which
-- that player has won.
--
-- The fields timesWon, timesLost and timesPlayed will be inserted into each
-- record.
--
-- The results come back as an array sorted according to the name of the
-- associated game.
function gamesPlayedInformationForPlayerID (db, playerID)
    local results = {}

    -- Collect the primary data. This is a huge evil join type operation that pulls select fields
    -- from SessionReportEntry, SessionReport and GameList that are all unified by the sessionID
    -- in SessionReportEntry and SessionList, and the gameID in the SessionList and GameList.
    for row in db:SQLrows ("SELECT SessionReport.sessionID, SessionReport.sessionDate, SessionReport.gameID, " .. 
                                  "SessionReportEntry.*, " .. 
                                  "GameList.gameID, GameList.name " .. 
                                  "FROM SessionReport,SessionReportEntry,GameList " ..
                                  "WHERE SessionReport.sessionID = SessionReportEntry.sessionID and GameList.gameID = SessionReport.gameID and playerID = ?", playerID) do

        -- We're going to create this table using the gameID as the key and the table as
        -- the value. As we iterate the main result set, if the gameID doesn't exist in the temporary list, we copy 
        -- it over and set up some fields. Otherwise, we just update some fields.
        if results[row.gameID] == nil then
            row.timesWon = 0
            row.timesLost = 0
            results[row.gameID] = row
        end

        -- Count this as either a win or a loss. Use the current row to determine if this
        -- is a winner, but update the result in the previously generated entry.
        if string.upper (row.winner) == "YES" then
            results[row.gameID].timesWon = results[row.gameID].timesWon  + 1
        else
            results[row.gameID].timesLost = results[row.gameID].timesLost  + 1
        end
    end

    -- Now create the actual return value. Insert all of the items in the result set
    -- into this table as an array, and then sort it by name.
    local retVal = {}
    for key,value in results do
        table.insert (retVal, value)
    end

    -- Sort
    table.sort (retVal, sortByGameName)

    -- All done
    return retVal
end

---------------------------------------------------------------------------

-- Takes a playerID and returns back information on the most recent win
-- (wantWins == true) or loss (wantWins == false) for the player provided. 
-- 
-- The information that comes back is similar to that returned by
-- gamesPlayedInformationForPlayerID, except that it does not have the extra
-- fields that count wins/losses.
--
-- The return value will be a single table with the information, or nil if it could
-- not be found.
function getLatestWinLossForPlayerID (db, wantWins, playerID)
    -- Based on the boolean passed in, set the value to be the proper YES/NO string.
    if wantWins then wantWins = "YES" else wantWins = "NO" end

    -- This query works similar to the one in gamesPlayedInformationForPlayerID, except here
    -- we get the first first, and we specify either win or loss. We also make sure that the
    -- session report is ordered by sessionDate, so that we get the most recent value.
    return db:SQLfirstRow ("SELECT SessionReport.sessionID, SessionReport.sessionDate, SessionReport.gameID, " .. 
                           "SessionReportEntry.*, " .. 
                           "GameList.gameID, GameList.name " .. 
                           "FROM SessionReport,SessionReportEntry,GameList " ..
                           "WHERE SessionReport.sessionID = SessionReportEntry.sessionID and GameList.gameID = SessionReport.gameID and SessionReportEntry.winner = ? and playerID = ? ORDER BY SessionReport.sessionDate DESC", wantWins, playerID) 
end

---------------------------------------------------------------------------

-- Given a playerID, return back a list of all session reports in which that player
-- is listed as a winner, ordered by the date of the session.
function getWinningSessionsForPlayer (db, playerID)
    local retVal = {}

    -- Select all session reports where this player is listed as a winner.
    local SQL = [[SELECT * from SessionReport,SessionReportEntry 
                  WHERE SessionReport.sessionID = SessionReportEntry.sessionID 
                      AND SessionReportEntry.playerID = ? 
                      AND SessionReportEntry.winner = "YES" 
                      ORDER by SessionReport.sessionDate DESC]]

    for row in db:SQLrows (SQL, playerID) do
        table.insert (retVal , row)
    end

    return retVal
end

---------------------------------------------------------------------------

-- Given a playerID, return back a list of all session reports in which that player
-- is not listed as a winner, ordered by the date of the session.
function getLosingSessionsForPlayer (db, playerID)
    local retVal = {}

    -- Select all session reports where this player is listed as a winner.
    local SQL = [[SELECT * from SessionReport,SessionReportEntry 
                  WHERE SessionReport.sessionID = SessionReportEntry.sessionID 
                      AND SessionReportEntry.playerID = ? 
                      AND SessionReportEntry.winner = "NO" 
                      ORDER by SessionReport.sessionDate DESC]]

    for row in db:SQLrows (SQL, playerID) do
        table.insert (retVal , row)
    end

    return retVal
end

---------------------------------------------------------------------------

