---------------------------------------------------------------------------
--- LOCAL VARS ------------------------------------------------------------
---------------------------------------------------------------------------

-- What section to use in page headers in this file.
local thisPageSection = "SESSION REPORTS"                         

-- Error messages
local ERR_SESS_NOGAMEID = [[An error has occured. I don't know what game you're trying to add a session for. Please try again.]]
local ERR_SESS_NOPLAYERID = [[An error has occured. I don't know what player you are trying to view sessions for. Please try again.]]
local ERR_SESS_BADDATE = [[The session report date entered is not valid.]]
local ERR_SESS_NOPLAYERS = [[Your session report does not list any players.]]
local ERR_SESS_NOSESSID = [[I was unable to get session report information for the report you requested. Please try again.]]
local ERR_SESS_NOGAMEID = [[I was unable to locate the game listed in this session report. Please try again.]]
local ERR_SESS_NOADDAUTH = [[You are not allowed to add new session reports to the system. Please check with your administrator.]]
local ERR_SESS_NODELAUTH = [[You are not allowed to delete session reports from the system. Please check with your administrator.]]
local ERR_SESS_NOEDITAUTH = [[You are not allowed to edit session reports. Please check with your administrator.]]
local ERR_DEL_NOSESSIONS = [[You did not select any session reports to delete.]]
local ERR_DEL_NOFINDSESSION = [[Unable to find the session report to delete. Possibly deleted from another session.]]

---------------------------------------------------------------------------
--- FUNCTIONS -------------------------------------------------------------
---------------------------------------------------------------------------

-- This takes an array of session report information and returns back the
-- index in the array for the session with the provided sessionID. If the
-- sessionID is not found, then nil is returned instead.
local indexOfSession = function (sessionList, sessionID)
    -- Convert the desired ID to a number to start.
    sessionID = tonumber (sessionID)

    -- Loop over all entries. If we find one with our sessionID, then
    -- return it's index.
    for index,session in sessionList do
        if tonumber (session.sessionID) == sessionID then return index end
    end

    -- Failure.
    return nil
end

---------------------------------------------------------------------------

-- This generates a control bar that allows for various operations on the
-- current session report. Navigation links allow for browsing of sessions,
-- while other links allow for switching between viewing/editing the session
-- and viewing and adding other sessions for the same game.
local generateSessionControlBar = function (db, sessionID, isEditPage)
    -- Get a complete session list. If it ends up empty or with only a
    -- single entry, then pretend that it's ni.
    local sessionList = getSessionList (db) or {}
    if table.getn (sessionList) < 2 then sessionList = nil end

    -- Open the bar
    html:addText ('<FONT SIZE="-2">')

    -- First thing is either edit/view, depending on if isEditPage is true or not.
    if isEditPage then
        wrappedLink (urls.sessionDetails .. sessionID , "View")
    else
        -- Only generate the edit link if this user has the permissions to
        -- edit properties.
        if userInfo.canEditSession then
            wrappedLink (urls.sessionProperties .. sessionID , "Edit")
        end
    end

    -- The base URL for everything else depends on if this is an edit page
    -- or not. This is used by the navigation code so that next/previous take
    -- you to either the view or edit page of the next/previous entry.
    local baseURL = urls.sessionDetails
    if isEditPage then baseURL = urls.sessionProperties end

    -- Get information for the session report that we're generating this
    -- control bar for.
    local sessionInfo = getSessionReportByID (db, sessionID)

    -- If we have a session list then we should add some naviation links
    -- for the reports before/after us globally.
    if sessionList then
        -- Figure out the index of the current session, if possible.
        local myIndex = indexOfSession (sessionList , sessionID)
        if myIndex then
            -- If we're not the first item in the list, then generate a previous
            -- link.
            if myIndex > 1 then
                wrappedLink (baseURL .. sessionList[myIndex - 1].sessionID, "Previous Report")
            else
                wrappedLink (nil, "No Previous")
            end

            -- If we're not the last item in the list, then generate a next link.
            if myIndex + 1 <= table.getn (sessionList) then
                wrappedLink (baseURL .. sessionList[myIndex + 1].sessionID, "Next Report")
            else
                wrappedLink (nil, "No Next")
            end
        end
    end

    -- Now, if we have our own session information, we can put in some other links.
    if sessionInfo then
        -- If this user is allowed to add session reports, add in a link that allows
        -- for the addition of a new report using the same game as is being reported
        -- on right now.
        if userInfo.canAddSession then
            wrappedLink (urls.sessionAdd .. sessionInfo.gameID, "Add Another Report")
        end

        -- Add in a link to view all reports for this gameID
        wrappedLink (urls.gamesBySession .. sessionInfo.gameID, "View Other Reports")
    end

    -- Close the bar
    html:addText ("</FONT>")
end

---------------------------------------------------------------------------

-- This takes a date in form form (all fields broken up) and returns it
-- back as an epoch time suitable for insertion into the database and
-- so on.
local formDateToOsTime = function (args)
    local dateInfo = {}

    -- The year, month, and day fields are easily copied over.
    dateInfo.year  = args.sessionYear
    dateInfo.month = args.sessionMonth
    dateInfo.day   = args.sessionDay

    -- Fill in minutes and seconds, they're easy.
    dateInfo.min = args.sessionMinute
    dateInfo.sec = 0

    -- Fill in the hour now. Now this works depends on if we're AM or PM.
    -- This is true when we're AM, false for pm.
    if string.upper (args.sessionAmPm) == "AM"  then
        -- AM. The hour provided is the proper hour, except for 12, which should
        -- be 0.
        dateInfo.hour = args.sessionHour
        if dateInfo.hour == 12 then dateInfo.hour = 0 end
    else
        -- PM. The hour provided should have 12 added to it if it's
        -- not already 12.
        dateInfo.hour = args.sessionHour
        if dateInfo.hour ~= 12 then dateInfo.hour = dateInfo.hour + 12 end
    end

    -- Now return back the structure converted to a time.
    return os.mktime (dateInfo.year, dateInfo.month, dateInfo.day, dateInfo.hour, dateInfo.min)
end

---------------------------------------------------------------------------

-- Handles a form submission for adding a new session report, after having
-- args (which are form arguments) validated and modified via the form
-- validator for new session reports. This will add in the session report,
-- and then provide some success text.
--
-- The return value is the sessionID of the newly added session report.
local addSessionReport = function (db, args, gameInfo)
    -- Convert from form date to a regular epoch date.
    local sessionDate     = formDateToOsTime (args)
    local sessionDuration = fromTimeValue (args.sessionDuration)

    -- Add in the core of the session report and store it's ID.
    local sessionID = addSessionReport (db, gameInfo.gameID, sessionDate, sessionDuration, args.sessionNotes)

    -- Now, loop over all of the player information, and for everything valid,
    -- add an entry for a player to SessionReportEntry.
    for thisPlayerID = 1, tonumber (gameInfo.maxPlayers) do
        -- Generate form fields.
        local nameField   = "sessionPlayerID"     .. thisPlayerID
        local scoreField  = "sessionPlayerScore"  .. thisPlayerID
        local winnerField = "sessionPlayerWinner" .. thisPlayerID

        -- For each valid player record, insert an entry in the session report entry list.
        if args[nameField] ~= -1 then
            addSessionReportPlayer (db, sessionID, args[nameField], args[scoreField], args[winnerField])
        end
    end

    -- Indicate success
    html:addText (successText ("Added a new session report for '" .. gameInfo.name .. "'"), "<P>")
    addLog (db, "Added a new session report for '" .. gameInfo.name .. "'")

    -- Return back the sessionID of the new session report.
    return sessionID
end

---------------------------------------------------------------------------

-- This takes form arguments for a session report property submission that
-- has been validated, along with the game information that is associated 
-- with the report, and updates the properties of an already existing
-- session.
local updateSessionProperties = function (db, args, gameInfo)
    -- Get the date for this session, converted  back into a single
    -- epoch time from the many disparate date arguments used in the form.
    local sessionDate     = formDateToOsTime (args)
    local sessionDuration = fromTimeValue (args.sessionDuration)

    -- Update the core session report, keeping the same sessionID
    updateSessionReport (db, args.sessionID, gameInfo.gameID, sessionDate, sessionDuration, args.sessionNotes)

    -- Delete all SessionReportEntry items for this session report, so we 
    -- can add them back new.
    deleteSessionReportEntriesForSessionID (db, args.sessionID)

    -- Now, loop over all of the player information, and for everything valid,
    -- add an entry for a player to SessionReportEntry.
    for thisPlayerID = 1, tonumber (gameInfo.maxPlayers) do
        -- Make the form fields.
        local nameField   = "sessionPlayerID"     .. thisPlayerID
        local scoreField  = "sessionPlayerScore"  .. thisPlayerID
        local winnerField = "sessionPlayerWinner" .. thisPlayerID

        -- If this player ID is valid, then insert a new record for them.
        -- Is this player a valid one?
        if args[nameField] ~= -1 then
            addSessionReportPlayer (db, args.sessionID, args[nameField], args[scoreField], args[winnerField])
        end
    end

    -- We're done.
    html:addText (successText ("Modified properties of Session Report for  '" .. gameInfo.name .. "'"), "<P>")
    addLog (db, "Modified properties of Session Report for  '" .. gameInfo.name .. "'")
end

---------------------------------------------------------------------------

-- Given a sessionID, and some game information that goes with it, fill out
-- and return a table that has the properties of the session with the given
-- ID in the proper form arguments, so that they can be used to generate a
-- form.
--
-- If gameInfo is nil, it will be filled in from the gameID in the session
-- provided.
--
-- Normally the return value is sessionInfo, gameInfo
--
-- In case of error, the return values are nil, errorReason
getSessionPropertiesAsForm = function (db, sessionID, gameInfo)
    -- Set up the return value
    local retVal = {}

    -- Start by putting our sessionID into the form arguments. This is
    -- what will allow the form generator to know what sort of form to
    -- generate.
    retVal.sessionID = sessionID

    -- Get the session report information for this session.
    -- If we can't do that, then return an error saying so.
    local sessionInfo = getSessionReportByID (db, sessionID)
    if sessionInfo == nil then return nil, ERR_SESS_NOSESSID end

    -- If we don't have any gameInfo, then fetch it now using the
    -- gameID that's in the session information. 
    if gameInfo == nil then
        -- If this fails, then we have to report an error.
        gameInfo = getGameByID (db, sessionInfo.gameID)
        if gameInfo == nil then return nil, ERR_SESS_NOGAMEID  end
    end

    -- Fill out the base easy fields.
    retVal.gameID          = sessionInfo.gameID
    retVal.sessionDuration = sessionInfo.sessionDuration
    retVal.sessionNotes    = sessionInfo.notes

    -- Convert the session date from the databae into a table.
    local sessionDate = os.date ("*t", sessionInfo.sessionDate)

    -- Copy the easy fields over. We'll need to noodle the hour seperately.
    retVal.sessionYear    = sessionDate.year
    retVal.sessionMonth   = sessionDate.month
    retVal.sessionDay     = sessionDate.day
    retVal.sessionMinute  = sessionDate.min

    -- Jigger the date a bit. The date we get is in 24 hour time.
    -- If the hour is 13+, then subtract 12 to get the real hour, 
    -- and this is pm.
    --
    -- For other hours, the hour is correct, except that hour 0 means midnight.
    -- If the hour is 12, then it's pm, otherwise the hour is always am.
    if sessionDate.hour > 12 then
        retVal.sessionHour = sessionDate.hour - 12
        retVal.sessionAmPm = "pm"
    else
        if sessionDate.hour == 12 then retVal.sessionAmPm = "pm" else retVal.sessionAmPm = "am" end
        if sessionDate.hour == 0 then retVal.sessionHour = 12 else retVal.sessionHour = sessionDate.hour end
    end

    -- Get the list of participants in this session
    local entries = getPlayersForSessionID (db, sessionID)

    -- Loop over all of the players for this game. 
    for thisPlayerID = 1, tonumber (gameInfo.maxPlayers) do
        -- Create the form fields that go with this information.
        local nameField   = "sessionPlayerID"     .. thisPlayerID
        local scoreField  = "sessionPlayerScore"  .. thisPlayerID
        local winnerField = "sessionPlayerWinner" .. thisPlayerID

        -- Try to get an entry for the player with this ID. If this
        -- works, then we will fill out the form elements with their
        -- information. Otherwise, we'll fill it out with default information.
        local playerInfo = entries[thisPlayerID]
        if playerInfo then
            -- Set in the playerID and score data. If they're flagged as being a
            -- winner, then set that field in as well.
            retVal[nameField]  = playerInfo.playerID
            retVal[scoreField] = playerInfo.score

            if playerInfo.winner == "YES" then
                retVal[winnerField] = "YES"
            end
        else
            -- Set in an ID of -1 and a score of 0. Leave the winner field alone,
            -- since specifying it makes the checkbox check.
            retVal[nameField] = "-1"
            retVal[scoreField] = 0
        end
    end

    -- Return now
    return retVal, gameInfo
end

---------------------------------------------------------------------------

-- This takes a list of games, and the ID of the one to select, and generates
-- a dropdown from them.
local genGameDropdown = function (items, userInfo)
    -- Figure out what game to select. This will be nil (hence -1) if nothing
    -- is selected yet.
    local selectedGame = tonumber (userInfo) or -1

    -- Loop over all games.
    for index,info in items do
        local thisID = tonumber (info.gameID)
        html.form:selectOption (info.name, thisID, thisID == selectedGame)
    end
end

---------------------------------------------------------------------------

-- This helper is used by generateDateFields to generate all of it's
-- dropdown date fields. Itrms should be a table with keys "name" and "value".
--
-- userInfo should be the value to select (always a number).
local genDateDropdown = function (items, userInfo)
    -- Convert the userInfo to a number once.
    userInfo = tonumber (userInfo)

    -- Loop over all items in the list and generate a select option for them.
    for index,info in items do
        html.form:selectOption (info.name, info.value, tonumber (info.value) == userInfo)
    end
end

---------------------------------------------------------------------------

-- This works similar to genDateDropdown, except that this one uses values
-- that are textual, so the userInfo passed in should be textual as well.
local genDateDropdownText = function (items, userInfo)
    -- Loop over all items in the list and generate a select option for them.
    for index,info in items do
        html.form:selectOption (info.name, info.value, info.value == userInfo)
    end
end

---------------------------------------------------------------------------

-- Generate a whole bunch of dropdown fields that allow for the selection of
-- a date for this session report to happen on. 
--
-- The information in the args (if any) is used as default values. The 
-- information from the current time will be used for any missing arguments.
local generateDateFields = function (args)
    -- Find out the current time, for use as default arguments.
    local now = os.date ("*t")

    -- The current time as returned by os.date is in 24 hour format,
    -- but 12 hour format is more cool. 
    --
    -- Convert from 24 hour to 12 hour time, and insert a bogus field
    -- "ampm" into the now time that is either "am" or "pm", depending
    -- on the current hour.
    if now.hour > 12 then
        -- Subtract 12 from the hour, the rest is ok.
        now.hour = now.hour - 12
        now.ampm = "pm"
    else
        -- If the hour is 12, then this is a pm time. Anything else is am.
        -- If the hour is 0, then it's 12 (midnight). Order here is important.
        if now.hour == 12 then now.ampm = "pm" else now.ampm = "am" end
        if now.hour == 0 then now.hour = 12 end
    end

    -- Generate information on years in the dropdown. A 10 year span should
    -- be enough to start with.
    local yearList = {}
    for year = now.year - 5,now.year + 5 do table.insert (yearList, {["name"] = year, ["value"] = year}) end

    -- Months, displayed by name, indexed by number
    local monthNames = {"January","February","March","April","May","June","July","August","September","October","November","December"}
    local monthList = {}
    for month = 1,12 do table.insert (monthList, {["name"] = monthNames[month], ["value"] = month}) end

    -- Days, from 1 to 31. Validation happens on the receiving end.
    local dayList = {}
    for day = 1,31 do table.insert (dayList, {["name"] = day, ["value"] = day}) end

    -- Hours, from 1 to 12.
    local hourList = {}
    for hour = 1,12 do table.insert (hourList, {["name"] = hour, ["value"] = hour}) end

    -- Minutes, from 0 to 59
    local minuteList = {}
    for minute = 0,59 do table.insert (minuteList, {["name"] = string.format ("%02d", minute) , ["value"] = minute}) end

    -- am/pm list
    local ampmList = {{["name"] = "am", ["value"] = "am"}, {["name"] = "pm", ["value"] = "pm"}}

    -- Generate the dropdown lists now, in order.
    html.form:select ("sessionMonth", 1, nil, nil, nil, genDateDropdown, monthList, args.sessionMonth or now.month)
    html:addText ("&nbsp;")
    html.form:select ("sessionDay", 1, nil, nil, nil, genDateDropdown, dayList, args.sessionDay or now.day)
    html:addText ("&nbsp;")
    html.form:select ("sessionYear", 1, nil, nil, nil, genDateDropdown, yearList, args.sessionYear or now.year)
    html:addText ("&nbsp;")
    html.form:select ("sessionHour", 1, nil, nil, nil, genDateDropdown, hourList, args.sessionHour or now.hour)
    html:addText ("&nbsp;")
    html.form:select ("sessionMinute", 1, nil, nil, nil, genDateDropdown, minuteList, args.sessionMinute or now.min)
    html:addText ("&nbsp;")
    html.form:select ("sessionAmPm", 1, nil, nil, nil, genDateDropdownText, ampmList, args.sessionAmPm or now.ampm)
end

---------------------------------------------------------------------------

-- This generates a dropdown list of player names, and marks the user with
-- the id provided in userInfo as the current one.
--
-- This also inserts an extra user of "None" with an ID of "-1", so that we
-- can possibly select no player in this position.
local generatePlayerDropdown = function (items, userInfo)
    -- Convert the userInfo to a number once to save time.
    userInfo = tonumber (userInfo)

    -- Generate an option tag for the None user first, so that it's at the
    -- top.
    html.form:selectOption ("None" , "-1", userInfo == -1)

    -- Now scan over the rest of the players and insert them in as well,
    -- possibly selecting them if their ID is right.
    for index,info in items do
        html.form:selectOption (info.name, info.playerID, tonumber (info.playerID) == userInfo)
    end
end

---------------------------------------------------------------------------

-- This is called repeatedly by generateSessionForm to output a row of form
-- elements in the form that allow for choosing a player, the score of that
-- player, and wether or not they won the game.
--
-- The args table (which can be empty but never null) may contain the information
-- to use as default values in the form.
local generatePlayerRow = function (db, thisPlayerID, playerList, args)
    -- Determine the names for these form fields.
    local nameField   = "sessionPlayerID"     .. thisPlayerID
    local scoreField  = "sessionPlayerScore"  .. thisPlayerID
    local winnerField = "sessionPlayerWinner" .. thisPlayerID

    -- What player is this?
    html.table:heading (true , nil, nil, "LEFT" , "TOP")
    html:addText ("Player:")
    html.table:cell (true, nil, nil, "LEFT" , "TOP")
    html.form:select (nameField, 1, nil, "gameDrop", nil, generatePlayerDropdown, playerList, args[nameField] or "-1")

    -- What is the score?
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Score:")
    html.table:cell (true, nil, nil, "LEFT", "TOP")
    html.form:text (scoreField , args[scoreField] or "0", 5, 10, "sessionText")

    -- Did they win?
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Winner:")
    html.table:cell (true, nil, nil, "LEFT", "TOP")
    html.form:checkbox (winnerField, "YES", args[winnerField])
    
    -- Start a new row for the rest of the stuff
    html.table:newRow ()
end

---------------------------------------------------------------------------

-- Generate a form that allows for the addition of a new session report
-- into the system, or for the properties editing of an existing session
-- report.
--
-- If args is nil, then the form generates a simple dropdown of games and
-- gets you to select one. That form selection will come back to us again
-- but with a gameID, which allows for a more complete form.
--
-- If the arguments contain a sessionID, then the form generated will be
-- a property editing form instead of an addition form. The only difference
-- is in what command gets reported back, and what the button label says.
--
-- gameInfo should be a database record for the game that will be associted
-- with the session report being added. If the args doesn't list a gameID
-- then this can be nil, since it won't be used in that case.
local generateSessionForm = function (db, args, gameInfo)
    -- Make sure args is always a table, in case we get called with a nil.
    args = args or {}

    -- Start the form that we're generating.
    html.form:open (env.scriptName , "POST")

    -- If the arguments have a sessionID attached, then insert a command
    -- for session property editing, and make sure that we know what our
    -- sessionID is. 
    --
    -- Failing that, we just need to put in a command for the add portion
    -- of the form instead.
    if args.sessionID then
        html.form:hidden ("command", "SessionProperties")
        html.form:hidden ("sessionID", args.sessionID)
        html.form:hidden ("submitted", "yes")
    else
        html.form:hidden ("command", "SessionAdd")
    end

    -- Now we can start the table that the form will be in.
    html.table:open (nil, nil, nil, nil, "5", "sessionList")

    -- Get the total list of games
    local gameList = getGameList (db, true)

    -- If we have a sessionID and a gameID, then we can display a control bar.
    -- We need both, because the control bar contains links to do stuff with this
    -- session report, and with the associated game. Without both, it doesn't work.
    --
    -- This basically only gets output when a properties form is in effect.
    if args.gameID and args.sessionID then
        html.table:heading (true, nil, nil, "LEFT", "TOP", nil, "6")
        generateSessionControlBar (db, args.sessionID, true)
        html.table:newRow ()
    end

    -- Generate a dropdown for the game that is being session reported on. If this is
    -- a first stage add form, then the first game will be selected by defaut. Otherwise,
    -- the previously picked game will be used.
    html.table:heading (true , nil, nil, "LEFT" , "TOP")
    html:addText ("Game:")
    html.table:cell (true, nil, nil, "LEFT" , "TOP", nil, nil, nil, "5")
    html.form:select ("gameID", 1, nil, "gameDrop", nil, genGameDropdown, gameList, args.gameID)

    -- If there is a gameID, then go ahead and generate a link that lets you view the
    -- game that the session report is currently associated with. This doesn't track to
    -- changes in the dropdown list, but does provide information as to what game was
    -- initially set in the session report.
    if args.gameID then
        html:addText ("&nbsp; (View ")
        html:link (urls.gameDetails .. args.gameID, gameInfo.name or "Unknown")
        html:addText (")")
    end

    html.table:newRow ()

    -- If there is no gameID in our args, then this is just a first stage form that
    -- lists the game, so finish up the form with a submit button and leave now.
    if args.gameID == nil then
        html.table:cell (true, nil, nil, "CENTER", "TOP", nil, nil, nil, "6")
        html.form:submit ("Button", "Select Game for Session Report", "session", nil)
        return 
    end

    -- Now output date fields for the report. The date field generator will
    -- use the date from the args, if present, or the current time otherwise.
    -- The date in the args will be from a previous failed submission, or from
    -- the database, depending on what's been going on with the submission.
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Date:")
    html.table:cell (true, nil, nil, "LEFT", "TOP", nil, nil, nil, "3")
    generateDateFields (args)
    html.table:heading (true, nil, nil, "LEFT", "TOP", nil, nil, nil)
    html:addText ("Duration:")
    html.table:cell (true, nil, nil, "LEFT", "TOP")
    html.form:text ("sessionDuration", toTimeValue (args.sessionDuration), 5, 10)
    html.table:newRow ()

    -- Now, we need to generate a list of items for the number of players in the
    -- game, so we can specify who played, their score, and if they won or not.
    -- We only need to show as many items are there are players listed as maximum
    -- for the game. We can get that from the gameInfo. Note that when a game is added
    -- or modified, the min/max players are checked, and inverted if the min is bigger
    -- than the max, so that max players is always exactly the right number.
    local playerList = getPlayerList (db, true)
    for playerCount = 1,tonumber (gameInfo.maxPlayers) do
        generatePlayerRow (db, playerCount, playerList, args)
    end

    -- Set up to allow the entering of session notes for this report.
    html.table:heading (true , nil, nil, "LEFT" , "TOP")
    html:addText ("Notes:")
    html.table:cell (true, nil, nil, "LEFT" , "TOP", nil, nil, nil, "5")
    html.form:textArea ("sessionNotes" , args.sessionNotes , "80" , "10", nil, "sessionTextArea")
    html.table:newRow ()

    -- Finish the form. The text for the submit button depends on what sort of form
    -- we're generating. If there is a sessionID, then this is a properties form.
    -- Otherwise, it's an add form.
    html.table:cell (true, nil, nil, "CENTER", "TOP", nil, nil, nil, "6")
    if args.sessionID then
        html.form:submit ("Button", "Change Session Properties", "session", nil)
    else
        html.form:submit ("Button", "Add Session Report", "session", nil)
    end

    -- All done
    html.table:close ()
    html.form:close ()
    return 
end

---------------------------------------------------------------------------

-- Given a record of game information, return back a table of form
-- fields which must exist and which may exist for a form submission on
-- a session report using this game to be considered valid.
--
-- This is needed because the number of fields for players in session reports
-- is controlled by how many players the game supports.
local formArgsNeeded = function (db, gameInfo)
    -- This is the list of for sure needed arguments.
    local retVal = {"gameID", "sessionYear", "sessionMonth", "sessionDay",
                    "sessionHour", "sessionMinute", "sessionAmPm", 
                    "sessionDuration", "sessionNotes"}
    local other = {}

    -- Add in per player form fields for the maximum number of players
    -- that this game supports.
    for count = 1, tonumber (gameInfo.maxPlayers) do
        table.insert (retVal, "sessionPlayerID"     .. count)
        table.insert (retVal, "sessionPlayerScore"  .. count)
        table.insert (other,  "sessionPlayerWinner" .. count)
    end

    -- Return them.
    return retVal, other
end

---------------------------------------------------------------------------

-- Given form arguments for an add form, validate them. Return nil on
-- success, and an error string on failure.
--
-- Success means that code should be able to process these arguments to
-- come up with a valid session report (so for example, there must be at
-- least one player listed).
--
-- In addition all numbers will be converted to numbers, etc.
local validateSessionForm = function (db, args, gameInfo)
    -- Convert the date parts into numbers.
    args.sessionYear     = tonumber (args.sessionYear)     or -1
    args.sessionMonth    = tonumber (args.sessionMonth)    or -1
    args.sessionDay      = tonumber (args.sessionDay)      or -1
    args.sessionHour     = tonumber (args.sessionHour)     or -1
    args.sessionMinute   = tonumber (args.sessionMinute)   or -1

    -- Trim up the whitespace on the notes.
    args.sessionNotes = string.trimleadingtrailing (args.sessionNotes)

    -- If the date or time is broken in some way, then complain.
    if args.sessionYear < 1 or args.sessionMonth < 1 or args.sessionDay < 1 or
       args.sessionHour < 0 or args.sessionMinute < 0 then
        return ERR_SESS_BADDATE
    end

    -- Make sure that the date is valid (that the day goes with the month
    -- in the given year, etc).
    if isValidDate (args.sessionYear, args.sessionMonth, args.sessionDay) == false then
        return ERR_SESS_BADDATE
    end

    -- Loop over all of the players. We want to first of all convert their
    -- ID's and scores into numbers, and the winner flag turned into a true boolean.
    --
    -- Along the way, we will keep track of how many player ID's are not -1, so
    -- that at the end we know how many real players have been entered.
    local totalPlayersListed = 0
    for thisPlayerID = 1, tonumber (gameInfo.maxPlayers) do
        local nameField   = "sessionPlayerID"     .. thisPlayerID
        local scoreField  = "sessionPlayerScore"  .. thisPlayerID
        local winnerField = "sessionPlayerWinner" .. thisPlayerID

        -- Fix ID and score to be numbers.
        args[nameField]  = tonumber (args[nameField])  or -1
        args[scoreField] = tonumber (args[scoreField]) or -1

        -- Booleanize. We're going to assume that any value means that
        -- we should be checked, since checkboxes that aren't checked
        -- come across as nil in the form arguments.
        if args[winnerField] then args[winnerField] = true else args[winnerField] = false end

        -- Count this as a player listed if it's not -1
        if args[nameField] ~= -1 then totalPlayersListed = totalPlayersListed + 1 end
    end

    -- There has to have been at least one valid player in the list
    -- for us to say that the arguments are ok.
    if totalPlayersListed == 0 then
        return ERR_SESS_NOPLAYERS 
    end

    -- Must be ok
    return nil
end

---------------------------------------------------------------------------

-- Takes a list of session reports and displays them in a table.
function generateSessionList (db, sessionList)
    -- Say how many reports are in the list. We use some anal retentive
    -- logic to make sure that the text looks grammatically ok no matter
    -- how many session reports there actually are.
    -- Say how many games there are in the list.
    local label1 = "are "
    local label2 = " session reports"
    if table.getn (sessionList) == 1 then label1 = "is " label2 = " session report" end
    html:addText ('<FONT SIZE="-1">')
    html:addText ("There " , label1, table.getn (sessionList) , label2, " in this list.")
    html:addText ("</FONT><P>")

    -- Track if this user can delete or edit session reports
    local canDelete = userInfo.canEditSession
    local canEdit   = userInfo.canEditSession

    -- If the user has javascript enabled, this confirmation will be used to get them to
    -- confirm what they are about to do.
    local deleteConfirm = "return confirm ('Any selected session reports will be completely removed from the system." ..
                          " Are you sure this is what you want to do?');"

    -- Now display the list of session reports
    html.form:open (env.scriptName , "POST", nil, nil, deleteConfirm)

    -- Display the sessions in a table
    html.table:open (nil, nil, nil, nil, "5", "sessionList")

    -- Output some column headers for the session information. These are the
    -- columns that are always used, no matter who the current user is.
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("&nbsp;")
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Date")
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Duration")
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Game")
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Winner")

    -- Can the user edit properties? If so, include this column
    if canEdit then
        html.table:heading (true, nil, nil, "LEFT", "TOP")
        html:addText ("&nbsp;")
    end
    
    -- Can the user delete? If so, use this column.
    if canDelete then
        html.table:heading (true, nil, nil, "LEFT", "TOP")
        html:addText ("Delete")
    end

    -- Start the first actual row.
    html.table:newRow ()

    -- Loop over all sessions
    for index,row in sessionList do
        -- A details link.
        html.table:cell (true, nil, nil, "LEFT" , "TOP")
        html:link (urls.sessionDetails .. row.sessionID, "Details")

        -- Date of this session.
        html.table:cell (true, nil, nil, "LEFT" , "TOP")
        local sessionDate = os.date ("%m/%d/%Y", row.sessionDate)
        html:addText (sessionDate)

        -- Duration of this session
        html.table:cell (true, nil, nil, "LEFT", "TOP")
        if row.sessionDuration == 0 then
            html:addText ("N/A")
        else
            html:addText (toTimeValue (row.sessionDuration))
        end

        -- Game Name associated with this session. This will be a link,
        -- if we have game information.
        html.table:cell (true, nil, nil, "LEFT" , "TOP")
        local gameInfo = getGameByID (db, row.gameID) or {}
        local gameName = gameInfo.name or "Unknown"
        html:link (urls.gameDetails .. row.gameID, gameName)

        -- Who won? Get a list of all of the players registered
        -- for this game that are flagged as having won. The first
        -- one listed will be put as the winner, in case there are
        -- more than one.
        --
        -- Get the complete list of players. If there is a first
        -- entry, then get the information for that player. If that
        -- works, then use their name as the winner name. The default
        -- is to use "Unknown".
        html.table:cell (true, nil, nil, "LEFT" , "TOP")
        local players = getPlayersForSessionID (db, row.sessionID, true)
        local winnerName = "Unknown"
        local playerID = 0
        if players[1] then
            local playerInfo = getPlayerByID (db, players[1].playerID)
            if playerInfo then
                winnerName = playerInfo.name
                playerID = playerInfo.playerID
            end
        end
        if playerID ~= 0 then
            html:link (urls.sessionsByPlayerWin .. playerID , winnerName)
        else
            html:addText (winnerName)
        end

        -- If the user can edit properties, then include a
        -- properties link as well.
        if canEdit then
            html.table:cell (true, nil, nil, "LEFT" , "TOP")
            html:link (urls.sessionProperties .. row.sessionID , "Properties")
        end

        -- If the user can delete, then add the delete row.
        if canDelete then
            html.table:cell (true, nil, nil, "LEFT" , "TOP")
            html.form:checkbox ("delSessionID" , row.sessionID)
        end

        -- Get ready for the next row
        html.table:newRow ()
    end

    -- If the user can delete and there is at least one session in the list of
    -- session reports, then we need to include one last row with a submit button
    -- for the integrated delete form.
    if table.getn (sessionList) > 0 and canDelete then
        -- The table has at least 6 columns in it. It will be 7 if this user
        -- can edit properties, 6 otherwise.
        local span = "6";
        if canEdit then span = "7" end

        -- Create a cell spanning the table and put a submit button in it.
        html.table:cell (true, nil, nil, "CENTER", "TOP", nil, nil, nil, span)
        html.form:submit ("Button", "Delete selected Session Reports", "session")
        html.form:hidden ("command", "SessionDelete")
    end

    -- Close up the table and the form now
    html.table:close ()
    html.form:close ()
end

---------------------------------------------------------------------------

-- This handles the deletion of session reports from the system.
function commandSessionDelete (db)
    -- Header for the page
    generatePageHeader (thisPageSection, "Delete Session Reports")

    -- Let the user get back to the main page
    html.form:singleButtonForm ("Return to Session Report Page" , "Session" , env.scriptName , "POST" , nil, nil, "session")

    -- If this user isn't allowed to delete a session report, tell them that and then
    -- leave. This happens first so that even if the user tries to craft a URL
    -- that will delete a session, it will still stop them.
    if userInfo.canEditSession == false then
        addLog (db , "Attempted to delete a session report")
        html:addText (errorText (ERR_SESS_NODELAUTH) , "<P>")
        return
    end
    
    -- Get the form arguments, which will be a list of session report
    -- ID's to delete.
    local formArgs = collectFormArguments ({"delSessionID"})

    -- At this point we should have a list of ID's to delete. If we
    -- didn't get anything, then complain.
    if formArgs == nil then 
        addLog (db, "Attempted delete with no sessions selected");
        html:addText (errorText (ERR_DEL_NOSESSIONS) , "<P>");
        return
    end

    -- The form argument might be a string if only a single session is being
    -- deleted, or a table if there was more than one. Make sure this is always
    -- a table to make the code below easier to handle.
    if type (formArgs.delSessionID) ~= "table" then
        formArgs.delSessionID = {formArgs.delSessionID}
    end

    -- We got confirmation on deletion prior to getting here, when the browser asked if
    -- they were sure. So, loop over all sessions and remove them.
    for index,session in formArgs.delSessionID do
        -- Get information on the session to delete
        local delSessionInfo = getSessionReportByID (db, session)
   
        -- It's an error if we can't find the information on the session report to
        -- delete.
        if delSessionInfo == nil then
            html:addText (errorText (ERR_DEL_NOFINDSESSION) , "<BR>");
        else
            -- Get rid of this base session report and all of the report entries
            -- that go with it.
            deleteSessionReportByID (db, delSessionInfo.sessionID)
            deleteSessionReportEntriesForSessionID (db, delSessionInfo.sessionID)

            -- Try to get the name of the game that this report was for, for
            -- logging purposes.
            local gameInfo = getGameByID (db, delSessionInfo.gameID) or {}
            local gameName = gameInfo.name or "Unknown"
            
            -- We're done.
            html:addText (successText ("Deleted session report for game '" .. gameName .. "'") , "<BR>")
            addLog (db , "Deleted session report for game '" .. gameName .. "'")
        end
    end
end

---------------------------------------------------------------------------

-- This finds the list of session reports that associte with a particular
-- game ID and then displays them.
--
-- Just displays complete session reporting lists, if no gameID is given.
function commandSessionViewGame (db)
    -- Try to find a gameID. If we find one, then try to find the information
    -- for it.
    local gameInfo = nil
    local args = collectFormArguments ({"gameID"})
    if args then
        gameInfo = getGameByID (db, args.gameID)
    end

    -- Generate a header for the page, which includes the game name, if we were
    -- able to find the game information.
    if gameInfo == nil then
        generatePageHeader (thisPageSection, "View Reports", "sessionsForGame")
    else
        generatePageHeader (thisPageSection, "View Reports for '" .. gameInfo.name .. "'", "sessionsForGame")
    end

    -- Generate a button to go back to the main session report page.
    html.form:singleButtonForm ("Return to Session Report Page" , "Session" , env.scriptName , "POST" , nil, nil, "session")
    html:addText ("<P>")

    -- If we have gameInfo, then we want to generate a list of session reports
    -- for only this game. If we don't have any gameInfo, then we need to 
    -- generate a list of global session reports 
    local sessionList = nil
    if gameInfo then
        sessionList = getSessionReportsForGameID (db, gameInfo.gameID)
    else
        sessionList = getSessionList (db)
    end

    -- Show the list now.
    generateSessionList (db, sessionList)
end

---------------------------------------------------------------------------

-- This displays a detailed version of database session report information.
function commandSessionDetails (db)
    -- The form arguments we were given should contain a sessionID that 
    -- tells us what session we should be showing the details of.
    local sessionInfo = nil
    local gameInfo = nil
    local args = collectFormArguments ({"sessionID"})

    -- We got the form arguments. In this caser, look up the session
    -- information that goes with this session. If that works, then
    -- also look up the information for the game that the session is
    -- for.
    if args then
        sessionInfo = getSessionReportByID (db, args.sessionID)
        if sessionInfo then
            gameInfo = getGameByID (db, sessionInfo.gameID)
        end
    end

    -- Now, we can generate the header for this page. If we managed to
    -- get game information, then we want to use a subsection that says what
    -- the game is. Otherwise, use a generic header.
    if gameInfo == nil then
        generatePageHeader (thisPageSection, "Detailed Session Report View", "sessionsView")
    else
        generatePageHeader (thisPageSection, "Detailed Session Report for '" .. gameInfo.name .. "'", "sessionsView")
    end

    -- We need session information, so if we didn't find that, complain and return.
    if sessionInfo == nil then
        html:addText (errorText (ERR_SESS_NOSESSID), "<P>")
        return
    end

    -- We got the session information. If we didn't also get the game information,
    -- that too is bad, so return.
    if gameInfo == nil then
        html:addText (errorText (ERR_SESS_NOGAMEID), "<P>")
        return
    end

    -- Generate a button to back to the session report page.
    html.form:singleButtonForm ("Return to Session Report Page" , "Session" , env.scriptName , "POST" , nil, nil, "session")
    html:addText ("<P>")

    -- Open a table for the session information.
    html.table:open (nil, "70%", nil, nil, "5", "sessionList")

    -- Output a session control bar for this session.
    html.table:heading (true, nil, nil, "LEFT", "TOP", nil, "4")
    generateSessionControlBar (db, sessionInfo.sessionID, false)
    html.table:newRow ()

    -- Display the date of the session report.
    html.table:heading (true, "10%", nil, "LEFT", "TOP")
    html:addText ("Session Date:")
    html.table:cell (true, nil, "10%", "LEFT", "TOP")
    local sessionDate = os.date ("%A %B %e, %Y @ %l:%M %P", sessionInfo.sessionDate)
    html:addText (sessionDate)

    -- Now display the name of the game. This will be a hotlink
    -- so that we can view the game in question if desired.
    html.table:heading (true, "10%", nil, "LEFT", "TOP")
    html:addText ("Game:")
    html.table:cell (true, "70%", nil, "LEFT", "TOP")
    html:link (urls.gameDetails .. sessionInfo.gameID, gameInfo.name or "Unknown")

    -- New row
    html.table:newRow ()

    -- Display the duration of this play, if this information is available. If no
    -- information was reported, then this would be set to 0, in which case we won't
    -- even bother trying.
    local sessionDuration = tonumber (sessionInfo.sessionDuration) or 0
    if sessionDuration ~= 0 then
        html.table:heading (true, "10%", nil, "LEFT", "TOP")
        html:addText ("Game Duration:")
        html.table:cell (true, "90%", nil, "LEFT", "TOP", nil, nil, nil, "3")
        html:addText (toTimeValue (sessionDuration))
    end

    -- New row
    html.table:newRow ()

    -- Now, get the list of players of this game.
    local players = getPlayersForSessionID (db, sessionInfo.sessionID)

    -- Do the following stuff twice, once for "YES", and once for "NO".
    -- That will put the winners first and the losers second in the view.
    for unused,playType in {"YES", "NO"} do
        -- Loop over all players.
        for index,playerInfo in players do
            -- Is this player of the type needed?
            if string.upper (playerInfo.winner) == playType then
                -- Yes, output some stuff for this player. Start by gettting
                -- the information on this player.
                local otherInfo = getPlayerByID (db, playerInfo.playerID) or {}
           
                -- Name of this player.
                html.table:heading (true, "10%", nil, "LEFT", "TOP")
                html:addText ("Player:")
                html.table:cell (true, "10%", nil, "LEFT", "TOP")
                html:addText (otherInfo.name or "Unknown")

                -- Now output the score and if they won or not
                html.table:heading (true, "10%", nil, "LEFT", "TOP")
                html:addText ("Score:")
                html.table:cell (true, "70%", nil, "LEFT", "TOP")
                html:addText (playerInfo.score)
    
                -- If we're doing the YES group, then this player is a winner in
                -- the game.
                if playType == "YES" then html:addText (" (WON)") else html:addText ("&nbsp;") end

                -- New row
                html.table:newRow ()
            end
        end
    end

    -- We will create a cell that will span all columns,
    -- and it will contain the description information for this report.
    html.table:cell (nil, nil, nil, "JUSTIFY" , "TOP", nil, nil, nil, "4") 
    local fixedNotes = string.gsub (sessionInfo.notes or "", "\n", "<p>")
    html:addText (fixedNotes)

    -- Close up the table, done now
    html.table:close ()
end

---------------------------------------------------------------------------

-- This handles the properties section of session reporting, which allows for
-- editing the contents of existing session reports.
function commandSessionProperties (db)
    -- Header for the page
    generatePageHeader (thisPageSection, "Edit Report", "sessionsEdit")

    -- Include a button that lets us go back to the main page.
    html.form:singleButtonForm ("Return to Session Report Page" , "Session" , env.scriptName , "POST" , nil, nil, "session")
    html:addText ("<P>")

    -- Stop right now and leave with an error if the current user can't edit session
    -- reports.
    if not userInfo.canEditSession then
        addLog (db, "Attempted to edit a session report")
        html:addText (errorText (ERR_SESS_NOEDITAUTH), "<P>")
        return
    end

    -- The absolute minimum requirement to get this page to do anything is to provide
    -- it the sessionID of the session report to edit via a form argument. Try to get
    -- that, and return with a failure if we can't.
    local args = collectFormArguments ({"sessionID"})
    if args == nil then
        html:addText (errorText (ERR_SESS_NOSESSID), "<P>")
        return
    end

    -- Using the sessionID we were given above, get the properties of this sessionID
    -- as fake form arguments. Since we pass in nil for gameInfo, the function will
    -- look it up for us. It is returned as the second return value. If the first return
    -- comes back nil, then the second return value is the reason for failure instead
    -- of the game information.
    --
    -- If we couldn't get the information, then explain why and leave.
    local fakeArgs,gameInfo = getSessionPropertiesAsForm (db, args.sessionID, nil)
    if fakeArgs == nil then
        html:addText (errorText (gameInfo), "<P>")
        return
    end

    -- Now, check our real arguments again, this time for all of the arguments that need
    -- to be there for a form submission, as well as any optional fields. Since we want
    -- to only use our arguments when the form has actually been submitted, force the
    -- variable "submitted" into the list of needed arguments. We also need to add sessionID,
    -- since the function doesn't add that for us (since that is unused when this function is
    -- called to get needed form arguments for new insertions).
    local needed, optional = formArgsNeeded (db, gameInfo)
    table.insert (needed, "sessionID")
    table.insert (needed, "submitted")

    -- Now, try to collect form arguments using these values.
    local formArgs = collectFormArguments (needed, optional)

    -- Now, if formArgs is not nil, then the form has been submitted. In that case,
    -- use those args as our form arguments for the code below, which will generate
    -- the form for us. Otherwise, use the fakeArgs. 
    --
    -- The idea is that the form arguments take precedence (they represent changed
    -- information), but if that hasn't happened, then use the real data.
    if formArgs then
        -- The form was submitted. First of all, use the form args instead of
        -- the fake ones (since they represent the state of the database and not
        -- the state of this edit).
        args = formArgs

        -- Now, verify the arguments, to see if things are ok. As usual, this returns nil
        -- on success and error reason on failure.
        local result = validateSessionForm (db, args, gameInfo)
        if result ~= nil then
            -- Not ok, the form is borked.
            html:addText (errorText (result), "<P>")
        else
            -- Form seems ok, so update properties for this session using these arguments.
            updateSessionProperties (db, args, gameInfo)

            -- Now that things are submitted, re-get the new changes from
            -- the database. This will set the form up with the current
            -- database contents.
            local errReason
            args, errReason = getSessionPropertiesAsForm (db, args.sessionID, gameInfo)
            if args == nil then
                html:addText (errorText (errReason), "<P>")
                return
            end
        end
    else
        -- The form wasn't submitted, use the fakes so that the form displays with
        -- the current database properties of things.
        args = fakeArgs
    end

    -- Generate a form now. 
    generateSessionForm (db, args, gameInfo)
end

---------------------------------------------------------------------------

-- This handles the addition of session reports to the system.
function commandSessionAdd (db)
    -- Header for the page
    generatePageHeader (thisPageSection, "Add Report", "sessionsAdd")

    -- Include a button to back to the main session report page.
    html.form:singleButtonForm ("Return to Session Report Page" , "Session" , env.scriptName , "POST" , nil, nil, "session")
    html:addText ("<P>")

    -- Stop right now if this user doesn't have permission to add session
    -- reports to the system.
    if not userInfo.canAddSession then
        addLog (db, "Attempted to add a session report")
        html:addText (errorText (ERR_SESS_NOADDAUTH), "<P>")
        return
    end

    -- Collect some form arguments, starting with just a gameID, which indicates
    -- what game the session report is for. If there isn't one, that's ok. 
    --
    -- gameInfo is always the information for the game that we're editing a session
    -- for. So, if there is no gameID, then gameInfo should be nil.
    local gameInfo = nil
    local args = collectFormArguments ({"gameID"})

    -- Did we get a gameID?
    if args then
        -- Try to get the information for the game that we're adding the session
        -- report for. We have to fail out if this doesn't work, because we need the
        -- game information to contiue.
        gameInfo = getGameByID (db, args.gameID)
        if gameInfo == nil then
            html:addText (errorText (ERR_SESS_NOGAMEID), "<P>")
            return 
        end

        -- Check to see if any other form arguments have been supplied. Start by getting
        -- the list of needed and optional form arguments for adding a report. This is
        -- needed because the number of fields for players is determined by the number of
        -- players the game supports.
        --
        -- The optional arguments are the checkboxes that flag the winners of games, since
        -- they will not appear in the submission if they're not checked.
        local needed, optional = formArgsNeeded (db, gameInfo)
        args = collectFormArguments (needed, optional)

        -- If we didn't get any arguments, the form was either submitted wrong (not enough
        -- fields, or the only part that's been submitted thus far is a gameID. When this
        -- happens, we just pretend that all we got was the gameID that we started with.
        if args == nil then
            args = {}
            args.gameID = gameInfo.gameID
        else
            -- We got form arguments. Validate the submission to see if it's correct.
            -- If this doesn't return nil, then the result is the reason why the
            -- submission has failed. The form arguments are left alone.
            local result = validateSessionForm (db, args, gameInfo)
            if result then
                -- The args didn't validate. If that's the case, then say why. We will let
                -- things continue on to re-generate the form.
                html:addText (errorText (result), "<P>")
            else
                -- The arguments were ok, so we can consider the form submitted. Add 
                -- a new session report, keeping track of it's ID. Next we fetch the
                -- contents of this session in the form of form arguments. This allows
                -- us to regenerate the form with the just populted data as seen in the
                -- database and have it be a property form instead of a new session report
                -- form.
                local errReason
                local sessionID = addSessionReport (db, args, gameInfo)
                args, errReason = getSessionPropertiesAsForm (db, sessionID, gameInfo)
                if args == nil then
                    html:addText (errorText (errReason), "<P>")
                    return
                end
            end
        end
    end

    -- Generate a session report form. If args is nil, then gameInfo is also nil,
    -- and what we get is a form that lists games and asks for us to pick the game
    -- to report on. When that form is submitted, we'll have a gameID, and then we
    -- get the real session form.
    --
    -- If a valid complete session report form was submitted successfully, then we
    -- replaced the submitted arguments with information from the database regarding
    -- the session, including a sessionID. This will make the form display as a property
    -- form instead of an addition form.
    generateSessionForm (db, args, gameInfo)
end

---------------------------------------------------------------------------

-- This finds the list of session reports where a particular player is
-- recorded as the winner and displays them.
function commandSessionViewWinsByPlayer (db)
    -- Try to find a playerID. If we find one, then try to find the information
    -- for it.
    local playerInfo = nil
    local args = collectFormArguments ({"playerID"})
    if args then
        playerInfo = getPlayerByID (db, args.playerID)
    end

    -- Generate a header for the page, which includes the player name, if we were
    -- able to find the player information.
    if playerInfo == nil then
        generatePageHeader (thisPageSection, "View Reports", "sessionsForPlayerWin")
    else
        generatePageHeader (thisPageSection, "View Reports where '" .. playerInfo.name .. "' won", "sessionsForPlayerWin")
    end
    
    -- Generate a button to go back to the main session report page.
    html.form:singleButtonForm ("Return to Session Report Page" , "Session" , env.scriptName , "POST" , nil, nil, "session")
    html:addText ("<P>")

    -- Bad if we can't get player information. 
    if playerInfo == nil then
        html:addText (errorText (ERR_SESS_NOPLAYERID), "<P>")
        return
    end

    -- Get the list of sessions where this player won
    local sessionList = nil
    if playerInfo then
        sessionList = getWinningSessionsForPlayer (db, args.playerID)
    end

    -- Show the list now.
    generateSessionList (db, sessionList)
end

---------------------------------------------------------------------------

-- This finds the list of session reports where a particular player is
-- recorded as a loser and displays them.
function commandSessionViewLossesByPlayer (db)
    -- Try to find a playerID. If we find one, then try to find the information
    -- for it.
    local playerInfo = nil
    local args = collectFormArguments ({"playerID"})
    if args then
        playerInfo = getPlayerByID (db, args.playerID)
    end

    -- Generate a header for the page, which includes the player name, if we were
    -- able to find the player information.
    if playerInfo == nil then
        generatePageHeader (thisPageSection, "View Reports", "sessionsForPlayerLoss")
    else
        generatePageHeader (thisPageSection, "View Reports where '" .. playerInfo.name .. "' lost", "sessionsForPlayerLoss")
    end
    
    -- Generate a button to go back to the main session report page.
    html.form:singleButtonForm ("Return to Session Report Page" , "Session" , env.scriptName , "POST" , nil, nil, "session")
    html:addText ("<P>")

    -- Bad if we can't get player information. 
    if playerInfo == nil then
        html:addText (errorText (ERR_SESS_NOPLAYERID), "<P>")
        return
    end

    -- Get the list of sessions where this player won
    local sessionList = nil
    if playerInfo then
        sessionList = getLosingSessionsForPlayer (db, args.playerID)
    end

    -- Show the list now.
    generateSessionList (db, sessionList)
end


---------------------------------------------------------------------------

-- This is the main function to view and edit session reports
function commandSession (db)
    -- Header for the page
    generatePageHeader (thisPageSection, nil, "sessionsBase")

    -- If the user has authorization, add a button that allows for the
    -- addition of a new session report.
    if userInfo.canAddSession then
        html.form:singleButtonForm ("Add Session Report"   , "SessionAdd" , env.scriptName , "POST" , nil, nil, "session")
    end

    -- Get the complete list of session reports, which are ordered
    -- according to date, and then display them.
    local sessionList = getSessionList (db)
    generateSessionList (db, sessionList)
end

---------------------------------------------------------------------------
--- MAIN CODE -------------------------------------------------------------
---------------------------------------------------------------------------

