-- This table should contain a list of SQL needed to create and initialize a table, with a key that is that
-- table's name. 
return {
    ["PlayerList"] = [[
-- Authors are the people who are allowed to work with the system. Each can be limited to
-- performing a set number of actions from all of those allowed. Each player also tracks
-- what player it was that created them
CREATE TABLE PlayerList (playerID INTEGER UNIQUE PRIMARY KEY, 
                         name TEXT, 
                         nickname TEXT, 
                         password TEXT, 
                         emailAddress TEXT, 
                         homepage TEXT, 
                         canEditCategory BOOLEAN,
                         canAddGame BOOLEAN,
                         canEditGame BOOLEAN,
                         canCreatePlayer BOOLEAN, 
                         canDeletePlayer BOOLEAN, 
                         canEditPlayer BOOLEAN,
                         canConfigSystem BOOLEAN,
                         canViewLog BOOLEAN, 
                         canEditLog BOOLEAN, 
                         canRawDB BOOLEAN, 
                         canEditDesigner BOOLEAN,
                         canEditPublisher BOOLEAN,
                         canAddSession BOOLEAN,
                         canEditSession BOOLEAN,
                         createdBy INTEGER);

-- We need a default player in the system who can do everything.
INSERT INTO PlayerList VALUES (NULL, "default player", "default", "password", "noone@nowhere.com", "http://www.homepage.com", "YES", "YES", "YES", "YES", "YES", "YES", "YES", "YES", "YES", "YES", "YES", "YES", "YES", "YES", 0);
]],

    ["SessionList"] = [[
-- This holds a list of currently active sessions in the system. authToken is
-- text created at login time, which is username:password crypted with a known
-- salt value, so that different browsers will generate the same auth token.
-- playerID is the numeric player id number of the player identified with this
-- session. lastAccessDate is the time that the last page was created/generated
-- for this session. 
--
-- At page generation time, before anything else occurs, all entries in this table
-- with a lastAccessDate more than X minutes in the past will be purged, to timeout
-- the user logged in.
CREATE TABLE SessionList (sessionID INTEGER UNIQUE PRIMARY KEY,
                          authToken TEXT,
                          playerID INTEGER,
                          lastAccessDate TIME);
]],

    ["ConfigInfo"] = [[
-- This table holds all of the configuration information for the system. The first column
-- is a configuration key, and the second is it's value. Both are text.
CREATE TABLE ConfigInfo (itemName TEXT UNIQUE PRIMARY KEY,
                         itemValue TEXT);

INSERT INTO ConfigInfo VALUES ("pageTitle" , "LUAweb Game Database");
INSERT INTO ConfigInfo VALUES ("loginTimeout" , "30");
INSERT INTO ConfigInfo VALUES ("defaultPage" , "Player")
]],

    ["Logs"] = [[
-- Logs are used to track what every user is doing and when they did it, also from what IP the
-- user was using at the time. eventTime is stored as an epoch time (integer).
--
-- playerID is 0 for system logs, some number for things the player with that playerID did, and
-- a textual string in cases where the user that generated the log has been deleted since the
-- log was created. The name used is the login name, in this case (PlayerList.nickname).
CREATE TABLE Logs (logID INTEGER UNIQUE PRIMARY KEY,
                   message TEXT,
                   ip TEXT,
                   playerID INTEGER,
                   eventTime TIME);

-- Create an initial log entry for when the database was initialized. Note that the time and IP here
-- will be wrong. We'll have to fill it in after the creation process finishes.
INSERT INTO Logs VALUES (NULL , "Database was initialized" , "127.0.0.1" , 0 , 0);
UPDATE Logs set ip = "172.36.1.15", eventTime = 2 where logID = 1;
]],

    ["PublisherList"] = [[
-- This table holds a list of game publishers, which are simple unique publisher
-- ID values with a textual name.
CREATE TABLE PublisherList (publisherID    INTEGER UNIQUE PRIMARY KEY,
                            publisherName  VARCHAR);  

-- Default to no publisher
INSERT INTO PublisherList VALUES (0 , "None");
]],

    ["DesignerList"] = [[
-- This table holds a list of game designers, which like publishers are unique
-- ID values bound to a textual name.
CREATE TABLE DesignerList (designerID INTEGER UNIQUE PRIMARY KEY,
                           designerName VARCHAR);

-- Default no designer
INSERT INTO DesignerList VALUES (0 , "None");
]],

    ["CategoryList"] = [[
-- This table holds a list of game categorizations, which like publishers and
-- desigbners, are unique ID values bound to a textual name.
CREATE TABLE CategoryList (categoryID INTEGER UNIQUE PRIMARY KEY,
                           categoryName VARCHAR);

-- Insert a default category
INSERT INTO CategoryList VALUES (0 , "None");
]],

    ["GameList"] = [[
-- This table holds a list of all games and their primary statistics.
-- Each game has a unique ID, and contains information about the game,
-- including the primary designer, publisher and category of the game. 
CREATE TABLE GameList (gameID INTEGER UNIQUE PRIMARY KEY,
                       gamegeekID INTEGER,
                       name VARCHAR,
                       description VARCHAR,
                       designerID INTEGER,
                       publisherID INTEGER,
                       categoryID INTEGER,
                       minPlayers INTEGER,
                       maxPlayers INTEGER,
                       playTimeMinutes INTEGER,
                       expandsGameID INTEGER);
]],

    ["PublisherPlacement"] = [[
-- This table is used when there needs to be more than one publisher assigned to
-- a game. Basically for any additional publishers, an entry is created here which
-- assignes a game ID to a publisher ID. 
CREATE TABLE PublisherPlacement (placementID  INTEGER UNIQUE PRIMARY KEY,
                                 gameID       INTEGER,
                                 publisherID  INTEGER);

]],

    ["DesignerPlacement"] = [[
-- Works like PublisherPlacement, but for designers. Again, for any designers besides
-- the primary designer, a placement is added that maps a designer ID to a game ID.
CREATE TABLE DesignerPlacement (placementID  INTEGER UNIQUE PRIMARY KEY,
                                gameID       INTEGER,
                                designerID   INTEGER);
]],

    ["CategoryPlacement"] = [[
-- Works like PublisherPlacement, but for categories. Again, for any categories besides
-- the primary one, a placement is added that maps a category ID to a game ID.
CREATE TABLE CategoryPlacement (placementID  INTEGER UNIQUE PRIMARY KEY,
                                gameID       INTEGER,
                                categoryID   INTEGER);
]],

    ["SessionReport"] = [[
-- This is the basis of a session report. Each session report has a unique
-- number to identify it, the date on which it occured, and the ID of the
-- game that was played.
CREATE TABLE SessionReport (sessionID INTEGER UNIQUE PRIMARY KEY,
                            sessionDate TIME,
                            sessionDuration TIME,
                            gameID INTEGER,
                            notes VARCHAR);
]],

    ["SessionReportEntry"] = [[
-- This is an entry into a session report. There is a record in this table for
-- each player in each game in each session report. Each entry uniquely identifies
-- itself, and contains the session report ID number that goes with this session
-- entry, the playerID of a player, their score, and wether or not they won.
--
-- playerID is either a number, to indicate an existing player, or a textual string
-- (PlayerList.name) if the player that used to be here has since been deleted.
CREATE TABLE SessionReportEntry (entryID INTEGER UNIQUE PRIMARY KEY,
                                 sessionID INTEGER,
                                 playerID INTEGER,
                                 score INTEGER,
                                 winner BOOLEAN);
]],
}
