---------------------------------------------------------------------------
--- LOCAL VARS ------------------------------------------------------------
---------------------------------------------------------------------------

-- What section to use in page headers in this file.
local thisPageSection = "TOOLS"                         

-- Error messages
local ERR_FINDER_BADPLAYERCOUNT = [[The number of players you have entered is not valid.]]
local ERR_FINDER_BADPLAYTIME = [[The play time you have entered is not valid.]]

---------------------------------------------------------------------------
--- FUNCTIONS -------------------------------------------------------------
---------------------------------------------------------------------------

-- Takes an array of numbers, and tells you if the value specified is 
-- contained in the array.
local numberArrayContains = function (list, value)
    -- If the list provided is nil, then the value isn't in it. No need to look.
    if list == nil then return false end

    -- Convert the value to a number.
    value = tonumber (value)

    -- Scan over the entire contents of the list.
    for index,content in list do
        -- If this item is the value we were given, then return true.
        if tonumber (content) == value then return true end
    end

    -- No good.
    return false
end

---------------------------------------------------------------------------

-- Generates a milti-select listbox. 
--
-- The listbox contains the value provided in list, an array of tables which 
-- should have fields nameField and valueField, telling the generator which 
-- field to use as the value of the item and which to use as the name. The 
-- value fields should be numbers.
--
-- The listbox will be named with the given fieldName.
--
-- Defaults is a list of items which should be selected by default. It is
-- expected to be an array of numbers.
--
-- If includeAny is true, then a bogus entry of "Any" with a value of "0"
-- will be included into the listbox as the first item. This item will be
-- selected if the defaults provided contains the value 0.
--
-- When includeAny is true and the defaults contain the value "0", all other
-- items in the defaults will no longer be selected, since Any encompasses
-- everything.
local generateMultiSelectList = function (list, nameField, valueField, fieldName, defaults, includeAny)
    -- Open the select tag.
    html.form:openSelect (fieldName, "5", true, "toolDrop")

    -- Should we generate a fake "Any" entry?
    if includeAny then
        -- Figure out if the value for "Any" is in the array.
        local anyIsSelected = numberArrayContains (defaults, "0")

        -- Insert the any option, possibly selecting it.
        html.form:selectOption ("Any", "0", anyIsSelected)

        -- If the any item was selected, then reset the defaults to be nil,
        -- since there is no need to select anything else.
        if anyIsSelected then defaults = nil end
    end

    -- Iterate over all of the values in the list that we were given.
    for unused,info in list do
        -- Generate a select option for this item, and select it if it's in the defaults.
        html.form:selectOption (info[nameField] , info[valueField] , numberArrayContains (defaults, info[valueField]))
    end

    -- Done now
    html.form:closeSelect ()
end

---------------------------------------------------------------------------

-- This will generate a drop box that selects how many recent games (if any)
-- should not be selected by the random picker.
--
-- The default is 0, which means to randomly pick whatever you want.
--
-- Given some number, that number of session reports (starting with most
-- recent and going backwards) is selected, and the games listed there are
-- discounted from being selected by the random picker. This only happens
-- when random pick is selected.
local generateDropBoxForRandomPickAvoid = function (fieldName, currentValue)
    -- Open the select tag
    html.form:openSelect (fieldName , nil, nil, "toolDrop")

    -- This contains a list of things to be included. The index is the value that
    -- will be submitted, and the value is the text that will be displayed for
    -- that option.
    local values = {"-- Pick Any Game --", "Last game played", "Last 2 games played", "Last 3 games played", 
                    "Last 4 games played", "Last 5 games played", "Last 6 games played", "Last 7 games played" ,
                    "Last 8 games played", "Last 9 games played", "Last 10 games played", "Last 11 games played",
                    "Last 12 games played", "Last 13 games played", "Last 14 games played"}

    -- Convert the value provided to a number.
    currentValue = tonumber (currentValue) or 0

    -- Scan over all items in the values array and generate a dropdown using them.
    -- Along the way, we need to convert the indexes to be 0 based instead of 1 based.
    for index,name in values do
        html.form:selectOption (name, (index - 1), (index - 1) == currentValue)
    end
    
    -- Done now
    html.form:closeSelect ()
end

---------------------------------------------------------------------------

-- Takes an array of tables, which should have keys "id" and "name".
--
-- The return value is a table in which the keys are the fields
-- with the given id, and the values are the associated name.
--
-- This allows faster searching of fields.
local unpackTable = function (table)
    -- Our return value
    local retVal = {}

    -- Iterate over everything, and use the ID field as the key
    -- to insert the provided name field.
    for index,data in table do
        retVal[tostring (data.id)] = data.name
    end

    -- Return
    return retVal
end

---------------------------------------------------------------------------

-- Take information on a game, a list of categories that are desired and a
-- list of categories to be excluded. Both are arrays of numbers which
-- represent category ID's
--
-- IF:
--   All of the categories in the given list are marked as categories for this
--   game
-- AND
--   none of the categories in the exclusion list are marked as categories for
--   this game
-- THEN
--   the return value will be true.
--
-- false is returned if the game doesn't list a category in the category list,
-- or if it DOES list a category from the exclusion list.
local gameInCategory = function (db, gameInfo, categoryList, exclusionList)
    -- Get the complete list of categories that this game is in
    -- The result is an array of values with keys 'id' and 'name'.
    -- Unpack it so that it is a table with ID values as keys.
    local gameCategories = unpackTable (getAllCategoriesForGameInfo (db, gameInfo))

    -- Easy first test. If the list doesn't contain the value "0", then go
    -- ahead and test the contents. If it DOES contain 0, then this is going
    -- to be in the list no matter what, so don't bother wasting any time 
    -- doing this.
    if not numberArrayContains (categoryList, "0") then 
        -- Now, loop over the whole category list.
        for index, catID in categoryList do
            -- Convert the categoryID to a string.
            catID = tostring (catID)

            -- If this isn't "0", then handle it.
            if catID ~= "0" then
                -- If the game isn't in this list, then return
                -- false.
                if gameCategories[catID] == nil then return false end
            end
        end
    end

    -- Now, if we've gotten here, check to see if this game is in any of the
    -- categories in the exclusion list. In this case, if the game is listed, 
    -- then we return false, since in this case the list is telling us to NOT
    -- list certain things.
    for index, catID in exclusionList do
        -- Convert to a string.
        catID = tostring (catID)

        -- If the game is in this category, then return false. 
        if gameCategories[catID] then return false end
    end

    -- If we get here, it must be ok
    return true
end

---------------------------------------------------------------------------

-- Take information on a game, a list of designers that are desired and a
-- list of designers to be excluded. Both are arrays of numbers which
-- represent designer ID's
--
-- IF:
--   All of the designers in the given list are marked as designers for this
--   game
-- AND
--   none of the designers in the exclusion list are marked as designers for
--   this game
-- THEN
--   the return value will be true.
--
-- false is returned if the game doesn't list a designer in the designer list,
-- or if it DOES list a designer from the exclusion list.
local gameByDesigner = function (db, gameInfo, designerList, exclusionList)
    -- Get the complete list of designers that this game is in
    -- The result is an array of values with keys 'id' and 'name'.
    -- Unpack it so that it is a table with ID values as keys.
    local gameDesigners = unpackTable (getAllDesignersForGameInfo (db, gameInfo))

    -- Easy first test. If the list doesn't contain the value "0", then go 
    -- ahead and test the contents. If it DOES contain 0, then this is going
    -- to be in the list no matter what, so don't bother wasting time.
    if not numberArrayContains (designerList, "0") then 
        -- Now, loop over the whole desitner list.
        for index, desID in designerList do
            -- Convert the designerID to a string.
            desID = tostring (desID)

            -- If this isn't "0", then handle it.
            if desID ~= "0" then
                -- If the game isn't in this list, then return
                -- false.
                if gameDesigners[desID] == nil then return false end
            end
        end
    end

    -- Now, if we've gottenhere, check to see if this game is designed by anyone
    -- in the exclusion list. If the game is listed, then return false, because
    -- we want to filter out games by this designer.
    for index, desID in exclusionList do
        -- Convert the designerID to a string.
        desID = tostring (desID)

        -- If the designer is in the list, then return false.
        if gameDesigners[desID] then return false end
    end

    -- If we get here, it must be ok
    return true
end

---------------------------------------------------------------------------

-- This takes verified arguments from the game finder form, and returns back
-- a list of all games in the system that match the criteria given in them.
local gamesMatchingGameFinder = function (db, args)
    -- Get the complete list of games in the system, sorted according to
    -- game name.
    local gameList = getGameList (db, true)

    -- If we are going to be randomly picking a game, then the optional
    -- setting stored in "randomPickSkip" is a number that represents how
    -- many of the most recently played games should not be randomly picked
    -- again. 
    --
    -- If this is the case, then the first thing we want to do is get the list
    -- of games we want to discount.
    local skipGameIDs = {}
    args.randomPickSkip = tonumber (args.randomPickSkip) or 0
    if args.randomPick and args.randomPickSkip ~= 0 then
        skipGameIDs = getGamesFromRecentSessionReports (db, args.randomPickSkip)
    end
    
    -- This will store games that match the criterea we need as we're going
    -- along.
    local retVal = {}

    -- Loop over all games in the game list.
    for index, gameInfo in gameList do
        -- Make sure that all of the numeric fields in the game information we're
        -- using is converted to numbers for proper comparisons with the form
        -- data (which was converted to numbers when it was validated).
        gameInfo.gameID          = tonumber (gameInfo.gameID)
        gameInfo.minPlayers      = tonumber (gameInfo.minPlayers)
        gameInfo.maxPlayers      = tonumber (gameInfo.maxPlayers)
        gameInfo.playTimeMinutes = tonumber (gameInfo.playTimeMinutes)
        gameInfo.expandsGameID   = tonumber (gameInfo.expandsGameID)

        -- If this game is in the list of games to skip, then we don't want to do anything.
        -- So, only go forward if we aren't in the list. 
        if not numberArrayContains (skipGameIDs, gameInfo.gameID) then
            -- Only process this game if the game isn't an expansion. The game isn't an expansion
            -- if expandsGameID is set to 0.
            if gameInfo.expandsGameID == 0 then
                -- Make sure that the number of players in the game and the play time of the game fall
                -- within the critera that we want. If not, then we don't want this item.
                if args.playerCount >= gameInfo.minPlayers and 
                   args.playerCount <= gameInfo.maxPlayers and 
                   gameInfo.playTimeMinutes <= args.playTime then

                   -- This game meets the player and time requirements. If it's in the list of categories
                   -- and designers, and not in any of the excluded lists, then consider it for inclusion in
                   -- the list.
                   if gameInCategory (db, gameInfo, args.categories, args.excludeCategories) and 
                      gameByDesigner (db, gameInfo, args.designers, args.excludeDesigners) then

                      -- Must be good, insert it
                      table.insert (retVal, gameInfo)
                   end
                end
            end
        end
    end

    -- Are we supposed to randomly pick a game? If we are and there is more than one
    -- game in the list, then we want to randomly pick something to return.
    --
    -- This form argument is either present or not, so we can use it as a simple
    -- boolean.
    if args.randomPick and table.getn (retVal) > 1 then
        -- Make sure the random number generator is properly seeded, then pick a
        -- game.
        math.randomseed (os.time ())

        -- This will generate a random number from 1 to the number of games in
        -- retVal, inclusive.
        local result = math.random (table.getn (retVal))
    
        -- Replace the return value with the single selected item.
        retVal = {retVal[result]} or {}
    end

    -- All done now
    return retVal
end

---------------------------------------------------------------------------

-- Takes the arguments from a game finder form and validates them. If things
-- are ok, nil is returned. If not, error text is returned and any bogus
-- form fields are reset to have default values.
--
-- Any numeric fields are converted to numbers.
local validateGameFinderForm = function (db, args)
    -- Convert the player count and play time fields to numbers. If the fields
    -- values are not numeric, then tonumber returns nil, and so they will get the
    -- values of -1.
    args.playerCount = tonumber (args.playerCount) or -1
    args.playTime    = tonumber (args.playTime)    or -1

    -- Set up the random skip to be numeric.
    args.randomPickSkip = tonumber (args.randomPickSkip) or 0

    -- Player count has to be >= 1
    if args.playerCount < 0 then args.playerCount = nil return ERR_FINDER_BADPLAYERCOUNT end

    -- Play time has to be >= 1
    if args.playTime < 0 then args.playTime = nil return ERR_FINDER_BADPLAYTIME end

    -- Must be ok
    return nil
end

---------------------------------------------------------------------------

-- Generates the finder form. If args is non-nil, it's assumed to be form
-- arguments from a previous submission, so that we can rebuild the form the
-- same way a second time.
--
-- Any fields within the args can be nil, which will cause a default to be
-- used instead. This is used for the first time generation, and also to
-- put in default when a numeric field in the submitted form is invalid.
local generateGameFinderForm = function (db, args)
    -- If args is nil, then make it an empty table. That way we can index
    -- it no matter what, and just worry about missing fields. We have to be
    -- careful here regarding the random pick button. We want it to be checked
    -- by default to begin with, but remain unchecked if someone unchecks it.
    -- 
    -- Since this field is optional, it will show up in the args as either nil
    -- or some value. Thus we can't distinguish between "I need to be on by
    -- default" and "Someone unchecked me" easily.
    --
    -- To get around this, we presume that if we have been given arguments, the
    -- default value should be whatever was stored in the args table. If we haven't
    -- been given arguments, then default it to on.
    local randomSetting
    if args then 
        randomSetting = args.randomPick
    else
        randomSetting = true
        args = {}
    end

    -- Present the form that will give us the criteria we need.
    html.form:open (env.scriptName , "POST")
    html.table:open (nil, nil, nil, nil, "5", "toolList")

    -- How many players will be playing the game
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Players:")
    html.table:cell (true, nil, nil, "LEFT", "TOP")
    html.form:text ("playerCount" , args.playerCount or "2", "5" , "40" , "toolText")

    -- How long do we want to play in minutes?
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Time (minutes):")
    html.table:cell (true, nil, nil, "LEFT", "TOP")
    html.form:text ("playTime" , args.playTime or "240", "5" , "40" , "toolText")
    html.table:newRow ()

    -- What categories are ok?
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Categorized as:")
    html.table:cell (true, nil, nil, "LEFT", "TOP")
    local list = getCategoryList (db, true, true)
    generateMultiSelectList (list, "categoryName", "categoryID", "categories" , args.categories or {"0"}, true)

    -- What categories to filter?
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("but not:")
    html.table:cell (true, nil, nil, "LEFT", "TOP")
    generateMultiSelectList (list, "categoryName", "categoryID", "excludeCategories" , args.excludeCategories or {}, false)
    html.table:newRow ()

    -- What designers are ok?
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Designed by:")
    html.table:cell (true, nil, nil, "LEFT", "TOP")
    list = getDesignerList (db, true, true)
    generateMultiSelectList (list, "designerName", "designerID", "designers" , args.designers or {"0"}, true)

    -- What designers to filter?
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("but not:")
    html.table:cell (true, nil, nil, "LEFT", "TOP")
    generateMultiSelectList (list, "designerName", "designerID", "excludeDesigners" , args.excludeDesigners or {}, false)
    html.table:newRow ()

    -- Should the game be chosen at random?
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Random Pick:")
    html.table:cell (true, nil, nil, "LEFT", "TOP")
    html.form:checkbox ("randomPick", "YES", randomSetting)

    -- How many recent games to discount because they have been played recently?
    html.table:heading (true, nil, nil, "LEFT", "TOP")
    html:addText ("Don't repick: ")
    html.table:cell (true, nil, nil, "LEFT", "TOP")
    generateDropBoxForRandomPickAvoid ("randomPickSkip" , args.randomPickSkip or "7")
    html.table:newRow ()

    -- The submit button and other hidden fields.
    html.table:cell (true, nil, nil, "CENTER", "TOP", nil, nil, nil, "4")
    html.form:submit ("Button", "Find games matching this criteria", "tool")
    html.form:hidden ("command", "ToolGameFinder")

    -- All done
    html.table:close ()
    html.form:close ()

    -- Seperate the form from the possible results that will shortly be displayed.
    html:addText ("<P>")
end

---------------------------------------------------------------------------

-- This is the game finder tool. This takes some criteria on games, and
-- selects one to be played.
function commandToolGameFinder (db)
    -- Header for the page
    generatePageHeader (thisPageSection, "Find a game to play", "toolsGameFinder")

    -- A button to take us back to the main tools page.
    html.form:singleButtonForm ("Return to Tools" , "Tool" , env.scriptName , "POST" , nil, nil, "tool")
    html:addText ("<P>")

    -- Try to get the form arguments. The elements for random pick, and the list of excluded 
    -- categories and designers are optional because if nothing is selected in them, they don't
    -- get submitted.
    local args = collectFormArguments ({"playerCount", "playTime", "categories" , "designers", "randomPickSkip"}, {"randomPick", "excludeCategories", "excludeDesigners"})

    -- If we got arguments, then process the form.
    if args then
        -- These items might be nil if nothing was selected in them. We want them to be empty
        -- tables in that case, so make sure they're ok.
        args.excludeCategories = args.excludeCategories or {}
        args.excludeDesigners  = args.excludeDesigners or {}

        -- The lists of categories, excluded categories, designers and excluded designers are
        -- multi select list boxes, so their values may be strings if only a single thing was
        -- selected, or tables for multiples. We alays want tables.
        
        -- Make sure that categories and designers are tables if only one item was selected.
        if type (args.categories)        ~= "table" then args.categories        = {args.categories} end
        if type (args.designers)         ~= "table" then args.designers         = {args.designers} end
        if type (args.excludeCategories) ~= "table" then args.excludeCategories = {args.excludeCategories} end
        if type (args.excludeDesigners)  ~= "table" then args.excludeDesigners  = {args.excludeDesigners} end

        -- Take our form arguments and validate them. This makes sure that the number values given are
        -- within a proper range. If this returns non-nil, then it is error text saying what was wrong.
        -- In this case, any offending arguments will have their values changed to a default.
        --
        -- We set a flag in the arguments saying that they're broken, so that later on we know if we
        -- should use them to do something.
        local result = validateGameFinderForm (db, args)
        if result then
            html:addText (errorText (result), "<P>")
            args.broken = true
        end
    end

    -- Generate the finder form using the form arguments we have now. If args is
    -- nil, then all defaults are used. Otherwise, the values in the form arguments
    -- are used to generate the form. This keeps it the same across submissions.
    generateGameFinderForm (db, args)

    -- Now, if we have no arguments, or if we do have arguments but something was wrong
    -- with them, we can stop now.
    if args == nil or args.broken then return end

    -- The arguments are ok. Find all games that match the criteria, and then feed
    -- that list into the function that will use it to generate a game list.
    local list = gamesMatchingGameFinder (db, args)
    generateGameList (db, list)
end

---------------------------------------------------------------------------

-- This displays a list of possible tools as single button forms and waits for
-- someone to select one.
function commandTool (db)
    -- Header for the page
    generatePageHeader (thisPageSection, nil, "toolsBase")

    -- Generate the tools. Each is a single button form.
    -- Some common tools
    html.form:singleButtonForm ("Find a game to play" , "ToolGameFinder" , env.scriptName , "POST" , nil, nil, "tool")
    html:addText ("<P>")
end

---------------------------------------------------------------------------
--- MAIN CODE -------------------------------------------------------------
---------------------------------------------------------------------------

