#!./lua-web2

---------------------------------------------------------------------------
--- CONFIGURATION ---------------------------------------------------------
---------------------------------------------------------------------------

-- The configuration file needs to be loaded up before we do anything.
-- To start with, set LUA_PATH to be the current directory only, then load
-- the file.
LUA_PATH = "./?"
gamedbConfig = require ("gamedb.cfg")

-- Now that the configuration is set up, we need to re-set the lua path
-- to where we will be able to find our other script files.
LUA_PATH = gamedbConfig.codeDir .. "?.lua"

-- The current version of gamedb.
GAMEDB_VERSION="1.0.3"

---------------------------------------------------------------------------
--- MODULES ---------------------------------------------------------------
---------------------------------------------------------------------------

-- All modules loaded here will have no access to the database. All they
-- can do is set up modules/variables. After all are loaded and setup is
-- complete, the main code can execute any methods in here that will pass
-- it the database and configuration information.
--
-- Also note that all code in modules has no access to any variables
-- declared as local, and cannot execute any function in their main
-- section that has not already been loaded.
require "helper"
require "login"
require "menu"
require "query"
require "player"
require "logs"
require "config"
require "designer"
require "publisher"
require "category"
require "game"
require "session"
require "tools"
require "cmdmap"

---------------------------------------------------------------------------
--- SETUP -----------------------------------------------------------------
---------------------------------------------------------------------------

-- Open the database
local db,errStr = sqlite3.open (gamedbConfig.databaseFilename)

-- Set a temporary title for the page, which will only be used if the page
-- can't be generated because the database didn't open.
html:setTitle ("LUAweb database error")

-- Set up the stylesheet action now, while it's still safe.
html:setStylesheet (gamedbConfig.cssURL)

-- Make sure the DB opened ok.
if checkDB (db, errStr) == false then return end

-- Get the config information.
--
-- This configuration comes from the database itself and controls
-- the behaviour of the script dynamically.
if getConfigInfo (db) == false then return end

-- We have config information, so change the page tile properly.
-- Change the page title now.
html:setTitle (configInfo.pageTitle)

-- Determine who the logged in user is. This returns one of 
-- three values:
--  "error" means that there is an authToken, but it specifies a 
--          user that doesn't exist. We have requested to erase the
--          authToken, but a different login will be needed.
--
--  "needLogin" means that there was no authToken cookie, so we need
--  for someone to log in.
--
--  "loggedIn" means that we have set up the userInfo table properly
-- and page generation can continue as needed. 
local result = getLoginUser (db)

-- If there is currently someone logged in, and there is a form submission
-- that is telling us to log out, then log the user out, add the event
-- to the system log, and make things seem as if there is no user
-- at all logged in.
if result == "loggedIn" and CGI.formData.command == "Logout" then 
    addLog (db, "Logged out");
    html:addText (successText ("You have been logged out.") , "<P>")
    logoutUser (db) 
    result = "needLogin"
end

-- If the result isn't "loggedIn", then we need to display some login 
-- stuff.
if result ~= "loggedIn" then
    -- Generate a login page now
    html:setTitle (html:getTitle () .. ": Login")
    generateLoginForm ()

    -- If the result is error, then display an error before displaying the
    -- login form.
    if result == "error" then
        html:addText (errorText ("Your login information is not valid. Your login may " ..
                      "have timed out, or it may have been logged out or had " ..
                      "it's credentials changed from another browser. Please " ..
                      "log in again.") , "<P>")
    end
    donePage ()
    return
end

---------------------------------------------------------------------------
--- MAIN CODE -------------------------------------------------------------
---------------------------------------------------------------------------

-- Set up the default command in the form if there isn't one there already.
-- When no command is specified, we get the one from the config info. This
-- only happens on an initial login.
CGI.formData.command = CGI.formData.command or configInfo.defaultPage or "None"

-- Stash the command for easy access.
local command = CGI.formData.command 

-- The main part of the page is a table with 2 columns. The first is a 
-- menu and the second is the rest of the page. Start off with a table
html.table:open ("0" , "100%" , nil , nil , "6")
html.table:cell (nil, "1%" , nil, "LEFT" , "TOP")

-- Generate the menu on the left side of the page
generateMenu ();

-- Now start the second cell for the main page content
html.table:cell (nil, "99%" , nil , "LEFT" , "TOP")

-- If the command given is in the command map, then use that
-- function to generate the body of the page. We have a fallback
-- if there is no mapping for the command, but if that is also not
-- set up properly, handle things somewhat gracefully.
local pageGenFunc = commandMap[command] or commandMap["ErrorFallback"]
if pageGenFunc ~= nil then 
    pageGenFunc (db)
else
    html:addText (errorText ("I don't know what to do, so I'm not going to do anything.") , "<P>")
end

-- Close the table now
html.table:close ()

-- Close the database and end the page
donePage ()

