#!./lua-web2

---------------------------------------------------------------------------
--- HOW HELP WORKS                                                      ---
---------------------------------------------------------------------------
--
-- The core of the help system is a simple system which ties help topics
-- to a function that knows how to generate help for that topic. Each help
-- generator gets two paramters, the first being the topic that it should
-- generate help for, and the second being the base URL to use if help
-- needs to generate a link to another topic. 
--
-- The help generator should use the regular HTML functions to add html
-- to the outgoing page to generate the help.
--
-- This core code maps topics to the functions that know how to generate
-- the help for them, and also does the work of setting up some page 
-- content that wraps the help, so that all help appears in a uniform way.
--
---------------------------------------------------------------------------

---------------------------------------------------------------------------
--- CONFIGURATION ---------------------------------------------------------
---------------------------------------------------------------------------

-- The configuration file needs to be loaded up before we do anything.
-- To start with, set LUA_PATH to be the current directory only, then load
-- the file.
LUA_PATH = "./?"
gamedbConfig = require ("gamedb.cfg")

-- Now that the configuration is set up, we need to re-set the lua path
-- to where we will be able to find our other script files.
LUA_PATH = gamedbConfig.helpDir .. "?.lua"

---------------------------------------------------------------------------
--- MODULES ---------------------------------------------------------------
---------------------------------------------------------------------------

-- These pull in the required help files for various sections.
require ("index")
require ("playerHelp")
require ("designHelp")
require ("publishHelp")
require ("categoryHelp")
require ("gameHelp")
require ("sessionHelp")
require ("toolHelp")
require ("configHelp")
require ("logHelp")

---------------------------------------------------------------------------
--- SETUP -----------------------------------------------------------------
---------------------------------------------------------------------------

-- Set up the stylesheet action now, while it's still safe.
html:setStylesheet (gamedbConfig.cssURL)

-- We have config information, so change the page tile properly.
-- Change the page title now.
html:setTitle ("GameDB Documentation")

---------------------------------------------------------------------------
--- FUNCTIONS -------------------------------------------------------------
---------------------------------------------------------------------------

-- Allows generated help pages to change the title of the page.
function setPageTitle (title)
    html:setTitle ("GameDB: " .. title)
end

---------------------------------------------------------------------------

-- Used to handle the situation of there being no help available for a
-- particular topic.
local noHelpTopic = function (topic, helpURL)
    html:addText ("There is no help available for the help topic '", topic , "'<p>")
end

---------------------------------------------------------------------------

-- Used as the help when the page generated was generated because an unknown command
-- was given.
local badCommandGiven = function (topic, helpURL)
    html:header (3, "Unknown Command")

    html:addText ([[
You are seeing this page because GameDB was given a command that it does not understand.<p>

If you manually entered the URL, this is most likely the issue. You should use the menu items
on the left to navigate around the site.<p>

If you got here by clicking on a link on a GameDB page, then please contact your administrator
and tell then what you were attempting to do when you got this error.
    ]])
end

---------------------------------------------------------------------------

-- This is a special link generator which outputs links wrapped in square
-- brackets and nonbreaking spaces, for use ine control bar generation.
--
-- If the url passed in is nil, then no link is generated and just regular
-- text is added instead.
function wrappedLink (url, text, target)
    html:addText ("&nbsp; [ &nbsp;")
    if url then html:link (url, text, target) else html:addText (text) end
    html:addText ("&nbsp; ]")
end

---------------------------------------------------------------------------

-- Maps help topics to the function that will return the HTML for that help.
local helpMap =
{
    -- This special fallback topic is used when the actual help topic desired
    -- is unavailable.
    ["ErrorFallback"] = noHelpTopic,
    ["badCommand"]    = badCommandGiven,

    -- Indexes
    ["index"] = helpIndex,     
    
    -- Help for the player section of the page.
    ["player"]              = basePlayerHelp,
    ["playerProfile"]       = playerProfileHelp,
    ["playerAdd"]           = playerAddHelp,
    ["playerView"]          = playerViewProfileHelp,
    ["playerProperties"]    = playerPropertiesHelp,

    -- Help for the designers section of the page
    ["designerBase"]        = designerBaseHelp,
    ["designerProperties"]  = designerPropertiesHelp,
    ["gamesByDesigner"]     = gamesByDesignerHelp,

    -- Help for the publishers section of the page
    ["publisherBase"]       = publisherBaseHelp,
    ["publisherProperties"] = publisherPropertiesHelp,
    ["gamesByPublisher"]    = gamesByPublisherHelp,

    -- Help for the categories section of the page
    ["categoryBase"]        = categoryBaseHelp,
    ["categoryProperties"]  = categoryPropertiesHelp,
    ["gamesByCategory"]     = gamesByCategoryHelp,

    -- Help for the games section of the page
    ["gamesBase"]           = gamesBaseHelp,
    ["gamesAdd"]            = gamesAddHelp,
    ["gamesEdit"]           = gamesEditHelp,
    ["gamesView"]           = gamesViewHelp,
    ["gamesExpansions"]     = gamesExpansionsHelp,

    -- Help for the sessions section of the page
    ["sessionsBase"]          = sessionsBaseHelp,
    ["sessionsAdd"]           = sessionsAddHelp,
    ["sessionsEdit"]          = sessionsEditHelp,
    ["sessionsView"]          = sessionsViewHelp,
    ["sessionsForGame"]       = sessionsForGameHelp,
    ["sessionsForPlayerWin"]  = sessionsForPlayerWinHelp,
    ["sessionsForPlayerLoss"] = sessionsForPlayerLossHelp,

    -- Help for the tools section of the page
    ["toolsBase"]           = toolsBaseHelp,
    ["toolsGameFinder"]     = toolsGameFinderHelp,

    -- Help for the configuration section of the page
    ["configHelp"]          = configHelp,

    -- Help for the activity log section of the page
    ["logsView"]            = logsViewHelp,
    ["logsDelete"]          = logsDeleteHelp,
}

---------------------------------------------------------------------------
--- MAIN CODE -------------------------------------------------------------
---------------------------------------------------------------------------

-- Figure out what help topic we want to display.
local helpTopic = CGI.formData.topic or "default"

-- Figure out the URL that we need to submit to if we want to generate a
-- link to some other help topic. We will need to append the actual topic
-- to this when it is used.
local baseHelpURL = gamedbConfig.baseURL .. "gamedbhelp.lua?topic="

-- Get the command that will generate the help html. We use a fallback
-- if the topic is unknown
local pageGenFunc = helpMap[helpTopic] or helpMap["ErrorFallback"]

-- Create the header table, which will provide help links
html.table:open ("0" , nil , nil , nil , "6", "helpTable")
html.table:cell (nil, nil , nil, "LEFT" , "TOP")
wrappedLink (baseHelpURL .. "index", "Help Index")
wrappedLink ("javascript:history.back ()", "Back")
html.table:close ()
html:addText ("<br>")

-- The actual help table
html.table:open ("0" , nil , nil , nil , "6", "helpTable")
html.table:cell (nil, nil , nil, "LEFT" , "TOP")

-- Generate the help
pageGenFunc (helpTopic, baseHelpURL)

-- Done now
html.table:close ()
html:endPage ()

---------------------------------------------------------------------------

