#!./lua-web2

-- This will import a file named import.txt that is in the
-- format of the export format of moveable type 3.17. 
-- You will get game name, description, gamegeekid and
-- categories, but that's it.
--
-- Takes care not to insert the same category multiple times,
-- but the same can't be said of games.

-- Load up the main config file, so that we can get the database
-- filename.
LUA_PATH = "./?"
cfg = require ("gamedb.cfg")

function prefix (str, text)
    local start,stop = string.find (str, text, 1, true)
    if start == 1 then return true else return false end
end

function stripprefix (str, pre)
    local result = string.sub (str , string.len (pre) + 1)
    return string.trimleadingtrailing (result)
end

function shouldIgnore (line)
    if line == "" or
       line == "-----" or
       prefix (line, "AUTHOR") or
       prefix (line, "STATUS") or
       prefix (line, "ALLOW COMMENTS") or
       prefix (line, "CONVERT BREAKS") or
       prefix (line, "ALLOW PINGS") or
       prefix (line, "PRIMARY CATEGORY") or
       prefix (line, "DATE") or
       prefix (line, "EXCERPT") or
       prefix (line, "KEYWORDS") or
       prefix (line, "BODY") 
    then return true end

    return false
end

function addCategory (name, list)
    -- Skip this recent entry
    if name == "RecentEntry" then return end

    -- Scan, leave if an item has this name
    for index,value in list do
        if value == name then return end
    end

    -- Insert this
    table.insert (list, name)
end

-- Handle the input line. If the record in thisEntry is
-- complete, then return true, else return false.
function handle (line, thisEntry, categoryList)
    -- This means that we're done with this entry
    if line == "--------" then 
        thisEntry.bodyComplete = nil
        return true 
    end

    -- This means we found the name of the game
    if prefix (line, "TITLE:") then
        title = stripprefix (line, "TITLE:")
        thisEntry.name = title
        return false
    end

    -- This means we found the name of a category this game is in
    if prefix (line, "CATEGORY:") then
        category = stripprefix (line, "CATEGORY:")
        addCategory (category, categoryList)
        return false
    end

    -- This means that the extended body is beginning, so the body
    -- is ending. The body is the description of the game.
    if prefix (line, "EXTENDED BODY:") then
        thisEntry.bodyComplete = true
        return false
    end

    -- Here, raw text. If the body is complete, then tack this onto
    -- the geekid. Otherwise, tack it to the description.
    if thisEntry.bodyComplete then
        thisEntry.geekid = thisEntry.geekid .. line
        return false
    else
        thisEntry.description = thisEntry.description .. line
        return false
    end

    html:addText ("Unhandled Line!->" , line , "<BR>")
    return false
end

local db,errStr = sqlite3.open (cfg.databaseFilename)

function donePage ()
    db:close ()
    html:endPage ()
end

-- Set the title of the page
html:setTitle ("DB Web Page")

-- If the Database wasn't opened, then complain and return.
if db == nil then
    html:addText ("Unable to open db: " , errStr)
    donePage ()
    return
end

-- Try to open the input file
inputHandle,errStr = io.open ("gamelist.txt", "rt")
if inputHandle == nil then
    html:addText ("Unable to open import file:" , errStr)
    donePage ()
    return
end

-- Iterate over lines
local gameList = {}
local categoryList = {}
local thisEntry = {["geekid"] = "", ["description"] = ""}
for input in inputHandle:lines () do
    -- Fix line
    line = string.trimleadingtrailing (input)

    -- If we shouldn't ignore this line, then do something with it
    if shouldIgnore (line) == false then
        if handle (line, thisEntry, categoryList) == true then
            thisEntry.geekid = tonumber (thisEntry.geekid) or 0
            thisEntry.categories = categoryList
            table.insert (gameList, thisEntry)
            categoryList = {}
            thisEntry = {["geekid"] = "", ["description"] = ""}
        end
    end
end

-- Close the file now
inputHandle:close ()

-- Make a global list of all categories
local globalCategories = {}
for index,gameInfo in gameList do
    for innerIndex,categoryItem in gameInfo.categories do
        addCategory (categoryItem, globalCategories)
    end
end
html:addText ("<P>", table.getn (gameList) , " games found to import in " , table.getn (globalCategories) , " total categories.<P>")

-- Scan over all categories, and add them to the database if they're not
-- there already. End up with a table in gameCategories where key == categoryID
-- and value == categoryName
local gameCategories = {}
for index,category in globalCategories do
    -- Try to get the info for this category
    info = db:SQLfirstRow ("SELECT * FROM CategoryList WHERE categoryName = ?" , category)

    if info ~= nil then
        -- It worked. If the gameCategories table doesn't already have an entry for this
        -- categoryID, then add it
        if gameCategories[info.categoryID] == nil then
            gameCategories[info.categoryID] = category
        end
    else
        -- Didn't work. Insert a new one, then capture the rowID
        db:execSQL ("INSERT INTO CategoryList VALUES (NULL, ?)", category)
        categoryID = db:last_insert_rowid ()

        if gameCategories[categoryID] == nil then
            gameCategories[categoryID] = category
        end
    end
end

-- Invert the gameCategories table so that the key is name
local tmp = {}
for key,value in gameCategories do
    tmp[value] = key
end
gameCategories = tmp

-- Loop over all items in the gameList
for index,gameInfo in gameList do
    -- Primary category is 0 if there is no list of categories
    primaryCategoryID = "0"

    -- Get the first item out of the category list if there is one.
    -- If there is, try to reverse lookup to get the info that we want.
    if table.getn (gameInfo.categories) > 0 then
        primaryCategoryName = table.remove (gameInfo.categories , 1)
        primaryCategoryID = gameCategories[primaryCategoryName] or "0"
    end

    -- Now, insert this game.
    db:execSQL ("INSERT INTO GameList VALUES (NULL, ?, ?, ?, 0, 0, ?, 2, 5, 60, 0)" , 
                gameInfo.geekid, gameInfo.name, gameInfo.description, primaryCategoryID)
    gameID = db:last_insert_rowid ()

    -- Loop over the rest of the categories and add category placements for them
    for catInd, catName in gameInfo.categories do
        categoryID = gameCategories[catName] or nil

        if categoryID ~= nil then
            db:execSQL ("INSERT INTO CategoryPlacement VALUES (NULL, ?, ?)" , gameID, categoryID)
        end
    end
    
end

html:addText ("Import is complete")

-- Close the database and end the page
donePage ()

